function mat2gif(mat,path,varargin)
% creates a gif from a matrix
%
%   mat2gif(mat,fs,path)
%   mat2gif(mat,fs,path,map)
%
%   REQUIRED
%   mat             - (3d numeric) 
%                     Input 3d matrix.
%   path            - (char)
%                     Output path. Include .gif in the path.
%   PARAMETERS
%   fps             - (numeric scalar)
%                     Video fps. Default is 30.
%   sampling_freq   - (numeric_scalar)
%                     The sampling frequency of the frames in Hz. 
%                       playback_speed must also be supplied if using 
%                     sampling_freq.
%                       fps is calculated using playback_speed, but has a
%                     maximum value of 30 fps. If playback_speed gives a
%                     higher fps than the limit frames will be skipped. 
%   playback_speed  - (numeric scalar)
%                     Playback speed as a factor of realtime, sampling_freq
%                     must be supplied. 
%   map             - (256x3 double/single)
%                     Matrix of a colormap. 
%                       default : begonia.colormaps.viridis(256)
%   limits          - (1x2 numeric) 
%                     The two values that will be mapped to 0 and 256. 
%                       default : [min(mat(:)),max(mat(:))]
p = inputParser;
p.addRequired('mat',...
    @(x) validateattributes(x,{'logical','numeric'},{'3d'}));
p.addRequired('path',...
    @(x) validateattributes(x,{'char'},{'nonempty'}));
p.addOptional('fps',30,...
    @(x) validateattributes(x,{'numeric'},{'scalar'}));
p.addParameter('sampling_freq',[],...
    @(x) validateattributes(x,{'numeric'},{'scalar'}));
p.addParameter('playback_speed',[],...
    @(x) validateattributes(x,{'numeric'},{'scalar'}));
p.addParameter('map',begonia.colormaps.viridis(256),...
    @(x) validateattributes(x,{'single','double'},{'size',[256,3]}));
p.addParameter('limits',[min(mat(:)),max(mat(:))], ...
    @(x) validateattributes(x,{'numeric'},{'numel',2}));
p.parse(mat,path,varargin{:});
begonia.util.dump_inputParser_vars_to_caller_workspace(p);

if ~isempty(sampling_freq) || ~isempty(playback_speed)
    assert(~isempty(sampling_freq),'sampling_freq must be supplied.');
    assert(~isempty(playback_speed),'playback_speed must be supplied.');

    max_fps = fps;
    st = 1;
    en = size(mat,3);

    fps = round(sampling_freq*playback_speed);
    
    if fps > max_fps
        r = fps/max_fps;
        fps = max_fps;
        I = round(st:r:en);
    else
        I = st:en;
    end
    mat = mat(:,:,I);
end

dt = 1/fps;

mat = begonia.mat_functions.normalize(mat,'type','uint8','limits',limits);

begonia.path.make_dirs(path);

for i = 1:size(mat,3)
    im = mat(:,:,i);
    if i == 1
        imwrite(im,map,path,'gif','LoopCount',Inf,'DelayTime',dt);
    else
        imwrite(im,map,path,'gif','WriteMode','append','DelayTime',dt);
    end
end

end

