classdef PrairieOutput < handle ...
        & begonia.data_management.DataLocation ...
        & begonia.data_management.DataInfo ...
        & dynamicprops
    
    properties
        uuid
        name
        type
        source
        cycles
        channels
        channel_names
        frames_in_cycle
        original_img_dim
        rasters_pr_frame
        dx
        dy
        dt
        optical_zoom
        record_date
        processing_state
        format
        xml_file
        metadata
        
        % DataInfo props:
        start_time_abs
        duration
        time_correction
    end
    
    methods
        function obj = PrairieOutput(path)
            import begonia.util.logging.*;
            import begonia.scantype.prairie.*;
            
            obj@begonia.data_management.DataLocation(path);

            if path(end) == filesep
                path(end) = [];
            end
            
            assert(exist(fullfile(path,'References'),'dir') == 7, ...
                'begonia:load:missing_Ref_folder', ...
                '''References'' folder not found, used as an indicator for PrarieOutput.');
            
            xml_files = begonia.path.find_files(path,'.xml',false);
            assert(~isempty(xml_files),'begonia:load:missing_xml','Missing xml file.');
            
            % ensure we have a uuid
            obj.dl_ensure_has_uuid();
            obj.uuid = obj.dl_unique_id;
            
            xml_file = xml_files{1};
            try
                meta = PrairieOutput.parse_xml_file_fast(xml_file);
            catch err
                error('begonia:load:faulty_xml', err.message);
            end
            obj.metadata = meta;

            obj.source = 'Prairie';
            obj.type = meta.type;
            obj.rasters_pr_frame = meta.rasters_pr_frame;
            obj.dx = meta.dx;
            obj.dy = meta.dx;
            obj.dt = meta.dt * obj.rasters_pr_frame;
            obj.optical_zoom = meta.optical_zoom;
            obj.cycles = meta.cycles;
            obj.frames_in_cycle = meta.frames_in_cycle;
            
            obj.channels = meta.channels;
            obj.record_date = meta.date;
            obj.channel_names = meta.channel_names;
            obj.path = path;
            obj.original_img_dim = meta.original_img_dim; 

            % Set the name of the stack from the directory.
%             name = strsplit(path,filesep);
%            name = strsplit(xml_file, filesep)
            [~,name,~] = fileparts(xml_file);
%            name = name{end};
            obj.name = char(name);
            obj.xml_file = xml_file;
            
            obj.processing_state = 'raw';
            obj.format = 'ome-xml';
            
            obj.load_timeinfo();
        end
        
        function load_timeinfo(obj)
        
   
            id = 'MATLAB:datetime:AmbiguousDateString';
            warning('off',id);
            if any(strcmp(obj.saved_vars, 'time_correction'))
                obj.time_correction = obj.load_var('time_correction');
            end
            try
                switch obj.metadata.type
                    case 'Linescan'
                        obj.start_time_abs = obj.metadata.date;
                        dur = seconds(obj.metadata.relative_times(end)+obj.metadata.line_period(end)*obj.metadata.linesPerFrame(end));
                        dur.Format = 'hh:mm:ss.SSS';
                        obj.duration = dur;
                    case 'TSeries Timed Element'
                        obj.start_time_abs = obj.metadata.date;
                        dur = seconds(obj.metadata.frames_in_cycle*obj.metadata.dt(1));
                        dur.Format= 'hh:mm:ss.SSS';
                        obj.duration = dur;
                    case 'ZSeries'
                        obj.start_time_abs = obj.metadata.date;
                        dur = seconds(obj.metadata.relative_times(end)+obj.metadata.dt(1));
                        dur.Format= 'hh:mm:ss.SSS';
                        obj.duration = dur;
                    case 'Single'
                        obj.start_time_abs = obj.metadata.date;
                        dur = seconds(obj.metadata.dt);
                        dur.Format= 'hh:mm:ss.SSS';
                        obj.duration = dur;
                end
            catch
                warning(['Error loading timeinfo for ' obj.metadata.type ' (' obj.path ')']);
            end
        end
    end
    
    methods (Static)
        
        function metadata = parse_xml_file_fast(xml_path)
            assert(exist(xml_path, 'file') == 2,'xml file does not exist.');
            
            text = fileread(xml_path);
            
            temp = regexp(text,'(?<="framePeriod" value=").*?(?=")','match');
            if length(temp) > 1
                metadata.dt = str2double(temp(2:end));
            else
                metadata.dt = str2double(temp(1:end));
            end

            temp = regexp(text,'(?<="linesPerFrame" value=").*?(?=")','match');
            metadata.linesPerFrame = temp(2:end);
            metadata.linesPerFrame = cellfun(@(x) str2num(x), metadata.linesPerFrame);

            temp = regexp(text,'(?<=<PVStateValue key="micronsPerPixel">).*?(?=</PVStateValue>)','match');
            temp = regexp(temp,'(?<="XAxis" value=").*?(?=")','match');
            metadata.dx = str2double(temp{1});

            temp = regexp(text,'(?<=<PVStateValue key="micronsPerPixel">).*?(?=</PVStateValue>)','match');
            temp = regexp(temp,'(?<="YAxis" value=").*?(?=")','match');
            metadata.dy = str2double(temp{1});

            temp = regexp(text,'(?<=type=").*?(?=")','match');
            metadata.type = temp{1};

            temp = regexp(text,'(?<="opticalZoom" value=").*?(?=")','match');
            metadata.optical_zoom = str2double(temp{1});

            temp = regexp(text,'(?<="pixelsPerLine" value=").*?(?=")','match');
            metadata.original_img_dim = [str2num(temp{1}) str2num(temp{1})];

            temp = regexp(text,'(?<=Freehand x=").*?(?=")','match');

            metadata.line_x = cellfun(@(x) str2double(x), temp);

            temp = regexp(text,'(?<= y=").*?(?=")','match');
            metadata.line_y = cellfun(@(x) str2double(x), temp);

            temp = regexp(text,'(?<=key="scanLinePeriod" value=").*?(?=")','match');
            metadata.line_period = cellfun(@(x) str2double(x), temp(2:end));

            temp = regexp(text,'(?<=relativeTime=").*?(?=")','match');
            metadata.relative_times = cellfun(@(x) str2double(x), temp);

            temp = regexp(text,'(?<=absoluteTime=").*?(?=")','match');
            metadata.absolute_times = cellfun(@(x) str2double(x), temp);
            
            % number of frames averaged for each frame:
            temp = regexp(text,'(?<="rastersPerFrame" value=").*?(?=")','match');
            metadata.rasters_pr_frame = cellfun(@(x) str2double(x), temp);

            temp = regexp(text,'(?<=date=").*?(?=")','match');
            metadata.date = datetime(temp,'InputFormat','MM/dd/uuuu hh:mm:ss aa');
            if ~isempty(metadata.absolute_times)
                metadata.date = metadata.date + seconds(metadata.absolute_times(1));
            end


            % Matches any characters that are not whitespace (\S) which is
            % also between 'filename="' and '"' .
            temp = regexp(text,'(?<=filename=")\S*?\.ome\.tif(?=")','match');
            metadata.files = temp;

            
            tmp_cy = cell(length(temp),1);
            tmp_ch = cell(length(temp),1);
            for i = 1:length(temp)
                tmp_cy{i} = temp{i}(end-23:end-19);
                tmp_ch{i} = temp{i}(end-17:end-15);
            end
            metadata.cycles = length(unique(tmp_cy));
            tmp_ch = unique(tmp_ch);
            metadata.channels = length(tmp_ch);
            metadata.channel_names = tmp_ch;
            metadata.files = reshape(metadata.files,metadata.cycles,metadata.channels,[]);
            % The previous code of sorting the files brakes if the cycles
            % dont have equally many frames, so we juse use that data to
            % create frames_in_cycle.
            metadata.frames_in_cycle = repmat(size(metadata.files,3),1,metadata.cycles);
            
        end
        
    end
end

