classdef GPNGReader < handle
    %PNGSTACK Summary of this class goes here
    %   Detailed explanation goes here
    
    properties
        path
        ds_ratio
        max_cache
        image_cache = [];
        config
        datatype
    end
    
    methods
        function obj = GPNGReader(path, max_cache)

            obj.path = path;

            % read json data:
            configpath = fullfile(path, 'mpng_config.json');
            json = fileread(configpath);
            config = jsondecode(json);
            obj.config = config;
            obj.ds_ratio = obj.config.ds_ratio;
            
            % default max cache is same as dim:            
            if nargin < 2
                max_cache = 10;
            end
            %obj.max_cache = max_cache;
            
            obj.datatype = class(obj.read_frame(1));
        end
       
        function mat = read_interpolated_frame(obj, fnum)
            % convert the frame requested based on downsampling ratio
            dsfnum = (fnum - 1)/obj.ds_ratio;
            lower_fnum = floor(dsfnum) + 1;
            upper_fnum = ceil(dsfnum) + 1;
            
            lower_real_frame = ((lower_fnum - 1) * obj.ds_ratio) + 1;
            upper_real_frame = ((upper_fnum - 1) * obj.ds_ratio) + 1;
            
            % extract the two closest frames 
            mat_low = obj.read_frame(lower_fnum);
            mat_high = obj.read_frame(upper_fnum);
            
            % interpolate the two images:
            dist = (fnum - lower_real_frame) / (obj.ds_ratio - 1);
            mat = (mat_low * (1-dist)) + (mat_high * dist);      
        end
        
        
        function mat = read_interpolated_frames(obj, range)
            if isa(range, 'char') && strcmp(range, ':')
                range = 1:obj.config.original_frames;
            end
            
            mat = obj.read_interpolated_frame(range(1));
            if length(range) < 2
                return;
            end
            
            for f = range(2:end)
                frame_mat = obj.read_interpolated_frame(f);
                mat = cat(3, mat, frame_mat);
            end
                
        end
        
        
        % loads a frame by it's internal number:
        function mat = read_frame(obj, fnum)
            
            % calculate where to find the frame:
            [fnum, row, col, ~] = begonia.scantype.proxy.frame_pos(...
                fnum ...
                , obj.config.dim);
            
            % get image from cache, or add it:
            img_mat = obj.image_from_cache(fnum);
            
            if isempty(img_mat)
                img_fname = obj.config.file_list{fnum,:};
                img_path = fullfile(obj.path, img_fname);
                img_mat = imread(img_path);
                obj.cache_image(fnum, img_mat);
            end
            
%             w = size(img_mat, 1) / obj.config.dim;
%             h = size(img_mat, 2) / obj.config.dim;
            
            w = obj.config.width;
            h = obj.config.height;
            
            x = 1 + ((col - 1) * w);
            y = 1 + ((row - 1) * h);
            
            mat = img_mat(x:x+w-1, y:y+h-1);
        end

        
        
        % override for builtin size function:
        function n = size(obj, d)
            size_vec = [obj.config.width ... 
                , obj.config.height ...
                , obj.config.original_frames];
            
            if nargin < 2
               n = size_vec;
            else
                n = size_vec(d);
            end
        end
        
        
        
        % override for allow using the reader with indexes:
        function result = subsref(obj, indicies)
            if(~strcmp(indicies(1).type,'()'))
                % we only want to deal with indexing, and pass other stuff
                % onto the builtin subsref function. This prevents this
                % modification from interfering with e.g. dot-notation.
                result = builtin('subsref', obj, indicies);
            else
                result = uint16([]);
                if length(indicies.subs) < 3
                    reshape = 2;
                    indicies.subs(3) =  {':'};
                    %error('Reader requires three indexes')
                else
                    reshape = 0;
                end
                
                % grab frames and extract subarea:
                xr = indicies.subs{1};
                yr = indicies.subs{2};
                zr = indicies.subs{3};
                
                % extract full matrix:
                fullmat = obj.read_interpolated_frames(zr);
                result = fullmat(xr, yr,:);
                
                if reshape == 2
                   result = reshape(result, [size(result,1), size(result,2) * size(result,3)]); 
                end
                
            end
        end
        
        
        function datatype = class(obj)
            datatype = obj.datatype;
        end
        
    end
    
    methods (Access = private)

        % return the image matrix if it is already im memory:
        function img = image_from_cache(obj, fnum)
            for cached_img = obj.image_cache
                if cached_img.number == fnum
                    img = cached_img.mat;
                    return;
                end
            end 
            
            img = [];
        end
        
        % adds an image to the cache:
        function cache_image(obj, fnum, mat)
            % clear least used image from cache:
            if length(obj.image_cache) > obj.max_cache
                obj.image_cache = obj.image_cache(2:end);
            end
            
            cached_image = struct();
            cached_image.number = fnum;
            cached_image.mat = mat;
            obj.image_cache = [obj.image_cache cached_image]; 
        end 
    end
end










