function proxy = create(stack, output_dir, ds_rate, dims, write_png, write_metadata)
    import begonia.util.logging.*;
    tic;

    if nargin < 5
        write_png = true;
        write_metadata = true;
    end

    if nargin < 3
        ds_rate = 10;
        dims = 5;
    end
    
    % write PNGs if asked:
    if write_png
        vlog(1, ['(create) Writing grid PNG files for stack: ' stack.name])
        create_gpng(stack, output_dir, ds_rate, dims);
    end
    
    
    % write other known files:
    if write_metadata
        vlog(1, ['(create) Copying stack metadata: ' stack.name])
        create_meatdata(stack, output_dir);
    end

    % this is partly a workaround due to problems with proxy stacks
    % generating these images, and partly to ensure the proxy stacks show
    % the correct source-based images for this:
    vlog(1, ['(create) Copying/generating auxillary files: ' stack.name])
    proxy = begonia.scantype.proxy.TSeriesProxy(output_dir);
    
    % copy reference images, if asked:
    for cycle = 1:stack.cycles
        for channel = 1:stack.channels
            % std-image:
            stack.get_std_img(cycle, channel);
            file_std = sprintf('std_cycle_%.3d_channel_%.2d.tif',cycle,channel);
            path_std = fullfile(stack.path,file_std);

            % average_image:
            stack.get_avg_img(cycle, channel);
            file_avg = sprintf('avg_cycle_%.3d_channel_%.2d.tif',cycle,channel);
            path_avg = fullfile(stack.path,file_avg);

            % max image:
            stack.get_max_img(cycle, channel);
            file_max = sprintf('max_cycle_%.3d_channel_%.2d.tif',cycle,channel);
            path_max = fullfile(stack.path,file_max);

            % copy images:
            if exist(path_std, 'file')
                copyfile(path_std, fullfile(proxy.path, file_std));
            end

            if exist(path_avg, 'file')
                copyfile(path_avg, fullfile(proxy.path, file_avg));
            end

            if exist(path_max, 'file')
                copyfile(path_max, fullfile(proxy.path, file_max));
            end
        end
    end

    vlog(1, '(create) Done');
end


function create_gpng(stack, output_dir, ds_rate, dims)
    import begonia.util.logging.*;

    for cycle = 1:stack.cycles

        for channel = 1:stack.channels
            channel_name = begonia.scantype.proxy.ch_dirname(cycle, channel);
            channel_path = fullfile(output_dir , channel_name);
            mkdir(channel_path);
            
            vlog(1, channel_name);
            if isa(stack, 'begonia.scantype.TSeries')
                mat = stack.get_mat(cycle, channel);
            else
                mat = stack.get_stack(cycle, channel);
            end
            
            % each cycle-channel pair gets a separate writer:
            title = channel_name;
            fprefix = [title ' - '];
            writer = begonia.scantype.proxy.GPNGWriter( ...
                channel_path ...
                , dims ...
                , ds_rate ...
                , fprefix ...
                , title);
            writer.open(mat);
            
            % average frames and build multi-PNG reader input:
            for frame = 1:ds_rate:stack.frames_in_cycle
                
                % avoid overshooting end of array:
                if frame + ds_rate < stack.frames_in_cycle
                    mean_frames = frame:(frame + ds_rate);
                else
                    mean_frames = frame:stack.frames_in_cycle;
                end
                
                mean_frame = uint16(mean(mat(:,:,mean_frames), 3));
                writer.write(mean_frame);
               
            end
            
            % ensures last file gets saved:
            writer.close();
        end
    end
end


% copies important stack files:
function create_meatdata(stack, output_dir)
        
    % we have an ID to connect this proxy to in the future:
    src_uuid = stack.dl_unique_id;
    
    % copy XML file
    xml_file = dir(fullfile(stack.path, '*.xml'));
    xml_path = fullfile(xml_file(1).folder, xml_file(1).name);
    copyfile(xml_path, output_dir);
    
    % copy metadata dir: 
    if exist(stack.dloc_metadata_dir, 'dir')
        copyfile(stack.dloc_metadata_dir, fullfile(output_dir, 'metadata'));
    end
    
    % ensure the proxy has a copy of the source ID:
    dloc = begonia.data_management.DataLocation(output_dir);
    dloc.save_var('proxy_src_uuid', src_uuid);
    dloc.save_var('proxy_method', 'gpng');
    dloc.save_var('proxy_api', '1.0');
    
    % save metadata file:
    begonia.scantype.procstack.save_metadata(...
        stack ...
        , output_dir ...
        , 'PNG-Proxy' ...
        ,  {});
end






