function stacks = find_by_class(varargin)
    import begonia.scantype.*;
    import begonia.util.logging.*;

    % find_stacks_by_class recursively searchs for spesific stacks at the specified directory.
    %   Looks for spesific stacks specified by stack_classes. 
    %
    %   stacks = find_stacks_by_class(dir_in,stack_classes)
    %   stacks = find_stacks_by_class(...,NAME,VALUE)
    %
    %   REQUIRED
    %   dir_in               - (char array) 
    %                           Directory.
    %   stack_classes       - (cell of begonia.stack.Stack classes) 
    %                           Each class in a cell array with @ in front.
    %   
    %   PARAMETERS
    %   show_waitbar        - (logical) 
    %                           Show a gui waiting window while searching for
    %                           stacks (default false).
    %   ignore_str          - (char,default:'') 
    %                           Directores containing ignore_str will not be 
    %                       entered when searching for stacks. This is useful 
    %                       in the case where stacks might contain many
    %                       files that slows down the search. E.g. if a stack
    %                       folder is named 'TSeries001', if input_str =
    %                       'TSeries' the stack will still be found be the
    %                       algorithm will not look inside 'TSeries001' for more
    %                       stacks. 
    %
    %   RETURNED
    %   stacks              - (begonia.stack.Stack) 
    %                           Empty if not found. 
    %
    % Example:
    %   stack_classes = {...
    %       @begonia.stack.StackPrarie, ...
    %       @begonia.stack.StackH5, ...
    %       @begonia.stack.StackOldV1};
    %   stacks = begonia.stack_search.find_stacks_by_class(dir_in,stack_classes);
    p = inputParser;
    p.addRequired('dir_in',...
        @(x) validateattributes(x,{'char', 'string'},{'nonempty'}));
    p.addRequired('stack_classes', ...
        @(x) validateattributes(x,{'cell'},{'nonempty'}));
    p.addParameter('show_waitbar',false, ...
        @(x) validateattributes(x,{'logical'},{}));
    p.addParameter('ignore_str','', ...
        @(x) validateattributes(x,{'char'},{}));
    p.addParameter('ignore_metadata',true, ...
        @(x) validateattributes(x,{'logical'},{}));
    p.parse(varargin{:});
    begonia.util.dump_inputParser_vars_to_caller_workspace(p);


    vlog(1,'Discovering directories');
    all_dirs = begonia.path.list_dirs(dir_in,ignore_str);
            
    if ignore_metadata
        % Ignore dirs where the name of the last folder is 'metadata'.
        last_dir = [filesep,'metadata'];
        I = cellfun(@(x)strcmp(x(end-length(last_dir)+1:end),last_dir),all_dirs);
        all_dirs(I) = [];
    end
    
    stacks = begonia.path.init_classes_from_paths(all_dirs,stack_classes, ...
        'show_waitbar',show_waitbar);

end

