function stacks = find_explicit(varargin)
    import begonia.scantype.*;
    import begonia.util.logging.*;


    % find_stacks_explicit
    %   Looks for spesific stacks specified by stack_classes. 
    %
    %   stacks = find_stacks_explicit(dirs_in,stack_classes)
    %   stacks = find_stacks_explicit(...,NAME,VALUE)
    %
    %   REQUIRED
    %   dirs_in             - (cell array of chars) 
    %                           Each directory to search for stacks. Each
    %                           directory will be ran with each stack class and
    %                           the first one that does not fail will be
    %                           returned.
    %   stack_classes       - (cell of begonia.stack.Stack classes) 
    %                           Each class in a cell array with @ in front.
    %   
    %   PARAMETERS
    %   show_waitbar        - (logical) 
    %                           Show a gui waiting window while searching for
    %                           stacks (default false).
    %
    %   RETURNED
    %   stacks              - (begonia.stack.Stack) 
    %                           Empty if not found. 
    p = inputParser;
    p.addRequired('dirs_in',...
        @(x) validateattributes(x,{'cell'},{}));
    p.addRequired('stack_classes', ...
        @(x) validateattributes(x,{'cell'},{'nonempty'}));
    p.addParameter('show_waitbar',false, ...
        @(x) validateattributes(x,{'logical'},{}));
    p.parse(varargin{:});
    begonia.util.dump_inputParser_vars_to_caller_workspace(p);

    global BEGONIA_VERBOSE

    stacks = begonia.stack.Stack.empty;

    if show_waitbar
        h = waitbar(0,'Finding stacks.');
    end

    % Try to find every stack, but continue if it fails because the folder is
    % not a valid stack.
    cnt = 1;
    for i = 1:length(dirs_in)
        if show_waitbar
            waitbar(i/length(dirs_in),h);
        end

        for j = 1:length(stack_classes)
            try
                stacks(cnt) = stack_classes{j}(dirs_in{i});
                cnt = cnt + 1;
                break
            catch e
                if contains(e.identifier,'begonia:load')
                    if BEGONIA_VERBOSE >= 3
                        disp(dirs_in{i});
                        disp(e.identifier);
                    end
                    continue
                else
                    rethrow(e);
%                     warning(['Could not load due to unknown exception: ' dirs_in{i}])
                end
            end
        end
    end

    if show_waitbar
        close(h)
    end

end

