%{
Problem: we need to smooth the stack in time and space, then for each pixel
keep a record of what values it has held. This consumes an unrelaistic
amount of memory for allot of scenarios.  

Solution: to compensate we will for each x,y-coordinate grab an area around
it large enough to apply the gaussean filter, then count the values in that
column, and move on to the next. We'll keep the raw values in memory long
enough after this to use them for smoothing the next pixels. We will only
retain in memory the dominating pixel value for each coordinate after the
calculation pass.  

This solution swaps memory usage for I/O and CPU.  It assumes the matrix
is a lazy reader of data.
%}
function f0 = read_lowmem(stack, cy, ch, divs, use_gpu, baseline_frames)
    import begonia.util.*;
    import begonia.util.logging.*;
    tic;
    
    if nargin < 2
        cy = 1;
        ch = 1;
    end
    
    if nargin < 4
       divs = 2; 
    end
    
    if nargin < 5
       use_gpu = false; 
    end
    
    if nargin < 6
        baseline_frames = 1:stack.frames_in_cycle;
    end
    
    vlog(1, 'Creating filter (not applying')
    sigma_dx = 1;
    sigma_dt = 0.1/stack.dt;

    win_dx = begonia.util.gausswin(sigma_dx);
    win_dt = begonia.util.gausswin(sigma_dt);
    win_dt = reshape(win_dt,1,1,[]);

    filter = win_dx*win_dx';
    filter = repmat(filter,1,1,length(win_dt));
    filter = filter.*win_dt;
    filter = filter/sum(filter(:));
    filter = single(filter);
    toc;
    
    overlap = (length(win_dt) - 1) / 2;
    
    func = @(m,~,~,~) chunk_func(m, filter, sigma_dt, use_gpu);
    
    vlog(1, 'Getting stack reader')
    mat = stack.get_mat(cy,ch);
    mat = mat(:,:,baseline_frames);
    toc
    
    vlog(1, 'Intiating XY-chunked processing of matrix')
    disp([num2str(divs^2) ' chunks']);
    f0 = chunkxyfun(func, mat, divs, overlap); 
end

function f0 = chunk_func(mat, filter, sigma_dt, use_gpu)
    import begonia.util.logging.*;

    toc;
    tic;
    
    
    vlog(1, '(Chunk) Applying filter'); 
    mat = convn(mat,filter,'same');
    
    vlog(1, '(Chunk) Downsampling'); 
    mat_down = mat(:,:,1:ceil(sigma_dt):end);
    mat_down = uint16(mat_down);
    
    if use_gpu        
        gpu = gpuDevice();
        reset(gpu)
        vlog(1, '(Chunk) Transfering to GPU'); 
        mat_down = gpuArray(single(mat_down));
    end
    
    vlog(1, '(Chunk) Calculating mode'); 
    f0 = mode(mat_down,3);
    
    if use_gpu
        f0 = single(gather(f0));
    else
        f0 = single(f0);
    end
    vlog(1, '(Chunk) done'); 
end





