function [roa, roa_ignore_mask] = read(stack, img_f0, mat, ignore_treshold, size_tresh_um2, dur_tresh_s, custom_ignore)
    import begonia.util.logging.*;
    if nargin < 7
        custom_ignore = [];
    end
    if nargin < 6
        dur_tresh_s = 1;
    end
    if nargin < 5
        size_tresh_um2 = 5;
    end
    if nargin < 4
        ignore_treshold = 100;
    end
    if nargin < 3
        mat = stack.load_var('roa_mask');
    end
    if nargin < 2
        img_f0 = stack.load_var('img_f0');
    end
    
    timer_id = tic; tic;
    vlog(1, ['(RoA) Stack ' stack.name]);

    % Find the areas to ignore.
    vlog(2, '(RoA) Finding ignore mask');
    
    vec = img_f0(:);
    vec(vec < ignore_treshold) = [];
    baseline = mode(floor(vec));
    mask = img_f0 < (baseline * 0.7);
    r = round(1/stack.dx);
    se = strel('sphere',r);
    mask = imdilate(mask,se);
    mask = ~mask;
    
    % If custom ignore mask is provided, add them:
    if ~isempty(custom_ignore)
        mask = mask .* double(custom_ignore);
    end
    
    % Ignore the edges
    c = 6;
    mask(1:c,:) = false;
    mask(end-c:end,:) = false;
    mask(:,1:c) = false;
    mask(:,end-c:end) = false;
    roa_ignore_mask = mask;
    
    % apply ignore mask:
    vlog(2, '(RoA) Applying ignore mask to RoA mask');
    
    % mat = mat .* mask; % causes serious memory issues for some reason
    chunk = 1000;
    for z = 1:chunk:size(mat,3)
        zr = z:z+chunk;
        if zr(end) > size(mat,3)
            zr = size(mat,3);
        end
        vlog(3, ['(RoA) ' num2str(zr(1)) ' - '  num2str(zr(end))]);
        mat(:,:,zr) = mat(:,:,zr) .* mask;
    end

    vlog(2, '(RoA) Doing something');
    size_thresh = size_tresh_um2/stack.dx/stack.dx;
    mat = begonia.binary_mat_functions.filter_by_size_2d(mat,size_thresh);
    time_thresh = dur_tresh_s/stack.dt;
    mat = begonia.binary_mat_functions.filter_by_time(mat,time_thresh);

    vlog(2, '(RoA) Detecting connected components');
    CC = bwconncomp(mat,6);

    
    vlog(2, '(RoA) Extracting RoA from components');
    dim = CC.ImageSize(1:2);
    cnt = 1;
    roa = struct;
    for i = 1:CC.NumObjects
        indices = CC.PixelIdxList{i};
        [y,x,t] = ind2sub(CC.ImageSize, indices);

        % For each t, find the area of the roi.
        area_trace = splitapply(@(x) length(x),x,t-t(1)+1);

        % Find the "total" roi area by finding all the unique xy combinations.
        xy = unique([x,y],'rows');
        x = xy(:,1);
        y = xy(:,2);
        % Find the time vector by finding all the unique t.
        t = unique(t);

        [~,t_peak] = max(area_trace);
        t_peak = t_peak + t(1) - 1;

        roa(cnt).x_idx = x;
        roa(cnt).y_idx = y;
        roa(cnt).t_idx = t;
        roa(cnt).t = t*stack.dt;
        roa(cnt).t_start = t(1)*stack.dt;
        roa(cnt).t_end = t(end)*stack.dt;
        roa(cnt).t_peak = t_peak*stack.dt;
        roa(cnt).area = length(x)*stack.dx*stack.dx;
        roa(cnt).area_trace = area_trace;
        roa(cnt).image_size = dim;
        roa(cnt).center_idx = [mean(x),mean(y)];
        roa(cnt).center = [mean(x),mean(y)]*stack.dx;
        cnt = cnt+1;
    end
    vlog(1, ['(RoA) Done ' stack.name ' in ' num2str(toc(timer_id)) 's']);
end

