function aligned_stacks = align_stacks(stacks, dirs_out, conf)
% align_stacks aligns the input stacks using normcorre. 
%
%   aligned_stacks = align_stacks(stacks,dirs_out)
%   aligned_stacks = align_stacks(stacks,dirs_out,conf)
%
%   REQUIRED
%   stacks              - (begonia.scantype.TSeries) 
%                       TSeries to be aligned. 
%
%   dirs_out            - (cell of char) 
%                       Array of output directories for each stack. Must 
%                       be equal length as stacks. 
%
%   Optional
%   conf                - (struct) Default: begonia.stack_operations.AlignmentSettings()
%                       Struct with configuration properties for the
%                       alignment.
%
%   RETURNED
%   aligned_stacks      - (begonia.scantype.h5.TSeriesH5) 
%                       Aligned stacks. Empty if not found.

if ischar(dirs_out)
    dirs_out = {dirs_out};
end

if nargin < 3
    conf = begonia.stack_operations.AlignmentSettings();
end

assert(isa(stacks,'begonia.scantype.TSeries'),'Invalid input type, must be ''begonia.scantype.TSeries''.');

aligned_stacks = begonia.scantype.h5.TSeriesH5.empty;

for i = 1:length(stacks)
    stack = stacks(i);

    begonia.util.logging.vlog(1, sprintf('Starting alignment of %s',stack.path));

    tic

    % Delete previous folders. 
    if exist(dirs_out{i},'dir')
        rmdir(dirs_out{i},'s');
    end

    metadata = begonia.scantype.h5.create_metadata(stack);
    if conf.rigid
        metadata.processing_state = 'normcorre_aligned_rigid';
    else
        metadata.processing_state = 'normcorre_aligned_nonrigid';
    end

    % Align
    for cycle = 1:stack.cycles
        alignment_channel = conf.channel;

        begonia.util.logging.vlog(1, 'Loading matrix');
        mat = stack.get_mat(cycle,alignment_channel);

        % These stacks tend to be lazy and not actually matlab matricies.
        mat = mat(:,:,:);

        % Align
        begonia.util.logging.vlog(1, 'Aligning with NoRMCoRRE');
        warning off
        [mat, shifts, ~, conf.nc_params] = normcorre(mat, conf.nc_params);
        warning on
        mat = uint16(mat);

        % Create the output folder.
        begonia.path.make_dirs(dirs_out{i});

        % Save as hdf5.
        begonia.util.logging.vlog(1, 'Saving matrix');
        path_h5 = fullfile(dirs_out{i},metadata.files{cycle,alignment_channel});
        mat_lazy = begonia.util.h5arr(path_h5,size(mat),'uint16');
        mat_lazy(:,:,:) = mat;

        % Shift and save the other channels.
        channels = 1:stack.channels;
        % Ignore the channel that was just aligned. 
        channels(alignment_channel) = [];
        for channel = channels
            begonia.util.logging.vlog(1, 'Loading matrix');
            mat = stack.get_mat(cycle,channel);
            mat = mat(:,:,:);

            begonia.util.logging.vlog(1, 'Applying shifts');
            mat = apply_shifts(mat,shifts,conf.nc_params);

            % Save as hdf5.
            path_h5 = fullfile(dirs_out{i},metadata.files{cycle, channel});
            begonia.util.logging.vlog(1, 'Saving matrix');
            mat_lazy = begonia.util.h5arr(path_h5,size(mat),'uint16');
            mat_lazy(:,:,:) = mat;
        end
    end

    % Save metadata:
    begonia.util.logging.vlog(2, 'Saving metadata');
    save(fullfile(dirs_out{i},'metadata.mat'), 'metadata');

    % Copy over any xml files_
    begonia.util.logging.vlog(1, 'Copying XML files');
    xml_files = begonia.path.find_files(stack.path,'.xml');
    for j = 1:length(xml_files)
        [~,xml_filename,~] = fileparts(xml_files{j});
        xml_new = fullfile(dirs_out{i},[xml_filename,'.xml']);
        copyfile(xml_files{j} ,xml_new);
    end


    % Get the new stack and create other images.
    begonia.util.logging.vlog(1, 'Creating reference images');
    stack_out = begonia.scantype.h5.TSeriesH5(dirs_out{i});
    for cy = 1:stack.cycles
        for ch = 1:stack.channels
            stack_out.get_avg_img(cy,ch);
            stack_out.get_std_img(cy,ch);
            stack_out.get_max_img(cy,ch);
        end
    end

    aligned_stacks(end+1) = stack_out;

    alignment_time = toc;
    str = sprintf('Alignment completed in %d seconds.',alignment_time);
    begonia.util.logging.vlog(1,str);
end

end

