function aligned_stacks = align_stacks_memmat(stacks, dirs_out, conf)
% align_stacks aligns the input stacks using normcorre. 
%
%   aligned_stacks = align_stacks(stacks,dirs_out)
%   aligned_stacks = align_stacks(stacks,dirs_out,conf)
%
%   REQUIRED
%   stacks              - (begonia.scantype.TSeries) 
%                       TSeries to be aligned. 
%
%   dirs_out            - (cell of char) 
%                       Array of output directories for each stack. Must 
%                       be equal length as stacks. 
%
%   Optional
%   conf                - (struct) Default: begonia.stack_operations.AlignmentSettings()
%                       Struct with configuration properties for the
%                       alignment.
%
%   RETURNED
%   aligned_stacks      - (begonia.scantype.h5.TSeriesH5) 
%                       Aligned stacks. Empty if not found.

if ischar(dirs_out)
    dirs_out = {dirs_out};
end

if nargin < 3
    conf = begonia.stack_operations.AlignmentSettings();
end

assert(isa(stacks,'begonia.scantype.TSeries'),'Invalid input type, must be ''begonia.scantype.TSeries''.');

%aligned_stacks = begonia.scantype.h5.TSeriesH5.empty;
aligned_stacks = [];

for i = 1:length(stacks)
    stack = stacks(i);

    begonia.util.logging.vlog(1, sprintf('Starting alignment of %s',stack.path));

    tic

    % Delete previous folders. 
    if exist(dirs_out{i},'dir')
        rmdir(dirs_out{i},'s');
    end

    metadata = begonia.scantype.h5.create_metadata(stack);
    if conf.rigid
        metadata.processing_state = 'normcorre_aligned_rigid';
    else
        metadata.processing_state = 'normcorre_aligned_nonrigid';
    end

    % Align
    for cycle = 1:stack.cycles
        alignment_channel = conf.channel;

        begonia.util.logging.vlog(1, 'Loading matrix');
        mat = stack.get_mat(cycle,alignment_channel);

        % These stacks tend to be lazy and not actually matlab matricies.
        mat = mat(:,:,:);
        % Align
        begonia.util.logging.vlog(1, 'Aligning with NoRMCoRRE');
        warning off
        begonia.util.logging.vlog(1, 'Saving temporary matrix');
        
        fileid = fopen('temp.dat','w');
        fwrite(fileid,mat,'uint16');
        fclose(fileid);
        sizY = size(mat);
        m = memmapfile('temp.dat','Format',{'uint16',sizY,'Y'}, 'Writable',true);
        %Y = m.Data.Y;
        inconf = conf.nc_params;
        inconf.output_type = 'h5';
        inconf.h5_groupname = 'dataset_1';
        begonia.path.make_dirs(dirs_out{i});
        inconf.h5_filename = [dirs_out{i} '/' metadata.files{alignment_channel}];
        [fn, shifts, ~, conf.nc_params, ~, avg_img, max_img, std_img] = begonia.util.normcorre_memmat(m, inconf);
        warning on
        avg_filename = [dirs_out{i} '/avg_cycle_00' num2str(cycle) '_channel_0' num2str(alignment_channel) '.tif'];
        max_filename = [dirs_out{i} '/max_cycle_00' num2str(cycle) '_channel_0' num2str(alignment_channel) '.tif'];
        std_filename = [dirs_out{i} '/std_cycle_00' num2str(cycle) '_channel_0' num2str(alignment_channel) '.tif'];
              
        saveastiff(avg_img, avg_filename);
        saveastiff(max_img, max_filename);
        saveastiff(std_img, std_filename);
        try
            delete temp.dat
        end
        % Shift and save the other channels.
        channels = 1:stack.channels;
       
        % Ignore the channel that was just aligned. 
        channels(alignment_channel) = [];
        for channel = channels
            begonia.util.logging.vlog(1, 'Loading matrix');
            mat = stack.get_mat(cycle,channel);
            mat = mat(:,:,:);
            fileid2 = fopen('temp2.dat','w');
            fwrite(fileid2, mat,'uint16');
            fclose(fileid2);
            sizY = size(mat);
            m = memmapfile('temp2.dat','Format',{'uint16',sizY,'Y'}, 'Writable',true);

            begonia.util.logging.vlog(1, 'Applying shifts');
            inconf.h5_filename = fullfile(dirs_out{i},metadata.files{cycle, channel});
            [~, avg_img, max_img, std_img] = begonia.util.apply_shifts_memmat(m,shifts,inconf);

            begonia.util.logging.vlog(1, 'Saving matrix');
            
            avg_filename = [dirs_out{i} '/avg_cycle_00' num2str(cycle) '_channel_0' num2str(channel) '.tif'];
            max_filename = [dirs_out{i} '/max_cycle_00' num2str(cycle) '_channel_0' num2str(channel) '.tif'];
            std_filename = [dirs_out{i} '/std_cycle_00' num2str(cycle) '_channel_0' num2str(channel) '.tif'];

            saveastiff(avg_img, avg_filename);
            saveastiff(max_img, max_filename);
            saveastiff(std_img, std_filename);
        end
%     end
    end

    % Save metadata:
    begonia.util.logging.vlog(2, 'Saving metadata');
    save(fullfile(dirs_out{i},'metadata.mat'), 'metadata');

    % Copy over any xml files_
    begonia.util.logging.vlog(1, 'Copying XML files');
    xml_files = begonia.path.find_files(stack.path,'.xml');
    for j = 1:length(xml_files)
        [~,xml_filename,~] = fileparts(xml_files{j});
        xml_new = fullfile(dirs_out{i},[xml_filename,'.xml']);
        copyfile(xml_files{j} ,xml_new);
    end


    % Get the new stack and create other images.
    begonia.util.logging.vlog(1, 'Creating reference images');
    stack_out = begonia.scantype.find_scans(dirs_out{i});
    alignment_time = toc;
    str = sprintf('Alignment completed in %d seconds.',alignment_time);
    begonia.util.logging.vlog(1,str);
    aligned_stacks = [aligned_stacks, stack_out];
end

end

