classdef Chunker < handle 
    properties (SetAccess = private)
        chunks
        chunk_axis
        chunk_size
        chunk_start_end
        chunk_padding
        dim
    end
    
    methods
        function obj = Chunker(mat,varargin)
            
            
            default_data_type = class(mat);
            
            p = inputParser;
            p.KeepUnmatched = true;
            p.addParameter('chunk_axis',ndims(mat),...
                @(x) validateattributes(x,{'numeric'},{'real'}));
            p.addParameter('chunk_size_mb',0,...
                @(x) validateattributes(x,{'numeric'},{'real'}));
            p.addParameter('chunk_size_RAM_fraction',4,...
                @(x) validateattributes(x,{'numeric'},{'real','integer'}));
            p.addParameter('chunk_size',[],...
                @(x) validateattributes(x,{'numeric'},{'real'}));
            p.addParameter('chunk_padding',0,...
                @(x) validateattributes(x,{'numeric'},{'real'}));
            p.addParameter('data_type',default_data_type,...
                @(x) begonia.validators.validatestring(x,{'double','single','uint16','uint8'}));
            p.parse(varargin{:});
            begonia.util.dump_inputParser_vars_to_caller_workspace(p);
            
            if chunk_size_mb == 0
                % Get the memory of the computer to use as a default value for
                % chunk size. 
                if ismac 
                    [~,txt] = system('sysctl -a | grep hw.memsize | awk ''{print $2}'''); 
                    memory_avail_gb = (eval(txt)/1024^3);
                    memory_avail_mb = memory_avail_gb*1e3;
                    chunk_size_mb = memory_avail_mb/chunk_size_RAM_fraction;
                elseif ispc  
                    m = memory;
                    memory_avail_mb = m.MemAvailableAllArrays/1e6;
                    chunk_size_mb = memory_avail_mb/chunk_size_RAM_fraction;
                else
                    chunk_size_mb = 16000000/chunk_size_RAM_fraction;
                end
            end
            
            if isempty(chunk_size)
                % Base the chunk_size on the size of a slice of the dat mat
                % along the chunk_axis and the desired chunk size in mega
                % bytes.
            
                switch data_type
                    % Find the bytes each number in mat.
                    case 'double'
                        sample_size_bytes = 8;
                    case 'single'
                        sample_size_bytes = 4;
                    case 'uint16'
                        sample_size_bytes = 2;
                    case 'uint8'
                        sample_size_bytes = 1;
                    otherwise
                        error('Unknown byte size of the matrix class. Please add it to the list.');
                end
                
                dim = size(mat);
                dim(chunk_axis) = 1;
                slice_size = prod(dim)*sample_size_bytes;
                
                chunk_size = round(chunk_size_mb*1e6/slice_size);
                chunk_size = max(chunk_size,1);
            end
            
            obj.chunk_size = chunk_size;
            
            chunk_dim = size(mat,chunk_axis);
            
            chunk_start = 1:obj.chunk_size:chunk_dim;
            chunk_end = circshift(chunk_start,-1) - 1;
            chunk_end(end) = chunk_dim;
            
            obj.dim = size(mat);
            
            obj.chunks = length(chunk_start);
            
            obj.chunk_start_end = zeros(obj.chunks,2);
            obj.chunk_start_end(:,1) = chunk_start;
            obj.chunk_start_end(:,2) = chunk_end;
            
            obj.chunk_axis = chunk_axis;
            obj.chunk_padding = zeros(obj.chunks,2);
            obj.chunk_padding(:,:) = chunk_padding;
            obj.chunk_padding(1,1) = 0;
            obj.chunk_padding(end,2) = 0;
        end
        
        
        function I = matrix_cell_index(self,chunk_number)
            I = repmat({':'},1,length(self.dim));
            idx_1 = self.chunk_start_end(chunk_number,1) - self.chunk_padding(chunk_number,1);
            idx_2 = self.chunk_start_end(chunk_number,2) + self.chunk_padding(chunk_number,2);
            I{self.chunk_axis} = idx_1:idx_2;
        end
        
        function mat = chunk(self,mat,chunk_number)
            I = self.matrix_cell_index(chunk_number);
            mat = mat(I{:});
        end
        
        
        function mat = unpad(self,mat,chunk_number)
            otherdims = repmat({':'},1,length(self.dim));
            N = size(mat,self.chunk_axis);
            idx_1 = 1 + self.chunk_padding(chunk_number,1);
            idx_2 = N - self.chunk_padding(chunk_number,2);
            otherdims{self.chunk_axis} = idx_1:idx_2;
            mat = mat(otherdims{:});
        end
        
    end
    
end

