classdef StackPrarie < begonia.stack.Stack
    %StackPrarie reads stacks from the ome-tif format from the recording machine. 
    
    properties (Access = private)
        files
    end
    
    methods
        function obj = StackPrarie(path, ensure_is_tseries)
            obj@begonia.stack.Stack(path);
            
            if nargin < 2
                ensure_is_tseries = true;
            end
            
            if path(end) == filesep
                path(end) = [];
            end
            
            if ensure_is_tseries
                assert(exist(fullfile(path,'References'),'dir') == 7, ...
                    'begonia:invalid_stack:missing_Ref_folder', ...
                    'References folder not found, used as an indicator for prarie tseries.');
            end
            
            xml_files = begonia.path.find_files(path,'.xml',false);
            assert(~isempty(xml_files),'begonia:invalid_stack:missing_xml','Missing xml file.');
            
            xml_file = xml_files{1};
            meta = begonia.stack.StackPrarie.parse_xml_file_fast(xml_file);
            
            if ensure_is_tseries
                assert(contains(meta.type,'TSeries'), ...
                    'begonia:invalid_stack:unsupported_type', ...
                    'Only supports TSeries as of now.');
            end
            
            obj.type = meta.type;
            obj.dx = meta.dx;
            obj.dy = meta.dx;
            obj.dt = meta.dt;
            obj.optical_zoom = meta.optical_zoom;
            obj.cycles = meta.cycles;
            obj.frames_in_cycle = meta.frames_in_cycle;
            
            obj.channels = meta.channels;
            obj.record_date = meta.date;
            obj.channel_names = meta.channel_names;
            obj.path = path;
            obj.original_img_dim = meta.original_img_dim; 
            
            obj.files = cellfun(@(filename) [obj.path,filesep,filename], ...
                meta.files, 'UniformOutput',false);
            
            % Set the name of the stack from the directory.
            name = strsplit(path,filesep);
            name = name{end};
            % This is a hotfix, not needed in most cases, but probably wont
            % hurt. 
            name = strrep(name,'-raw','');
            obj.name = name;
            
            % Assert this is a prarie stack by verifying that the last file
            % exists. 
            if ensure_is_tseries
                files = obj.files(:);
                assert(~isempty(files),'begonia:invalid_stack:empty_xml','The xml file has no tif files.');
                assert(exist(files{end},'file') == 2,'begonia:invalid_stack:missing_tiffs','Missing tiff files!');
            end 
            
            global BEGONIA_VERBOSE
            if BEGONIA_VERBOSE >= 1
                fprintf('Found prarie stack: %s\n',obj.path);
            end
            
            %obj.original_img_dim = size(imread(obj.files{1}));
            obj.processing_state = 'raw';
            obj.format = 'ome-xml';
        end
        
        function mat = get_stack(self,cycle,channel)
            files = self.files(cycle,channel,:);  %#ok<*PROPLC>
            mat = begonia.frame_providers.PrarieFrameProvider(files);
        end
        
        
        function [complete,missing_files] = validate_files(self)
            complete = true;
            missing_files = {};
            for s = self
                fprintf('Validating %s\n',s.path);
                for i = 1:numel(s.files)
                    if exist(s.files{i},'file') == 0
                        complete = false;
                        missing_files{end+1} = s.files{i};
                    end
                end
            end
        end
        
        
        function resolve_intermettent_missing_files(self)
            [~,missing_files] = self.validate_files();
            for i = 1:length(missing_files)
                num = regexp(missing_files{i},'\d+','match');
                num = num{end};
                num_prev = str2num(num);
                num_prev = sprintf('%.6d',num_prev-1);

                path = missing_files{i};
                path_prev = strrep(path,num,num_prev);
                disp('');
                disp(path);
                disp(path_prev);
                copyfile(path_prev,path);
            end
        end
    end
    
    methods (Static)
        
        function metadata = parse_xml_file_fast(xml_path)
            if ~contains(xml_path,'.xml')
                % Assume it is a directory.
                xml_path = begonia.path.find_files(xml_path, '.xml');
                xml_path = xml_path{1};
            end
%             metadata.date = '';
%             metadata.dt = '';
%             metadata.dx = [];
%             metadata.dy = [];
%             metadata.cycles = [];
%             metadata.type = [];
%             metadata.frames_in_cycle = [];
%             metadata.files = {};
%             metadata.channels = [];
%             metadata.channel_names = {};
            
            text = fileread(xml_path);
            
            temp = regexp(text,'(?<=date=").*?(?=")','match');
            metadata.date = temp{1};
            
            temp = regexp(text,'(?<="framePeriod" value=").*?(?=")','match');
            metadata.dt = str2double(temp{1});
            
            temp = regexp(text,'(?<=<PVStateValue key="micronsPerPixel">).*?(?=</PVStateValue>)','match');
            temp = regexp(temp,'(?<="XAxis" value=").*?(?=")','match');
            metadata.dx = str2double(temp{1});
            
            temp = regexp(text,'(?<=<PVStateValue key="micronsPerPixel">).*?(?=</PVStateValue>)','match');
            temp = regexp(temp,'(?<="YAxis" value=").*?(?=")','match');
            metadata.dy = str2double(temp{1});
            
            temp = regexp(text,'(?<=type=").*?(?=")','match');
            metadata.type = temp{1};
            
            temp = regexp(text,'(?<="opticalZoom" value=").*?(?=")','match');
            metadata.optical_zoom = str2double(temp{1});
            
            temp = regexp(text,'(?<="pixelsPerLine" value=").*?(?=")','match');
            metadata.original_img_dim = [str2num(temp{1}) str2num(temp{1})];
            
            temp = regexp(text,'(?<=Freehand x=").*?(?=")','match');
            
            metadata.line_x = cellfun(@(x) str2double(x), temp);
            
            temp = regexp(text,'(?<= y=").*?(?=")','match');
            metadata.line_y = cellfun(@(x) str2double(x), temp);
            
            temp = regexp(text,'(?<=key="scanLinePeriod" value=").*?(?=")','match');
            metadata.line_period = cellfun(@(x) str2double(x), temp);
            
            
            % Matches any characters that are not whitespace (\S) which is
            % also between 'filename="' and '"' .
            temp = regexp(text,'(?<=filename=")\S*?\.ome\.tif(?=")','match');
            metadata.files = temp;
            
            
            
            
            tmp_cy = cell(length(temp),1);
            tmp_ch = cell(length(temp),1);
            for i = 1:length(temp)
                tmp_cy{i} = temp{i}(end-23:end-19);
                tmp_ch{i} = temp{i}(end-17:end-15);
            end
            metadata.cycles = length(unique(tmp_cy));
            tmp_ch = unique(tmp_ch);
            metadata.channels = length(tmp_ch);
            metadata.channel_names = tmp_ch;
            metadata.files = reshape(metadata.files,metadata.cycles,metadata.channels,[]);
            % The previous code of sorting the files brakes if the cycles
            % dont have equally many frames, so we juse use that data to
            % create frames_in_cycle.
            metadata.frames_in_cycle = repmat(size(metadata.files,3),1,metadata.cycles);
            
        end
        
        function metadata = parse_xml_file(xml_path)
            if ~contains(xml_path,'.xml')
                % Assume it is a directory.
                xml_path = begonia.path.find_files(xml_path, '.xml');
                xml_path = xml_path{1};
            end

            metadata = struct();

            doc = xmlread(xml_path);

            metadata.date = char(doc.getElementsByTagName('PVScan').item(0).getAttribute('date'));

            seq_list = doc.getElementsByTagName('PVStateValue');
            for i = 0:seq_list.getLength()-1
                item = seq_list.item(i);
                key = item.getAttribute('key');
                value = item.getAttribute('value');

                if strcmp(key, 'framePeriod')
                    metadata.dt = str2num(char(value));
                elseif strcmp(key, 'micronsPerPixel')
                    var = item.getElementsByTagName('IndexedValue');
                    metadata.dx = str2num(char(var.item(0).getAttribute('value')));
                    metadata.dy = str2num(char(var.item(1).getAttribute('value')));
                end
            end

            seq_list = doc.getElementsByTagName('Sequence');
            metadata.cycles = seq_list.getLength;
            metadata.type = char(seq_list.item(0).getAttribute('type'));

            for cycle = 0:seq_list.getLength - 1
                frame_list = seq_list.item(cycle).getElementsByTagName('Frame');
                metadata.frames_in_cycle(cycle+1) = frame_list.getLength;

                for frame = 0:frame_list.getLength - 1
                    channel_list = frame_list.item(frame).getElementsByTagName('File');
                    for channel = 0:channel_list.getLength - 1
                        metadata.files{cycle+1,channel+1,frame+1} = char(channel_list.item(channel).getAttribute('filename')); 
                    end
                end
            end

            file_list = frame_list.item(0).getElementsByTagName('File');
            metadata.channels = file_list.getLength;
            for i = 0:file_list.getLength - 1
                metadata.channel_names{i+1} = lower(char(file_list.item(i).getAttribute('channelName')));
            end
        end
        
    end
    
end


