function stacks = find_stacks_by_class(varargin)
% find_stacks_by_class recursively searchs for spesific stacks at the specified directory.
%   Looks for spesific stacks specified by stack_classes. 
%
%   stacks = find_stacks_by_class(dir_in,stack_classes)
%   stacks = find_stacks_by_class(...,NAME,VALUE)
%
%   REQUIRED
%   dir_in               - (char array) 
%                           Directory.
%   stack_classes       - (cell of begonia.stack.Stack classes) 
%                           Each class in a cell array with @ in front.
%   
%   PARAMETERS
%   show_waitbar        - (logical) 
%                           Show a gui waiting window while searching for
%                           stacks (default false).
%   ignore_str          - (char,default:'') 
%                           Directores containing ignore_str will not be 
%                       entered when searching for stacks. This is useful 
%                       in the case where stacks might contain many
%                       files that slows down the search. E.g. if a stack
%                       folder is named 'TSeries001', if input_str =
%                       'TSeries' the stack will still be found be the
%                       algorithm will not look inside 'TSeries001' for more
%                       stacks. 
%
%   RETURNED
%   stacks              - (begonia.stack.Stack) 
%                           Empty if not found. 
%
% Example:
%   stack_classes = {...
%       @begonia.stack.StackPrarie, ...
%       @begonia.stack.StackH5, ...
%       @begonia.stack.StackOldV1};
%   stacks = begonia.stack_search.find_stacks_by_class(dir_in,stack_classes);
p = inputParser;
p.addRequired('dir_in',...
    @(x) validateattributes(x,{'char'},{'nonempty'}));
p.addRequired('stack_classes', ...
    @(x) validateattributes(x,{'cell'},{'nonempty'}));
p.addParameter('show_waitbar',false, ...
    @(x) validateattributes(x,{'logical'},{}));
p.addParameter('ignore_str','', ...
    @(x) validateattributes(x,{'char'},{}));
p.parse(varargin{:});
begonia.util.dump_inputParser_vars_to_caller_workspace(p);

global BEGONIA_VERBOSE

if BEGONIA_VERBOSE >= 1
    disp('Discovering directories ...')
end

all_dirs = begonia.path.list_dirs(dir_in,ignore_str);

if BEGONIA_VERBOSE >= 1
    disp('Looking for stacks ...')
end

stacks = begonia.stack_search.find_stacks_explicit(all_dirs,stack_classes, ...
    'show_waitbar',show_waitbar);

end

