function stacks_h5 = write_h5(stacks,paths_new,verbose)
% write_h5 creates h5 stacks from stack objects at a new location.
%
%   write_h5(stacks,new_paths)
%   write_h5(stacks,new_paths,verbose)
%
%   REQUIRED
%   stacks          - (begonia.stack.Stack) 
%                       Array of stack objects to write to a new format.
%   paths_new       - (cell of char)
%                       Cell array of paths to save the stacks to. Must
%                       have equal length as number of stacks.
%
%   PARAMETERS
%   verbose         - (int)
%                       0   : Dont display any text.
%                       1   : Displays text while running.
%
%   RETURNED
%   stacks_h5       - (begonia.stack.StackH5)
%                       The newly created stacks.
if nargin < 3
    verbose = 1;
end
if ischar(paths_new)
    paths_new = {paths_new};
end

assert(length(paths_new)==length(stacks));

stacks_h5 = begonia.stack.Stack.empty(0);

for i = 1:length(stacks)
    stack = stacks(i);
    if verbose
        fprintf('Saving stack to : %s\n',paths_new{i});
    end
    begonia.path.make_dirs([paths_new{i},filesep]);
    
    % Output filenames.
    files_h5 = cell(stack.cycles,stack.channels);
    for cycle = 1:stack.cycles
        for channel = 1:stack.channels

            files_h5{cycle,channel} = ...
                sprintf('%s_cycle_%.3d_ch%.2d.h5', ...
                stack.name,cycle,channel);
        end
    end
    
    for cycle = 1:stack.cycles
        for channel = 1:stack.channels
            
            % Load stack.
            if verbose
                fprintf('Loading matrix  : cycle %d channel %d \n',cycle,channel);
            end
            mat = stack.get_stack(cycle,channel);
            mat = mat(:,:,:);
            
            % Save as hdf5.
            if verbose
                fprintf('Saving matrix   : cycle %d channel %d \n',cycle,channel);
            end
            path = fullfile(paths_new{i},files_h5{cycle,channel});
            mat_new = begonia.util.h5arr(path,size(mat),'uint16');
            mat_new(:,:,:) = mat;
        end
    end
    
    % Save metadata
    if verbose
        disp('Saving metadata.mat');
    end
    metadata.name = stack.name;
    metadata.type = stack.type;
    metadata.cycles = stack.cycles;
    metadata.channels = stack.channels;
    metadata.channel_names = stack.channel_names;
    metadata.frames_in_cycle = stack.frames_in_cycle;
    metadata.original_img_dim = stack.original_img_dim;
    metadata.dx = stack.dx;
    metadata.dy = stack.dy;
    metadata.dt = stack.dt;
    metadata.record_date = stack.record_date;
    metadata.files = files_h5;
    metadata.processing_state = stack.processing_state;
    metadata.format = 'StackH5';
    
    save(fullfile(paths_new{i},'metadata.mat'), 'metadata');
    
    
    % Copy over the xml file.
    xml_files = begonia.path.find_files(stack.path,'.xml');
    if length(xml_files) == 1
        [~,xml_filename,~] = fileparts(xml_files{1});
        if verbose
            fprintf('Copying %s\n',[xml_filename,'.xml']);
        end
        xml_new = fullfile(paths_new{i},[xml_filename,'.xml']);
        copyfile(xml_files{1},xml_new);
    end
    
    stack_new = begonia.stack.StackH5(paths_new{i});
    stack_new.create_additional_images();
    stacks_h5(end+1) = stack_new;
end

fprintf('\n')
end
