function mat = custom(x, c, boundary)
% custom convolves the columns of the input with a custom filter. 
%
%   mat_out = custom(mat,filter_vec)
%   mat_out = custom(mat,filter_vec,boundary)
%
%   REQUIRED
%   mat             - (2d numeric)
%                     Columns of mat are convolved.
%   filter_vec      - (1d numeric) 
%                     mat is convolved with filter_vec.
%   
%   PARAMETERS
%   boundary        - (char)
%                       'replicate'     : (default) The returned convolution 
%                           has the same size as the input mat. The 
%                           convolution is padded with the same values as   
%                           the value the boundaries.
%                       'zero'          : The returned convolution 
%                           has the same size as the input mat. The 
%                           convolution is padded with 0.
%                       'symmetric'     : The returned convolution 
%                           has the same size as the input mat. The 
%                           convolution is padded with symmetrical values
%                           (the boundaries are mirrored).
%   
%
%   RETURNED
%   mat_out        - (2d numeric) 
if nargin < 3
    boundary = 'replicate';
end

if ~isfloat(x)
    x = single(x);
end

% If mat is a row vector, make it a column vector.
if isrow(x)
    x = x';
    row_vec = true;
else
    row_vec = false;
end

if isrow(c)
    c = c';
end

N_c = length(c);
N_x = size(x,1);
N_x_cols = size(x,2);

assert(mod(N_c,2) == 1,'Even length filters gave the developer a headache, please use odd length filters.');
assert(N_c <= N_x,'Filter is longer than the input matrix, this will probably crash the program.');

if N_x_cols == 0
    mat = x;
    return
end

switch boundary
    case 'symmetric'
        pad_x = ceil(N_c/2);
        x = padarray(x,pad_x,'symmetric');
        
        pad_c = size(x,1) - N_c;
        c = padarray(c,pad_c,0,'post');
        
        % Center the convolution
        shift = - (N_c-1)/2;
        x = circshift(x,shift);
        
        mat = ifft(fft(x).*fft(c));
        
        % Remove padding introduced in ifft fft
        mat = mat(1+pad_x:end-pad_x,:);
        
    case 'zero'
        % Pad x on each side, pad c at the end.
        pad_x = ceil(N_c/2);
        x = padarray(x,pad_x);
        
        pad_c = size(x,1) - N_c;
        c = padarray(c,pad_c,0,'post');
        
        % Center the convolution
        shift = - (N_c-1)/2;
        x = circshift(x,shift);
        
        mat = ifft(fft(x).*fft(c));
        
        % Remove padding introduced in ifft fft
        mat = mat(1+pad_x:end-pad_x,:);
    case 'replicate'
        pad_x = ceil(N_c/2);
        x = padarray(x,pad_x,'replicate');
        
        pad_c = size(x,1) - N_c;
        c = padarray(c,pad_c,0,'post');
        
        % Center the convolution
        shift = - (N_c-1)/2;
        x = circshift(x,shift);
        
        mat = ifft(fft(x).*fft(c));
        
        % Remove padding introduced in ifft fft
        mat = mat(1+pad_x:end-pad_x,:);
%         pad_length = ceil(N_c/2);
% 
%         pad_top = ones(pad_length,N_x_cols).*x(1,:);
%         pad_bot = ones(pad_length,N_x_cols).*x(end,:);
%         x = cat(1,pad_top,x,pad_bot);
% 
%         % The filter will be run along the first dimension of the matrix.
%         A = convmtx(c,size(x,1));
% %         A = begonia.util.sparseconvmtx(filter_vec,size(mat,1));
%         mat_out = A*x;
%         % Remove the padding we introduced.
%         % convmtx uses zeropadding so we must remove that too, basically removing
%         % the padding 2 times. 
%         if mod(N_c,2) == 0
%             mat_out = mat_out(2*pad_length:end-2*pad_length,:);
%         else
%             mat_out = mat_out(2*pad_length:end-2*pad_length+1,:);
%         end
    otherwise 
        error('Invalid boundary conditions.');
end

if row_vec
    mat = mat';
end

end