classdef MaskEditor < handle
    
    properties (SetObservable,AbortSet)
        mask
        video_axes
        alpha = 0.5
        mode = 'add';
        brush_radius = 10;
    end
    
    properties (SetAccess = private)
        ready
    end
    
    properties (Access = private)
        img
        image_handle
        mouse_down = false
    end
    
    methods
        function val = get.ready(self)
            val = false;
            if isempty(self.video_axes) || isempty(self.alpha) || isempty(self.mask)
                return;
            end
            val = true;
        end
        
        function set.mask(self,mask)
            self.mask = mask;
            if ~isequal(size(self.img),size(self.mask))
                color = [0,0,1];
                self.img = ones(size(mask,1),size(mask,2),3);
                self.img(:,:,1) = color(1);
                self.img(:,:,2) = color(2);
                self.img(:,:,3) = color(3);
            end
            self.update();
        end
        
        function set.video_axes(self,video_axes)
            if ~isempty(self.video_axes)
                f.WindowButtonMotionFcn = [];
                f.WindowButtonUpFcn = [];
                f.WindowButtonDownFcn = [];
            end
            
            self.video_axes = video_axes;
            
            % Hijack the axes click callback
            f = ancestor(text,'figure','toplevel');
            f.WindowButtonMotionFcn = @(~,~) self.callback_mouse_move();
            f.WindowButtonUpFcn = @(~,~) self.callback_mouse_up();
            f.WindowButtonDownFcn = @(~,~) self.callback_mouse_down();
            
            self.update();
        end
        
        function set.alpha(self,alpha)
            self.alpha = alpha;
            self.update();
        end
    end
    
    methods (Access = private)
        function update(self)
            if self.ready
                if isempty(self.image_handle)
                    hold(self.video_axes,'on');
                    self.image_handle = imshow(self.img,'Parent',self.video_axes);
                    uistack(self.image_handle,'top');
                end
                self.image_handle.Parent = self.video_axes;
                if ~isequal(size(self.img),size(self.image_handle.CData))
                    self.image_handle.CData = self.img;
                end
                self.image_handle.AlphaData = double(self.mask)*self.alpha;
            end
        end
        
        function [point,inside] = get_mouse_coordinates_in_axes(self)
            % Get mousepoint coordinates in ax and figure units
            point = self.video_axes.CurrentPoint(1,1:2);
            
            if point(1) < self.video_axes.XLim(1) || ...
                    point(1) > self.video_axes.XLim(2) || ...
                    point(2) < self.video_axes.YLim(1) || ...
                    point(2) > self.video_axes.YLim(2)
                inside = false;
            else
                inside = true;
            end
        end
        
        function callback_mouse_down(self)
            self.mouse_down = true;
        end
        
        function callback_mouse_up(self)
            self.mouse_down = false;
        end
        
        function callback_mouse_move(self)
            
            point = self.get_mouse_coordinates_in_axes();
            
            if self.ready && self.mouse_down
                r = round(self.brush_radius);
                side = round(r);
                [i,j] = meshgrid(-side:side,-side:side);
                mat = i.*i + j.*j <= r*r;
                [x,y] = find(mat);
                x = x + point(2) - r;
                y = y + point(1) - r;
                
                x = round(x);
                y = round(y);
                
                mask = self.mask;
                dim = size(mask);
                I = x < 1 | x > dim(1) | y < 1 | y > dim(2);
                x(I) = [];
                y(I) = [];
                I = sub2ind(dim,x,y);
                switch self.mode
                    case 'add'
                        mask(I) = true;
                    case 'subtract'
                        mask(I) = false;
                end
                self.mask = mask;
            end
        end
    end
end

