classdef RoiArrayModel < handle
    
    properties (SetObservable,AbortSet)
        selected_rois
        tseries_model
        roi_array
    end
    
    properties (SetAccess = private)
        ready
        
        roi_names
        
        ROI_GROUPS_LONG = {'Neuronal Soma', 'Neuronal Dendrite', 'Neuronal Axon', 'Neuropill','Astrocyte Soma','Astrocyte Endfoot','Astrocyte Process','Gliopill', 'Artery', 'Vein', 'Capillary'};
        ROI_GROUPS = {'NS', 'ND', 'NA','Np','AS', 'AE', 'AP','Gp','Ar','Ve','Ca'}
        ROI_COLORS = {'Red', [.96 .45 .027], [.96 .65 .027], [.75 .5 0], 'Green', [0.016 .61 .51], [.63 .90 .02], [.067 .48 0], [.24 .09 .66], [.43 .051 .64], [.76, .02 .47]}
        
    end
    
    properties (Access = private)
        listener_tseries_model
    end
    
    methods
        function val = get.ready(self)
            val = false;
            
            if isempty(self.tseries_model)
                return;
            end
            
            if ~self.tseries_model.ready
                return;
            end
            
            val = true;
        end
        
        function set.tseries_model(self,tseries_model)
            if ~isempty(self.listener_tseries_model)
                delete(self.listener_tseries_model);
            end
            self.tseries_model = tseries_model;
            self.listener_tseries_model = tseries_model.addlistener('tseries','PostSet',@(~,~) self.load_roi_array());
            self.listener_tseries_model(2) = tseries_model.addlistener('tseries_index','PostSet',@(~,~) self.load_roi_array());
            self.load_roi_array();
        end
        
        function set.roi_array(self,roi_array)
            self.roi_array = roi_array;
            
            % Create incrementing numbers for each group.
            groups = {roi_array.group};
            nums = zeros(1,length(groups));
            [C,IA,IC] = unique(groups);
            for i = 1:length(C)
                ind = IC == i;
                nums(ind) = 1:sum(ind);
            end

            % Create tags based on group and number.
            roi_names = {};
            for i = 1:length(roi_array)
                tag = sprintf([roi_array(i).group,'%02d'],nums(i));
                roi_names{i} = tag;
            end

            % Add the tag of the connected roi
            for i = 1:length(roi_array)
                roi = roi_array(i);
                if isempty(roi.id_connected_roi); continue; end
                % Find the connected roi.
                id = roi.id_connected_roi;
                ids = {roi_array.id};
                connected_tag = roi_names{strcmp(id,ids)};
                roi_names{i} = [roi_names{i},'_',connected_tag];
            end
            self.roi_names = roi_names;
        end
        
        function clear_offsets(self,I)
            if isempty(I)
                return;
            end
            self.roi_array(I).clear_xy_offsets();
        end
        
        function save_rois(self)
            tseries = self.tseries_model.tseries;
            roi_array = self.roi_array;
            roi_array = roi_array.to_struct();
            begonia.util.logging.vlog(1,'Saving ROIs');
            tseries.save_var(roi_array);
        end
        
        function I = get_rois_at_point(self,xy,frame)
            if nargin < 3
                frame = [];
            end
            
            xy = round(xy);
            I = false(1,length(self.roi_array));
            for i = 1:length(self.roi_array)
                xy_roi = [self.roi_array(i).x, self.roi_array(i).y];
                
                if ~isempty(frame) && ~isempty(self.roi_array(i).xy_offsets)
                    xy_roi = xy_roi + self.roi_array(i).xy_offset_vector(frame,:);
                end
                
                if ismember(xy,xy_roi,'rows')
                    I(i) = true;
                end
            end
            I = find(I);
        end
    end
    
    methods (Access = private)
        
        function load_roi_array(self)
            if ~self.ready
                self.roi_array = [];
                self.selected_rois = [];
                return;
            end
            
            ts = self.tseries_model.tseries(self.tseries_model.tseries_index);
            
            roi_array = ts.load_var('roi_array',[]);
            roi_array = xylobium.moving_rois.ROI(roi_array);
            for i = 1:length(roi_array)
                roi_array(i).frames = self.tseries_model.frames(roi_array(i).cycle,roi_array(i).channel);
            end
            self.roi_array = roi_array;
        end
    end
end

