classdef TSeriesVideo < handle
    
    properties (SetObservable,AbortSet)
        parent
        
        tseries_model
        frame
        clim
        cycle
        channel
    end
    
    properties (SetObservable,AbortSet,SetAccess = private)
        video_axes
        video_imhandle
        frames
        ready
    end
    
    properties (Access = private)
        tseries_model_listener
    end
    
    methods
        function val = get.ready(self)
            val = false;
            if isempty(self.video_axes) || ...
                    isempty(self.video_imhandle) || ...
                    isempty(self.clim) || ...
                    isempty(self.frame) || ...
                    isempty(self.cycle) || ...
                    isempty(self.channel) || ...
                    isempty(self.tseries_model)
                return;
            end
            
            if ~self.tseries_model.ready
                return;
            end
            
            frames = self.tseries_model.frames(self.cycle,self.channel);
            
            if self.frame < 1 || self.frame > frames
                return;
            end
            
            val = true;
        end
        
        function set.parent(self,parent)
            delete(self.video_axes);
            delete(self.video_imhandle);
            
            if isempty(parent)
                self.parent = [];
                self.video_axes = [];
                self.video_imhandle = [];
            else
                video_axes = axes('Parent',parent);

                imhandle = imagesc(video_axes,zeros(512,512));

                colormap(gray);

                video_axes.XTick = [];
                video_axes.YTick = [];
                video_axes.DataAspectRatio = [1,1,1];
                video_axes.ActivePositionProperty = 'position';
                
                self.parent = parent;
                self.video_imhandle = imhandle;
                self.video_axes = video_axes;
            end
            
            if self.ready
                self.update();
            else
                self.reset_derived_values();
            end
        end
        
        function set.tseries_model(self,tseries_model)
            if ~isempty(self.tseries_model)
                delete(self.tseries_model_listener)
            end
            
            self.tseries_model = tseries_model;
            
            if ~isempty(self.tseries_model)
                self.tseries_model_listener = tseries_model.addlistener('tseries','PostSet',@self.handle_events);
                self.tseries_model_listener(2) = tseries_model.addlistener('tseries_index','PostSet',@self.handle_events);
                self.set_default_values();
            end
            
            if self.ready
                self.update();
            else
                self.reset_derived_values();
            end
        end
        
        function set.frame(self,frame)
            if self.ready
                self.frame = frame;
                if self.ready
                    self.update_frame();
                else
                    self.reset_derived_values();
                end
            else
                self.frame = frame;
                if self.ready
                    self.update();
                else
                    self.reset_derived_values();
                end
            end
        end
        
        function set.clim(self,clim)
            if self.ready
                self.clim = clim;
                if self.ready
                    self.update_clim();
                else
                    self.reset_derived_values();
                end
            else
                self.clim = clim;
                if self.ready
                    self.update();
                else
                    self.reset_derived_values();
                end
            end
        end
        
        function set.cycle(self,cycle)
            self.cycle = cycle;
            if self.ready
                self.update();
            else
                self.reset_derived_values();
            end
        end
        
        function set.channel(self,channel)
            self.channel = channel;
            if self.ready
                self.update();
            else
                self.reset_derived_values();
            end
        end
    end
    
    methods (Access = private)
        
        function update(self)
            cycle = self.cycle;
            channel = self.channel;
            tseries_model = self.tseries_model;

            self.video_axes.CLim = self.clim;
            self.video_axes.XLim = [1,tseries_model.video_dims{cycle,channel}(2)];
            self.video_axes.YLim = [1,tseries_model.video_dims{cycle,channel}(1)];

            self.frames = tseries_model.frames(cycle,channel);
            self.update_frame();
            self.update_clim();
        end
        
        function update_frame(self)
            self.video_imhandle.CData = self.tseries_model.get_img(floor(self.frame),self.cycle,self.channel);
        end
        
        function update_clim(self)
            self.video_axes.CLim = self.clim;
        end
        
        function reset_derived_values(self)
            self.frames = [];

            if ~isempty(self.video_imhandle)
                self.video_imhandle.CData = zeros(512,512);
            end
        end
        
        function handle_events(self,s,e)
            self.set_default_values();
            if self.ready
                self.update();
            end
        end
        
        function set_default_values(self)
            if isempty(self.cycle) || isempty(self.tseries_model.cycles) || self.cycle < 1 || self.cycle > self.tseries_model.cycles
                self.cycle = 1;
            end
            if isempty(self.channel) || isempty(self.tseries_model.channels) || self.channel < 1 || self.channel > self.tseries_model.channels
                self.channel = 1;
            end
            if isempty(self.frame) || isempty(self.tseries_model.frames) || ...
                    self.frame < 1 || self.frame > self.tseries_model.frames(self.cycle,self.channel)
                self.frame = 1; 
            end
            if isempty(self.clim) && ~isempty(self.tseries_model.clim_minmax)
                self.clim = self.tseries_model.clim_minmax;
            end
        end
    end
end

