classdef RoiArrayModel < handle
    %ROIARRAYMODEL Summary of this class goes here
    %   Detailed explanation goes here
    
    properties (SetObservable,AbortSet)
        selected_rois
        tseries_model
        roi_array
    end
    
    properties (SetAccess = private)
        roi_count
        roi_names
        
        ROI_GROUPS_LONG = {'Neuronal Soma', 'Neuronal Dendrite', 'Neuronal Axon', 'Neuropill','Astrocyte Soma','Astrocyte Endfoot','Astrocyte Process','Gliopill', 'Artery', 'Vein', 'Capillary'};
        ROI_GROUPS = {'NS', 'ND', 'NA','Np','AS', 'AE', 'AP','Gp','Ar','Ve','Ca'}
        ROI_COLORS = {'Red', [.96 .45 .027], [.96 .65 .027], [.75 .5 0], 'Green', [0.016 .61 .51], [.63 .90 .02], [.067 .48 0], [.24 .09 .66], [.43 .051 .64], [.76, .02 .47]}
        
    end
    
    events
        RoiArrayChanged
        SelectedRoisChanged
    end
    
    methods
        function self = RoiArrayModel(tseries_model)
            tseries_model.addlistener('TSeriesChanged',@(s,~) self.load_tseries(s.tseries));
            self.tseries_model = tseries_model;
        end
        
        function clear_offsets(self,I)
            if isempty(I)
                return;
            end
            self.roi_array(I).clear_xy_offsets();
            self.notify('RoiArrayChanged');
        end
        
        function save_rois(self)
            tseries = self.tseries_model.tseries;
            roi_array = self.roi_array;
            roi_array = roi_array.to_struct();
            begonia.util.logging.vlog(1,'Saving ROIs');
            tseries.save_var(roi_array);
        end
        
        function load_tseries(self,tseries)
            roi_array = tseries.load_var('roi_array',[]);
            try
                roi_array = xylobium.moving_rois.old.ROI(roi_array);
            catch e
                begonia.util.logging.vlog(1,'Error during loading ROIs.');
                roi_array = [];
            end
            roi_array.update_tseries_info(tseries);
            
            self.roi_array = roi_array;
            self.update_roi_names();
            
            self.notify('RoiArrayChanged');
            self.selected_rois = [];
        end
        
        function I = get_rois_at_point(self,xy)
            xy = round(xy);
            I = [];
            for i = 1:self.roi_count
                xy_roi = [self.roi_array(i).x, self.roi_array(i).y];
                
                if ~isempty(self.roi_array(i).xy_offsets)
                    xy_roi = xy_roi + self.roi_array(i).xy_offset_vector(self.tseries_model.frame,:);
                end
                
                if ismember(xy,xy_roi,'rows')
                    I = [I,i];
                end
            end
        end
        
        function update_roi_names(self)
            if isempty(self.roi_array)
                self.roi_names = {};
            end
            % Create incrementing numbers for each group.
            groups = {self.roi_array.group};
            nums = zeros(1,length(groups));
            [C,IA,IC] = unique(groups);
            for i = 1:length(C)
                ind = IC == i;
                nums(ind) = 1:sum(ind);
            end

            % Create tags based on group and number.
            roi_names = {};
            for i = 1:length(self.roi_array)
                tag = sprintf([self.roi_array(i).group,'%02d'],nums(i));
                roi_names{i} = tag;
            end

            % Add the tag of the connected roi
            for i = 1:length(self.roi_array)
                roi = self.roi_array(i);
                if isempty(roi.id_connected_roi); continue; end
                % Find the connected roi.
                id = roi.id_connected_roi;
                ids = {self.roi_array.id};
                connected_tag = roi_names{strcmp(id,ids)};
                roi_names{i} = [roi_names{i},'_',connected_tag];
            end
            self.roi_names = roi_names;
        end
        
        function set.selected_rois(self,val)
            if isempty(self.roi_array)
                self.selected_rois = [];
            else
                self.selected_rois = val;
            end
            self.notify('SelectedRoisChanged');
        end
        
        function val = get.roi_count(self)
            val = length(self.roi_array);
        end
    end
    
    
end

