classdef RoiAxesOverlay < handle
    
    properties
        mode
    end
    
    properties (SetAccess = private)
        roi_array_model
        tseries_model
        
        overlay_axes
        
        roi_plot_handles
        
        mouse_down = false;
        mouse_down_pos
        
        modes = {'Draw','Edit','Dynamic Positon'}
    end
    
    events
        ModeChanged
    end
    
    methods
        function self = RoiAxesOverlay(overlay_axes,roi_array_model,tseries_model)
            self.roi_array_model = roi_array_model;
            self.tseries_model = tseries_model;
            
            %% configure overlay
            self.overlay_axes = overlay_axes;
            
            %% Other
            
            % Hijack the figure mouse callbacks
            f = ancestor(text,'figure','toplevel');
            f.WindowButtonMotionFcn = @(~,~) self.callback_mouse_move();
            f.WindowButtonUpFcn = @(~,~) self.callback_mouse_up();
            f.WindowButtonDownFcn = @(~,~) self.callback_mouse_down();
            
            tseries_model.addlistener('FrameChanged',@(~,~) self.on_frame_change());
            
            roi_array_model.addlistener('RoiArrayChanged',@(~,~) self.on_roi_array_change);
            roi_array_model.addlistener('SelectedRoisChanged',@(~,~) self.update_roi_colors());
            
            self.mode = self.modes{3};
        end
        
        function set.mode(self,val)
            if ~ismember(val,self.modes)
                warning('''mode'' does not match any exisiting modes.')
                return;
            end
            self.mode = val;
            self.on_mode_change();
            self.notify('ModeChanged');
        end
        
    end
    
    methods (Access = private)
        
        function on_roi_array_change(self)
            self.create_roi_plot_handles();
            self.update_roi_colors();
        end
        
        function on_frame_change(self)
            
            switch self.mode
                case 'Dynamic Positon'
                    self.update_roi_positions();
                    self.update_roi_colors();
            end
        end
        
        function on_mode_change(self)
            self.create_roi_plot_handles();
            self.update_roi_colors();
        end
        
        function point = get_mouse_coordinates_in_axes(self)
            % Get mousepoint coordinates in ax and figure units
            point = self.overlay_axes.CurrentPoint(1,1:2);
            point = round(point);
            
            % Return if outside axis.
            if point(1) < self.overlay_axes.XLim(1) || ...
                    point(1) > self.overlay_axes.XLim(2) || ...
                    point(2) < self.overlay_axes.YLim(1) || ...
                    point(2) > self.overlay_axes.YLim(2)
                point = [];
                return
            end
        end
        
        function callback_mouse_down(self)
            self.mouse_down = true;
            
            point = self.get_mouse_coordinates_in_axes();
            if isempty(point) 
                return;
            end
            
            self.mouse_down_pos = point;
            
            switch self.mode
                case 'Dynamic Positon'
                    self.roi_array_model.selected_rois = ...
                        self.roi_array_model.get_rois_at_point(point);
            end
        end
        
        function callback_mouse_up(self)
            self.mouse_down = false;
            
            point = self.get_mouse_coordinates_in_axes();
            % TODO: some bugs might be fixed by doing this differently.
            if isempty(point) 
                return;
            end
            
            m = self.roi_array_model;
            t = self.tseries_model;
            selected_rois = m.selected_rois;
            
            switch self.mode
                case 'Dynamic Positon'
                    if ~isempty(selected_rois)
                        offset = point - self.mouse_down_pos;
                        
                        rois = m.roi_array(selected_rois);
                        rois.insert_xy_offset(t.frame,offset(1),offset(2));
                    end
            end
            
        end
        
        function callback_mouse_move(self)
            
            point = self.get_mouse_coordinates_in_axes();
            if isempty(point) 
                self.callback_mouse_up();
                return;
            end
            
            m = self.roi_array_model;
            selected_rois = m.selected_rois;
            
            switch self.mode
                case 'Dynamic Positon'
                    if self.mouse_down && ~isempty(selected_rois)
                        self.update_roi_positions();
                        self.update_roi_colors();
                    end
            end
            
        end
        
        function update_roi_positions(self)
            m = self.roi_array_model;
            t = self.tseries_model;
            
            for i = 1:m.roi_count
                xdata = m.roi_array(i).plot_boundary(:,1);
                ydata = m.roi_array(i).plot_boundary(:,2);
                
                if ~isempty(m.roi_array(i).xy_offset_vector)
                    xdata = xdata + m.roi_array(i).xy_offset_vector(t.frame,1);
                    ydata = ydata + m.roi_array(i).xy_offset_vector(t.frame,2);
                end
                
                switch self.mode
                    case 'Dynamic Positon'
                        if self.mouse_down && ismember(i,m.selected_rois)
                            point = self.get_mouse_coordinates_in_axes();
                            xdata = xdata + point(1) - self.mouse_down_pos(1);
                            ydata = ydata + point(2) - self.mouse_down_pos(2);
                        end
                    otherwise
                end
                
                self.roi_plot_handles(i).XData = xdata;
                self.roi_plot_handles(i).YData = ydata;
            end
        end
        
        function update_roi_colors(self)
            r = self.roi_array_model;
            t = self.tseries_model;
            if isempty(self.roi_plot_handles)
                return;
            end
            if length(self.roi_plot_handles) ~= r.roi_count
                warning('wo wowowo, bad');
                return;
            end
            
            for i = 1:r.roi_count
                roi = r.roi_array(i);
                if ismember(i,r.selected_rois)
                    switch self.mode
                        case 'Dynamic Positon'
                            if ~isempty(roi.xy_offsets) && ismember(t.frame,roi.xy_offsets(:,1))
                                color = 'red';
                            else
                                color = 'white';
                            end
                        otherwise
                            color = 'white';
                    end
                else
                    color_idx = strcmp(roi.group,r.ROI_GROUPS);
                    color = r.ROI_COLORS{color_idx};
                end
                self.roi_plot_handles(i).Color = color;
            end
        end
        
        function create_roi_plot_handles(self)
            
            r = self.roi_array_model;
            
            %% Delete the old handles
            for i = 1:length(self.roi_plot_handles)
                if isvalid(self.roi_plot_handles(i))
                    delete(self.roi_plot_handles(i));
                end
            end
            %% Create new plot handles
            self.roi_plot_handles = gobjects(0);
            hold(self.overlay_axes,'on');
            for i = 1:r.roi_count
                self.roi_plot_handles(i) = plot( ...
                    r.roi_array(i).plot_boundary(:,1), ...
                    r.roi_array(i).plot_boundary(:,2), ...
                    'Parent',self.overlay_axes);
                self.roi_plot_handles(i).LineWidth = 2;
            end
            %%
            self.update_roi_positions();
        end
    end
end

