classdef ROI < handle
    
    properties
        group           % (char)    E.g. 'ND' for neuron dendrite, 'AS' for astrocyte soma. 
        tseries_id      % (char)    Unique ID of tseries.
        frames          % (char)
        channel         % (int)     Channel it is connected to.
        cycle           % (int)     Cycle it is connected to.
        shape           % (char)    Eg. 'polygon', 'circle', 'autodetect'
        id_connected_roi    % (char) The RoI.id of another roi.
        xy_offsets      % (double Nx3) 
        xy_offset_vector % (double Nx2)
        
        % Is set in the constructor and should never be empty.
        id              % (char) Unique identifier.
        id_shared       % (char) ID that can be shared between multiple rois
                        %        to identify them as the same. 
        area            % (int) 
        center          % (double 1x2)
        x               % (int 1xN)
        y               % (int 1xN)
        mask            % (logical NxM)
        boundary        % (cell)
        plot_boundary   % (Nx2) x and y data to plot, seperated by NaNs.
        
        version = '2018-11'
    end
    
    methods (Static)
        function updated_rois = old_to_new(roi_structs)
            % Converts from old versions to the newest version.
            if isempty(roi_structs)
                updated_rois = struct.empty;
            end
            
            for i = 1:length(roi_structs)
                roi = roi_structs(i);
                
                if ~isfield(roi,'id_shared')
                    roi.id_shared = roi.id;
                end
                
                if iscell(roi.boundary)
                    % Create the x and y points to outline a roi. 
                    % The older ROIs have a boundary property which has a bit
                    % convoluted format. As an improvement they should 
                    % just have saved boundary in this format instead (a NaN
                    % seperated list of points).
                    if isempty(roi.boundary)
                        roi.plot_boundary = [];
                    else
                        xdata = roi.boundary{1}(:,2);
                        ydata = roi.boundary{1}(:,1);
                        xdata = reshape(xdata,[],1);
                        ydata = reshape(ydata,[],1);

                        for j = 2:length(roi.boundary)
                            x = roi.boundary{j}(:,2);
                            y = roi.boundary{j}(:,1);
                            x = reshape(x,[],1);
                            y = reshape(y,[],1);
                            xdata = vertcat(xdata,nan,x);
                            ydata = vertcat(ydata,nan,y);
                        end
                        roi.plot_boundary = [xdata,ydata];
                    end
                end
                roi.version = '2018-11';
                
%                 switch roi.version
%                     case '2018-04'
%                         %%
%                     case '2018-11'
%                         %% pass
%                     otherwise
%                         error('Unknown roi version');
%                 end
                
                updated_rois(i) = roi;
            end
        end
    end
    
    methods
        function self = ROI(mask_or_struct)
            % RoI object.
            %
            %   roi = calendula.RoI(mask)
            %   roi = calendula.RoI(struct)
            % 
            %   REQUIRED
            %   mask        - (logical mat)
            %                   Creates a new roi based on the positive
            %                   values in mask.
            %   struct      - (struct(s) of a calendula.RoI or calendula.RoI)
            %                   A valid struct is created by RoI.to_struct().
            %                   The output is an exact copy of the rois in
            %                   the struct/objects, just instantiated as seperate 
            %                   roi objects.
            %
            %   EXAMPLE 1
            %   roi = calendula.RoI(mask);
            %
            %   EXAMPLE 2
            %   % A list of RoI objects created with masks. 
            %   roi_array = ...
            %   roi_structs = roi_array.to_struct();
            %   % Make the structs back into RoIs
            %   roi_array = calendula.RoI(roi_structs);
            if nargin == 0
                return
            end
            
            if isempty(mask_or_struct)
                self = xylobium.moving_rois.old.ROI.empty(0);
                return;
            end
            
            switch class(mask_or_struct)
                case 'struct'
                    % Update the rois to the newest version.
                    rois = self.old_to_new(mask_or_struct);
                    
                    % Define output. 
                    self = xylobium.moving_rois.old.ROI.empty;
                    
                    % Assign every matching field to the property with the
                    % same name.
                    class_properties = properties(self);
                    roi_fields = fieldnames(rois);
                    
                    for i = 1:length(rois)
                        for j = 1:length(roi_fields)
                            field = roi_fields{j};
                            if ismember(field,class_properties)
                                self(i).(field) = rois(i).(field);
                            end
                        end
                    end
                case 'logical'
                    self.set_mask(mask_or_struct);
                    self.id = begonia.util.make_uuid();
                    self.id_shared = self.id;
                otherwise
                    error('Wrong input of class ''%s''',class(mask_or_struct));
            end
        end
        
        
        function insert_xy_offset(roi,frame,x_offset,y_offset,relative)
            if nargin < 5
                relative = true;
            end
            
            if isempty(roi.frames)
                error('Cannot create xy offsets when the ''frames'' property is empty.')
            end
            
            frame = round(frame);
            x_offset = round(x_offset);
            y_offset = round(y_offset);
            
            if isempty(roi.xy_offsets)
                roi.xy_offsets(1,:) = [1,0,0];
                roi.xy_offsets(2,:) = [roi.frames,0,0];
                roi.update_xy_offset_vector();
            end
            
            % Change the offsets such that they are relative the position
            % of the roi at frame 'frame'.
            if relative
                x_offset = x_offset + roi.xy_offset_vector(frame,1);
                y_offset = y_offset + roi.xy_offset_vector(frame,2);
            end
            
            roi.xy_offsets(end+1,:) = [frame,x_offset,y_offset];
            
            [~,I] = sort(roi.xy_offsets(:,1));
            roi.xy_offsets = roi.xy_offsets(I,:);
            roi.xy_offsets = unique(roi.xy_offsets,'rows');
            
            roi.update_xy_offset_vector();
        end
        
        
        function update_xy_offset_vector(roi)
            if isempty(roi.xy_offsets)
                roi.xy_offset_vector = [];
            end
            
            roi.xy_offset_vector = zeros(roi.frames,2);
            for i = 1:size(roi.xy_offsets,1)-1
                i1 = roi.xy_offsets(i  ,1);
                i2 = roi.xy_offsets(i+1,1);
                
                N = i2-i1+1;
                
                x1 = roi.xy_offsets(i  ,2);
                x2 = roi.xy_offsets(i+1,2);
                
                y1 = roi.xy_offsets(i  ,3);
                y2 = roi.xy_offsets(i+1,3);
                
                % x offset vector.
                roi.xy_offset_vector(i1:i2,1) = linspace(x1,x2,N);
                roi.xy_offset_vector(i1:i2,2) = linspace(y1,y2,N);
            end
            roi.xy_offset_vector = round(roi.xy_offset_vector);
        end
        
        function clear_xy_offsets(rois)
            for i = 1:length(rois)
                roi = rois(i);
                roi.xy_offsets = [];
                roi.xy_offset_vector = [];
            end
        end
        
        
        function update_tseries_info(rois,tseries,cycle,channel)
            for i = 1:length(rois)
                rois(i).tseries_id = tseries.uuid;
                if nargin > 2
                    rois(i).cycle = cycle;
                    rois(i).channel = channel;
                end
                rois(i).frames = tseries.frames_in_cycle(rois.cycle);
            end
        end
        
        
        function create_new_ids(rois)
            % Give all the rois new ids. 
            for roi = rois
                rois.replace_id(roi.id,begonia.util.make_uuid());
            end
        end
        
        
        function replace_id(rois,id_old,id_new)
            % Replaces RoI.id and RoI.id_connected_roi from id_old with id_new.
            
            old = strcmp({rois.id},id_old);
            [rois(old).id] = deal(id_new);
            
            old = strcmp({rois.id_connected_roi},id_old);
            [rois(old).id_connected_roi] = deal(id_new);
            
            old = strcmp({rois.id_shared},id_old);
            [rois(old).id_shared] = deal(id_new);
        end
        
        
        function structs = to_struct(rois)
            if isempty(rois)
                structs = struct([]);
            end
            warning('off','all')
            for i = 1:length(rois)
                structs(i) = struct(rois(i));
            end
            warning('on','all')
        end
        
        function new_roi = merge(rois)
            % Create a new roi with the union of the masks of the input
            % rois.
            mask = rois(1).mask;
            for i = 2:length(rois)
                mask = mask | rois(i).mask;
            end
            
            new_roi = calendula.RoI(mask);
            new_roi.channel = rois(1).channel;
            new_roi.cycle = rois(1).cycle;
            new_roi.shape = rois(1).shape;
            new_roi.group = rois(1).group;
        end
        
        function set_mask(roi, mask)
            assert(length(size(mask)) == 2);
            mask = logical(mask);
            roi.mask = mask;
            [roi.y,roi.x] = find(mask);
            roi.boundary = bwboundaries(mask);
            cx = mean(roi.x);
            cy = mean(roi.y);
            roi.center = [cx, cy];
            roi.area = length(roi.y);
        end
        
        function move(rois, shift)
            shift = round(shift);
            for roi = rois
                mask = circshift(roi.mask, [shift(2), shift(1)]);
                roi.set_mask(mask);
            end
        end
        
        function bool = is_in_roi(roi, x, y)
            %ISINROI check if the point (x,y) is a part of the roi.
            %   val = isInRoi(roi,x,y,)
            %
            % roi       - Single RoI object.
            % x         - (int) Position in image as pixels.
            % y         - (int) Position in image as pixels.
            %
            % See also: RoI

            v1 = uint32([x, y]);
            v2 = [roi.x, roi.y];
            bool = ismember(v1,v2,'rows');
        end
        
        function vec = calculate_signal_from_mat(roi, mat)
                
            frames = size(mat,3);
            
            if isempty(roi.xy_offsets)
                % This is a static ROI. 

                % Crop a square around the roi.
                min_x = min(roi.x);
                min_y = min(roi.y);
                max_x = max(roi.x);
                max_y = max(roi.y);

                min_x = max(min_x,1);
                min_y = max(min_y,1);
                max_x = min(max_x,size(mat,2));
                max_y = min(max_y,size(mat,1));

                roi_mask = roi.mask(min_y:max_y, min_x:max_x);
                roi_mask = repmat(roi_mask, 1, 1, frames);

                % Crop a 3d matrix of the same size out of the recording. 
                chunk = mat(min_y:max_y, min_x:max_x, :);

                % Areas not in the mask are zero. 
                chunk(~roi_mask) = 0;

                vec = sum(sum(chunk, 1), 2) / length(roi.x);
                vec = reshape(vec,[],1);
            else
                % This is a moving ROI. 
                
                % Create a 3d mask which is offset correctly.
                frames = size(roi.xy_offset_vector,1);
                roi_mask = roi.mask;
                roi_mask = repmat(roi_mask,1,1,frames);
                
                for i = 1:frames
                    x_offset = roi.xy_offset_vector(i,1);
                    y_offset = roi.xy_offset_vector(i,2);
                    roi_mask(:,:,i) = circshift(roi_mask(:,:,i),x_offset,2);
                    roi_mask(:,:,i) = circshift(roi_mask(:,:,i),y_offset,1);
                end
                
                % Find a crop around the ROI. 
                min_x = min(roi.x);
                min_y = min(roi.y);
                max_x = max(roi.x);
                max_y = max(roi.y);
                
                % Add offset from the moving roi. 
                min_x = min_x + min(roi.xy_offset_vector(:,1));
                min_y = min_y + min(roi.xy_offset_vector(:,2));
                max_x = max_x + max(roi.xy_offset_vector(:,1));
                max_y = max_y + max(roi.xy_offset_vector(:,2));
                
                % Safety to keep all indices within the image 
                min_x = max(min_x,1);
                min_y = max(min_y,1);
                max_x = min(max_x,size(mat,2));
                max_y = min(max_y,size(mat,1));
                
                % Crop mask and mat
                roi_mask = roi_mask(min_y:max_y, min_x:max_x, :);
                chunk = mat(min_y:max_y, min_x:max_x, :);
                
                % Areas not in the mask are zero. 
                chunk(~roi_mask) = 0;
                
                vec = sum(sum(chunk, 1), 2) / length(roi.x);
                vec = reshape(vec,[],1);
            end
        end
        
    end
end

