% this class loads and sets up a custom xylobium data editor that will be 
% the bases for the stack processor.  StackProcessorEditor is a class that
% inherits from Editor and adds a config property to it. This property is
% then read by the actions as needed (actions are passed the dloc, editor
% and model when called, so extending the editor is a way to pass custom
% information to them, or giving then calls specific to their context)
classdef StackProcessor < handle
    
    properties
        editor
        config
    end
    
    methods
        function obj = StackProcessor(input, config)
            import xylobium.stackproc.*;
            
            % if no arguments, try to grab it:
            if nargin < 1
                [sheet, path] = uigetfile({'*.xlsx';'*.xls';'*.csv'}, 'Setup sheet (XLSX) or path to scan');
                if sheet == 0
                    error('Must provide a table of tseries to process');
                end
                input = fullfile(path, sheet); 
            end
            
            % default configuration if none is provided:
            if nargin < 2
                config = StackProcessorConfig();
            end
            obj.config = config; 
            
            % are we given a path:
            if isa(input, 'char') 
                if endsWith(input, '.xlsx', 'IgnoreCase', true)
                    % we are given a Excel-sheet with series:
                    obj.config = StackProcessorSheetConfig(input);
                    dlocs = obj.loadSeries();
                    dlocs = fliplr(dlocs);  % usually we add to bottom
                else
                    % we are given a path - shall we scan?
                    choice = questdlg('Scan selected path for raw stacks? Might take a really long time for large disks'...
                        , 'Scan path?'...
                        , 'Yes', 'No thanks', 'No thanks');
                    if choice == "Yes"
                        scans = begonia.scantype.find_scans(input);
                        dlocs = scans(arrayfun(@(d) isa(d, 'begonia.scantype.TSeries'), scans));
                    else
                        return;
                    end
                end
                
            elseif isa(input, 'begonia.scantype.TSeries')
                dlocs = input;
            else
                error('Stack processor can only read paths, XLSX sheets or begonie.scantype.TSeries objects');
            end
            
            % variables:
            vars = {'name', 'start_time' , 'path' , 'phase_offset_px'...
                , 'ncconfig', 'zip', 'aligned', 'proxy'...
                , 'vidprev', 'fulltrace', 'roa_array'};
            
            % set up mods:
            mods = xylobium.dledit.model.Modifier.empty();
            mods(1) = xylobium.dledit.mods.DateTimeMod('start_time', 'dd/MM/y HH:mm:ss');
            mods(2) = xylobium.stackproc.mods.AlignmentSettingsModifier('ncconfig');
            
            % actions
            actions = obj.setupActions();
            
            %obj.editor = xylobium.dledit.Editor(dlocs, actions, vars);
            obj.editor = StackProcessorEditor(dlocs, actions, vars, mods);
            
            % this will give actions additional configuration information
            obj.editor.config = obj.config;
            
            % better default column width for this setup:
            obj.editor.datagrid.table.ColumnWidth = {180, 150};
        end
      
        
        function dlocs = loadSeries(obj)
            tab = readtable(obj.config.setup_sheet);
            cachefile = [char(obj.config.setup_sheet) '_cached.mat'];
            
            % check for cache:
            wbar = waitbar(0, 'Checking cache');
            if exist(cachefile, 'file')
                cache = load(cachefile);
                dlocs = cache.dlocs;
                warning(['Using cached tseries - if problematic, delete:' cachefile]);
            else
                dlocs = begonia.data_management.DataLocation.empty;
            end
            
            % load stacks not in cache:
            waitbar(0, wbar, 'Loading stacks');
            for i = 1:length(tab.Series)
                p = char(tab.Series(i));
                dloc = dlocs(strcmp({dlocs.path}, p));
                if isempty(dloc)
                    try
                        dloc = begonia.scantype.prairie.TSeriesPrairie(p);
                        dlocs = [dlocs dloc];
                    catch err
                        obj.handleStackLoadError(err, p)
                    end
                end

                
                waitbar(i/length(tab.Series), wbar);
            end
            
            % save dlocs to cache to load faser in future:
            waitbar(1, wbar, 'Saving cache..');
            save(cachefile, 'dlocs');
            disp(['Cached saved to: ' cachefile])
            close(wbar);
        end
        
        
        function actions = setupActions(obj)
            import xylobium.dledit.*;
            import xylobium.stackproc.actions.*;
            
            % open somehow:         
            ac1 = Action('MatVewier...' ...
                , @(d, m, e) matViewer(d, m, e), false, true);
            
            ac14 = Action('List processed...' ...
                , @(d, m, e) processed_stacks(d, m, e), false, true);
           
            ac2 = Action('Conf. phase corr.' ...
                , @(d, m, e) configPhaseCorrect(d, m, e), false, true);
            
            ac3 = Action('Conf. NoRMcorre' ...
                , @(d, m, e) configNoRMCorre(d, m, e), false, true);
            
            ac4 = Action('Compress ZIP' ...
                , @(d, m, e) compressZip(d, m, e), true);
            
            ac5 = Action('(1) Align' ...
                , @(d, m, e) align_memmat(d, m, e), true, false, 'control-1');
            
            ac6 = Action('(2) Preview gen.' ...
                , @(d, m, e) preview(d, m, e), true, false, 'control-2');
            
            ac8 = Action('(3) RoA extraction' ...
                , @(d, m, e) extract_roa(d, m, e), true, false, 'control-3');
            
            ac12 = Action('(4) Make proxy stack' ...
                , @(d, m, e) makeProxy(d, m, e), true, false, 'control-4');
            
            ac13 = Action('(5) Sync stack vars' ...
                , @(d, m, e) sync_stacks(d, m, e), true, false, 'control-5');
            
            
            actions = [ac1 ac14 ac2 ac3 ac4 ac5 ac6 ac8 ac12 ac13];
        end
        
        function [a, b] = handleStackLoadError(~, ~, varargin)
            disp(['Cannot convert (faulty series?) : ' char(varargin)])
            a = [];
            b = [];
        end
    end
end

