classdef Analyser < handle
    %VALLEYDISTANCEANALYSER Summary of this class goes here
    %   Detailed explanation goes here
    
    properties
        stack
        mat
        config
        
        fig_hidden
        ax

        status
        analysed
        current_frame
        
        results
        diagnostic
       
        keep_diagnostic_data;
    end
    
    properties (Transient)
       current_frame_nr
       is_final_frame
    end
    
    properties (Access=private)
        current_frame_nr_
    end
    
    methods
        function obj = Analyser(stack, keep_diagn, config)
            
            % if no config is given, extract it from the current tsereies,
            % or squeel if no such thing is found:
            if nargin < 3
                if ~stack.has_var('vestool_config')
                    error('No config given, and tseries provided has not config set');
                else
                    config = stack.load_var('vestool_config');
                end
            end
            
            if nargin < 2
               keep_diagn = false; 
            end
            
            % setup object:
            obj.stack = stack;
            obj.config = config;
            obj.keep_diagnostic_data = keep_diagn;
            obj.mat = stack.get_mat(obj.config.cycle, obj.config.channel);
            
            % initial frame settings = fram 1:
            obj.current_frame_nr_ = 1;
            obj.analysed = false;
            
            % resulting values from the analysis goes into preallocated
            % tables:
            obj.results = struct();
            len_t = stack.frames_in_cycle;
            len_m = length(config.markers);

            obj.results.maker_names = {obj.config.markers.name }';

            obj.results.time_s = (1:len_t)' * stack.dt;  % time in seconds
            obj.results.frame_analysed = false(len_t, 1);
            obj.results.diam_min_mean = nan(len_t, len_m); 
            obj.results.diam_left_right_dynamic_mean = nan(len_t, len_m); 
            obj.results.diam_locked_mean = nan(len_t, len_m);
            obj.results.diam_half_peak = nan(len_t, len_m);
            obj.results.area_under_min_treshold = nan(len_t, len_m);
            
            obj.results.errors = strings(len_t, len_m);
            
            obj.results.channel_mean_fluo = nan(len_t,1);

%             obj.results.fcirc_center = nan(len_t, len_m);
%             obj.results.fcirc_radius = nan(len_t, len_m);
%             obj.results.fcirc_count = nan(len_t, len_m);

            % hidden figure for functions that need an axes in their
            % analysis:
            obj.fig_hidden = figure('Visible', 'off');  
            imagesc(rand(4,4));
            obj.ax = obj.fig_hidden.CurrentAxes();
            
        end
        
        
        % gets the current frame number:
        function fnr = get.current_frame_nr(obj)
            fnr = obj.current_frame_nr_;
        end
        
        % sets the current frame number:
        function set.current_frame_nr(obj, fnr)
            obj.current_frame_nr_ = fnr;
            obj.current_frame = [];
        end
        
        % set next frame as current frame. Error if frame is beyon last.
        function next(obj)
            % cannot advance on final frame:
            assert(~obj.is_final_frame, 'At final frame. Use: is_current_frame');
            obj.current_frame_nr = obj.current_frame_nr + 1;
        end

        % true if current frame is last frame of the stack:
        function isfin = get.is_final_frame(obj)
            isfin = obj.current_frame_nr == obj.stack.frames_in_cycle;
        end
        
        % analyse current frame:
        function analyse_current(obj)
            import xylobium.vesseltool.vdist.*;
            
            % extract current frame's mat:
            fnr = obj.current_frame_nr;
            avg_grab = round(obj.config.average/2);
            fs = fnr-avg_grab:fnr+avg_grab; % frames we'll average:
            
            % trim frames to ensure we dont exceed matrix bounds, then
            % retrieve and average:
            fs = fs(fs > 0 & fs < obj.stack.frames_in_cycle);
            mat_fs = obj.mat(:,:,fs);
            obj.current_frame = mean(mat_fs,3);
            
            % for each marker, run the analysis and collect the results:
            for m_idx = 1:length(obj.config.markers)
                m = obj.config.markers(m_idx);
                err = "";
                try
                [diam_m, dima_lrd, diam_l, diam_hp, aumt, chmean, lscan, means, mat_rot] = analyse_frame( ...
                    obj.current_frame ...
                    ,  [m.start_point(1) m.end_point(1)] ...
                    , [m.start_point(2) m.end_point(2)] ...
                    , obj.ax ...
                    , 3 ...
                    , m.locked_means);
                catch err_caught
                    diam_m = nan;
                    dima_lrd = nan;
                    diam_l = nan;
                    diam_hp = nan;
                    aumt = nan;
                    chmean = nan;
                    err = string(err_caught.getReport());
                end
                
                % basic measured values:
                obj.results.diam_min_mean(fnr, m_idx) = diam_m;
                obj.results.diam_left_right_dynamic_mean(fnr, m_idx) = dima_lrd;
                obj.results.diam_locked_mean(fnr, m_idx) = diam_l;
                obj.results.diam_half_peak(fnr, m_idx) = diam_hp;
                obj.results.area_under_min_treshold(fnr, m_idx) = aumt;
                obj.results.channel_mean_fluo(fnr) = chmean;
                
                obj.results.errors(fnr, m_idx) = err;
                
                % extended diagnostic values if asked to keep:
                if obj.keep_diagnostic_data
                    
                    
                end
                
                % mark frame as analysed:
                obj.results.frame_analysed(fnr,1) = 1;
            end
            
            obj.analysed = true; 
        end
        
        
        % loops through the matrix at given frames, and 
        function analyse_frames(obj, frames, save)
            if nargin < 1
                frames = 1:obj.config.ds:obj.stack.frames_in_cycle;
            end
            
            if nargin < 3 
                save = true;
            end
            
            % loop through all frames, forcing update:
            for fnr = frames
               obj.current_frame_nr = fnr;
               obj.analyse_current();
            end
            
            if save
                obj.write_dloc_vars(obj.stack);
            end
        end
        
        function write_dloc_vars(obj, dloc)
            dloc.save_var('vestool_results', obj.results);
        end
        
        
        
    end
end

