classdef ConfigView < handle
    %VALLEYDISTANCESETUPVIEW Summary of this class goes here
    %   Detailed explanation goes here
    
    properties
        fig
        fig_hidden % for frame analyser
        
        % dyamic property to get the current config at any time:
        config
        
        % matrix and stck:
        tseries
        mat
        
        channel
        cycle
        
        % axis:
        ax_zoomed;
        ax_valley;
        ax_timeseries;
        ax_overview;
        ax_hidden % for frame analyser
        
        % event handles to be deleted when figure is closed:
        ev_frame_listener;
        ev_marker_list_changed_listener;
        ev_marker_changed_listener;
        
        % timer to update only at max 1/2 Hz and avoid choppy line dragging
        % on events etc.
        update_timer
        timer_process_tick % the tick being processed, or last processed
        timer_newest_tick % the most recent 
        
        % struct to associate imdistline with marker:
        marker_lines
        line_selected
        marker_selected
        marker_selected_index
        
        % analyser object to perform analysis on mat for previews
        analyzer
        
        % matrix of current frame and reference frame:s
        frame_cur
        frame_ref
        frames_samples
        
    end
    
    properties (Transient)
        % current frame from stack of zoomed view and preview graph:
        current_frame_nr
    end
    
    properties (Access = private)
        current_frame_nr_;
    end
    
    events
       on_marker_added % fired when a marker is added
       on_marker_list_changed % fired when the list of markers is changed
    end
    
    methods
        function obj = ConfigView(tseries)
            obj.tseries = tseries;
            
            % simply use defaults for now:
            obj.channel = 1;
            obj.cycle = 1;
            
            obj.mat = obj.tseries.get_mat(obj.cycle, obj.channel);
            obj.frames_samples = [];
            
            % setup:
            obj.marker_lines = struct('line',{},'marker',{});
            obj.update_timer = timer('ExecutionMode', 'fixedDelay', 'Period', 0.1);
            obj.update_timer.TimerFcn = @(~,~) obj.handle_timer();
            obj.timer_newest_tick = 1;
            obj.timer_process_tick = 0; % forces redraw on start
            
            obj.setup_graphs();
            obj.setup_reference_image();

            % for debugging:
            obj.load_config();
            if length(obj.marker_lines) < 1
                obj.add_marker_line();
            end
            % go to first frame:
            obj.current_frame_nr = 1;
            start(obj.update_timer);
            
        end
        
        % creates axis of the subplot:
        function setup_graphs(obj)
            obj.fig = figure();
            obj.fig.Name = 'Vessel picker';
            obj.fig.MenuBar = 'none';
            obj.fig.ToolBar = 'figure';
            obj.fig.CloseRequestFcn = @(~,~) obj.close();
            
            % overview of current tseries:
            subplot(2,2,2);
            plot(1);
            obj.ax_overview = obj.fig.CurrentAxes();
            
            
            % selection area:
            subplot(2,2,1);
            plot(1);
            obj.ax_zoomed = obj.fig.CurrentAxes();
            
            % overview of current tseries:
            subplot(2,2,3);
            plot(1);
            obj.ax_timeseries = obj.fig.CurrentAxes();
            
            % view of current linescan valley:
            subplot(2,2,4);
            plot(1);
            obj.ax_valley = obj.fig.CurrentAxes();
            
            % events:
            obj.fig.KeyPressFcn = @(s,e) obj.handle_keyboard(s,e);
            
            % hidden figure for analyser:
            obj.fig_hidden = figure('Visible', 'off');  
            plot(1);
            obj.ax_hidden = obj.fig_hidden.CurrentAxes();
        end

        
        % get or create reference image (also used to get locked means):
        function setup_reference_image(obj)
            % FIXME: should use f0 image!
            if obj.tseries.has_var('img_f0')
                obj.frame_ref = obj.tseries.load_var('img_f0');
            else
                obj.frame_ref = obj.tseries.get_avg_img(obj.config.cycle, obj.config.channel);
            end
            
            % make a stack-wide averge of 100 frames for reference:
%             ref_samples = floor(linspace(1, size(obj.mat,3)-1000, 10));
%             mat_ref = double(obj.mat(:,:,ref_samples));
%             obj.frame_ref = mean(mat_ref, 3);
            himg = imagesc(obj.frame_ref, 'parent', obj.ax_overview);
            title(obj.ax_overview, 'Line selector')
            himg.ButtonDownFcn = @(s,ev) disp('click');
        end
        
        
        % handle the 
        function handle_keyboard(obj, ~, ev)
            code = xylobium.shared.keyboard_event_to_str(ev);
            
            switch (code)
                
                case 'shift-a'
                    obj.add_marker_line();
                    
                case 'shift-r'
                    wb = waitbar(10, 'Calculating preview');
                    obj.run_multiframe_test();
                    delete(wb);
                    
                case 'shift-c'
                    obj.clear_multiframe_test();
                    
                case 'shift-s'
                    obj.save_config();
                    
                case 'rightarrow'
                    obj.select_marker('next')
                    
                case 'leftarrow'
                    obj.select_marker('prev');
                    
                otherwise
                    disp(code);
            end
        end
        
        
        % executes the timer:
        function handle_timer(obj)
            if obj.timer_process_tick < obj.timer_newest_tick
                try
                    obj.update_preview();
                    obj.timer_process_tick = obj.timer_newest_tick;
                catch err
                    disp('Timer error:')
                    err.stack.name
                end
            end
        end
        
        
        function handle_line_change(obj, line)
            obj.invalidate_gui();
            obj.line_selected = line;
            
            obj.marker_selected = [];
            
            for ml = obj.marker_lines
                if ml.line == line
                    obj.marker_selected = ml.marker;
                end
            end
            
            obj.fig.Name = obj.marker_selected.name;
            obj.fig.NumberTitle = 'off';
        end
        
        
        function invalidate_gui(obj)
            obj.timer_newest_tick = obj.timer_newest_tick + 1;
        end
        
        
        
        % getter and setter for current frame:
        function set.current_frame_nr(obj, fnr)
            obj.current_frame_nr_ = fnr;
            obj.frame_cur = obj.mat(:,:,fnr);
            obj.invalidate_gui();
        end
        
        function fnr = get.current_frame_nr(obj)
            fnr = obj.current_frame_nr_;
        end
        
        
        % adds a marker with an associated imdistline, or creats a new
        % marker if no paramter is given:
        function add_marker_line(obj, marker)
            % if no marker given, assume we want to create a new one:
            if nargin < 2
                marker = xylobium.vesseltool.VesselMarker();
            end
            
            ml_idx = length(obj.marker_lines) + 1;
            line = imdistline(obj.ax_overview);
            line.setLabelVisible(false);
            obj.marker_lines(ml_idx) = struct('marker', marker, 'line', line);
            
            line.setPosition( ...
                [marker.start_point(1) marker.end_point(1)] ...
                , [marker.start_point(2) marker.end_point(2)]);
            
            % event to update preview when the line is changed:
            line.addNewPositionCallback(@(~) obj.handle_line_change(line));
            
            notify(obj, 'on_marker_added');
            notify(obj, 'on_marker_list_changed');
        end
        
        
        % selects a given marker:
        function select_marker(obj, marker)
            line = [];
            % unmark all makers:
            for ml = obj.marker_lines
                ml.line.setColor('white');
                
                if ml.marker == marker
                    line = ml.line;
                    line.setColor('red');
                end
            end
            
            obj.marker_selected = marker;
            obj.update_preview();
        end
        
        function next_marker(obj)
            
        end
        
        function prev_marker(obj)
        
        end
        
        
        % updates the preview and valley data of 
        function update_preview(obj)
            obj.clean_invalid_markers();
            
            % skip if no marker selected:
            if isempty(obj.line_selected) || isempty(obj.marker_selected)
               return; 
            end
            
            % draw 
            marker = obj.marker_selected;
            obj.update_markers();
            xs = [marker.start_point(1) marker.end_point(1)];
            ys = [marker.start_point(2) marker.end_point(2)];
                        
            % analyse current frame
            [diam, diam_lrd, diam_l, diam_hp, aut, chmean, lscan, means, mat_rot, mat_aut, peaks, peak_interceps] = ...
                xylobium.vesseltool.vdist.analyse_frame( ...
                    obj.frame_ref, xs, ys, obj.ax_hidden, 3, [100 100]);
            
            % print the zoomed preview:
            hold(obj.ax_zoomed, 'off');
            imagesc(obj.ax_zoomed, mat_rot);
            title(obj.ax_zoomed, marker.name);
            
            % update complex graphs:
            obj.update_lscan_graph(obj.ax_valley, lscan, means, peaks, peak_interceps);
            
            % keep means as the locked means:
            obj.marker_selected.locked_means = means;
        end
        
        
        function update_lscan_graph(obj, ax, lscan, means, peaks, peak_interceps)
            mid = floor(length(lscan)/2);

            cla(ax);
            
            % draw the lscan graph:          
            hold(ax, 'off');
            plot(ax, lscan, ':');
            hold(ax, 'on');
            plot(ax, lscan(1:mid));
            
            left_mean = mean(lscan(1:mid));
            right_mean = mean(lscan(mid:end));
            
            line(ax, [0, mid], [left_mean left_mean]);
            yyaxis(ax, 'right')
            cla(ax);
            line(ax, [mid, length(lscan)], [right_mean right_mean]);
            hold(ax, 'on');
            yyaxis(ax, 'left')
            
            % draw line on mid:
            ylims = ylim(ax);
            line(ax, [mid, mid], [ylims(1) ylims(2)]);
            
            if ~isnan(peaks(1))
                plot(ax, peaks,[lscan(peaks(1)) lscan(peaks(2))],'r*');
            end
            
            title(ax, 'Distance measurement');
        end
        
        % tests the current marker accross multipel frames in stack
        function run_multiframe_test(obj, ax, marker)
            ax = obj.ax_timeseries;
            marker = obj.marker_selected;
            
            fns = floor(linspace(1,obj.tseries.frames_in_cycle-1000,30));
            if isempty(obj.frames_samples)
                obj.frames_samples = obj.mat(:,:,fns);
            end
            xs = [marker.start_point(1) marker.end_point(1)];
            ys = [marker.start_point(2) marker.end_point(2)];
            
            diams = zeros(length(fns), 1);
            diams_l = zeros(length(fns), 1);
            for f = 1:size(obj.frames_samples,3)
                frame = obj.frames_samples(:,:,f);
                [diam, diam_lrd, diam_l, diam_hp] = ...
                    xylobium.vesseltool.vdist.analyse_frame( ...
                        frame, xs, ys, obj.ax_hidden, 3, [100 100]);
                diams(f,1) = diam;
                diams_l(f,1) = diam_l;
            end
            
            %cla(ax);
            hold(ax, 'on');
            plot(ax, fns * obj.tseries.dt, smooth(diams) * obj.tseries.dx);
            %plot(ax, fns * obj.tseries.dt, smooth(diams_l), ':');
            title(ax, 'Preview (shift-c to clear)');
            xlabel(ax, 'Time (s)');
            ylabel(ax, 'Diameter (µm)');
            
        end
        
        
        function clear_multiframe_test(obj)
            cla(obj.ax_timeseries);
        end
        
        
        % constructs a config object based on the current markers:
        function conf = get.config(obj)
            % get the markers for each imdistline, and reload the values
            % for that marker:
            obj.clean_invalid_markers();
            obj.update_markers();
            conf = xylobium.vesseltool.vdist.Config();
            for ml = obj.marker_lines
                
%                 % update the marker's position to match line:
%                 m = ml.marker;
%                 lpos = ml.line.getPosition();
%                 m.start_point = [lpos(1,1), lpos(1,2)];
%                 m.end_point = [lpos(2,1), lpos(2,2)];
%                 
                % add marker details from imdisline:
                conf.markers(end + 1) = ml.marker;
            end
        end
        
        % saves the current config to the tseries:
        function save_config(obj) 
            obj.clean_invalid_markers()
            obj.update_markers();
            obj.tseries.save_var('vestool_config', obj.config);
            
            obj.fig.Name = ['Saved markers' char(datetime)];
        end
        
        
        % loads a previouslky saved config, if present
        function loaded = load_config(obj)
            % simply return false if no config is loaded. If no markers are
            % loaded, the tool will add some new lines in the costructor:
            loaded = false;
            if ~obj.tseries.has_var('vestool_config')
               return; 
            end
            
            % load previosu config from this tseries:
            conf = obj.tseries.load_var('vestool_config');
            
            % load markers from config:
            for marker = conf.markers
                obj.add_marker_line(marker);
            end
            
            % this means markers have changed:
            notify(obj, 'on_marker_list_changed');
        end
        
        
        % update all markers with their references line's position:
        function update_markers(obj)
            
            for i = 1:length(obj.marker_lines)
                ml = obj.marker_lines(i);
                
                % update the marker's position to match line:
                m = ml.marker;
                lpos = ml.line.getPosition();
                m.start_point = [lpos(1,1), lpos(1,2)];
                m.end_point = [lpos(2,1), lpos(2,2)];
                
                % update marker val
                obj.marker_lines(i).marker = m;
            end
        end
        
        % there is no easy way to detect if imdistline is deleted AFAIK, so
        % this function cleans the list and is called when an update list
        % is needed:
        function clean_invalid_markers(obj)
            pre_count = length(obj.marker_lines);
            valids = arrayfun(@(mlc) isvalid(mlc.line), obj.marker_lines);
            obj.marker_lines = obj.marker_lines(valids);
            
            if pre_count ~= length(obj.marker_lines)
                notify(obj, 'on_marker_list_changed');
            end
        end

        
        % clears event listeners to prevent actions on deleted objects
        function close(obj, save)
            if nargin < 2
                save = true;
            else
                save = false;
            end
            if save
                disp('Auto-saving on close');
                obj.save_config();
            end
            
            % clean up timer and objects:
            stop(obj.update_timer)
            delete(obj.update_timer);
            delete(obj.fig_hidden);
            delete(obj.fig);
            delete(obj.ev_frame_listener);
            delete(obj.ev_marker_changed_listener);
            delete(obj.ev_marker_list_changed_listener);
            delete(obj);
        end
        

    end
end








