% Low level function that analyses individual frames. Use Analyser object
% to iterate throug a stack, and recieve the outputs of this function i a
% more structured manner.
function [diam_minmean, diam_lrd, diam_l, diam_hp, aut, chmean, lscan, means, mat_rot, mat_aut, peaks, peak_intercepts] = ...
        analyse_frame(mat, xs, ys, ax, span, locked_means)
    
    if nargin < 5
        span = 3;
    end

    % draw image on fig, measure angle of the line:
    imagesc(mat, 'parent', ax);
    mline = imdistline(ax, xs, ys);
    
    % make measurements: we will rotate the image, then grab the central
    % matrix from top to bottom, and mean each row to get a profile of the
    % valley of low fluoresence expected in the middle. We'll then assume
    % the vessel walls are steep, andmeasure between the median fluoresence
    % values on either side to get a distance. We'll do this using: locked
    % mean from f0 image, median values on the half of the scan that is
    % lowest, and medial from either side:
    ang = 90 - mline.getAngleFromHorizontal();
    dist = mline.getDistance();
    padding = 0;
    
    % capture channel mean fluoresence:
    chmean = mean(mat(:));
    
    % crop to area around the line, then rotatE:
    cx = xs(1) +  (xs(2) - xs(1)) / 2;
    cy = ys(1) + (ys(2) - ys(1)) / 2;
    
    croprect = [cx - dist/2 - padding ...
        , cy - dist/2 - padding ...
        , dist + padding * 2 ...
        , dist + padding * 2];
    
    mat_crop = imcrop(mat, croprect);   % select are
    %mat_gaus = mat_crop
    mat_gaus = imgaussfilt(mat_crop,2); % smooth a little in space
    mat_rot = imrotate(mat_gaus, ang);  % rotate

    % get the imddle part, and use that as a linescan to get profile of
    % vessel:
    mat_rot_mid = floor(size(mat_rot,2)/2);
    lscan_mat = mat_rot(:,mat_rot_mid-span:mat_rot_mid+span)';
    lscan = mean(lscan_mat,1);
    
    % remove 10% on either side to counter the rotation:
    cut = floor(length(lscan) * 0.12);
    lscan = lscan(cut:end-cut);
    
    % 1) First analysis: treshold intersect with profile from midline:
    [diam_minmean, diam_lrd, means, lp, rp] = get_unlocked_dist(lscan);  
    [diam_l] = get_dist(lscan, locked_means);  
    if isempty(diam_l)
        diam_l = nan;   % this occurs rarely, but nan is better
    end
    
    % mark center of mat_rot + left and right points in the rot. matrix:
    mx = ceil(size(mat_rot,1)/2);
    my = ceil(size(mat_rot,2)/2);
    
    mat_rot(my, mx) = max(mat_rot(:));
    mat_rot(lp + cut, mx-1:mx+1) = repmat(max(mat_rot(:)),1,3);
    mat_rot(rp + cut, mx-1:mx+1) = repmat(max(mat_rot(:)),1,3);
   
    % 2) Second analysis::
    % calculate area of pixels under treshold value in cropped matrix. The
    % locked means are used to find valley side intersection, and less
    % usefull here. We will use values under the lowest of the means to
    % ensure we always are inside the valley. The rotated matrix will
    % contain an area of zeroes after the imrotate operation, and these
    % will be included, unfortunately. 
    aut_threshold = min(means);
    mat_aut = mat_crop < aut_threshold;
    aut_vals = mat_aut(:);
    aut = length(aut_vals(aut_vals));
    
    % 3) Analyse peak distance from half height to first peak from midline:
    [diam_hp, peaks, peak_intercepts] = get_half_peak_dist(lscan);
end


function [diameter_m, diameter_lrd, means, left_point, right_point] = get_unlocked_dist(lscan)
    % calculate left and right half means separately as the fluoresence in
    % dynamic measurements might be quite different on either side:
    lscan_s = smooth(lscan);
    mid = floor(length(lscan)/2);
    
    left_min = min(lscan_s(1:mid));
    right_min = min(lscan_s(mid:end));
    
    left_amp = max(lscan_s(1:mid)) - left_min;
    right_amp = max(lscan_s(mid:end)) - right_min;
    
    left_mean = left_min + (left_amp / 2);
    right_mean = right_min + (right_amp / 2);
    
%     left_mean = mean(lscan(1:mid));
%     right_mean = mean(lscan(mid:end));
    
    means = [left_mean right_mean];
    
    % pick the smallest of the means to do the dynamic measurement:
    means_min = [min(means) min(means)];
    [diameter_m] = get_dist(lscan, means_min);
    
    % get diameter using individual means from left and right side:
    [diameter_lrd, left_point, right_point] = get_dist(lscan, means);
end

function [diameter, left_point, right_point] = get_dist(lscan, means)
    mid = floor(length(lscan)/2);
    left_mean = means(1);
    right_mean = means(2);
    
    % find intersection of mean with graph: last on left side, and first on
    % the right side. Distance is the length of the line between the
    % interception:
    left_point = find(lscan(1:mid) > left_mean, 1, 'last');
    right_point = find(lscan(mid+1:end) > right_mean, 1, 'first') + mid;
    diameter = right_point - left_point;
end

function [diam_hp, peaks, intercepts] = get_half_peak_dist(lscan)

    % find peaks to the left and right of the center of scan:
    lscan_smooth = smooth(lscan);
    mid = floor(length(lscan)/2);
    min_prom = (max(lscan_smooth) - min(lscan_smooth))/2;
    [~,locs] = findpeaks(lscan);
    %[~,locs] = findpeaks(lscan, 'MinPeakProminence',50);
    
    peaks_left = locs(locs < mid);
    peaks_right = locs(locs > mid);
    if isempty(peaks_left) || isempty(peaks_right) 
        diam_hp = nan;
        peaks = [nan nan];
        intercepts = [nan nan];
        return;
    end
    
    % if we have peaks on both sides, we can get the closest ones to mid:
    pl = peaks_left(end);
    pr = peaks_right(1);
    peaks = [pl pr];
    
    % get values of peaks, minimum values, and tresholds to intercept:
    peak_vals = [lscan(pl), lscan(pr)];
    min_val = min(lscan(pl:pr));
    tresholds = [...
        min_val + (peak_vals(1) - min_val)/2 ...
        , min_val + (peak_vals(2) - min_val)/2];
    
    % find intercepts treshold-linescan:
    left_intercept = find(lscan(1:mid) > tresholds(1), 1, 'last');
    right_intercept = find(lscan(mid+1:end) > tresholds(2), 1, 'first') + mid;
    intercepts = [left_intercept right_intercept];
    diam_hp = right_intercept - left_intercept;
end



