# Begonia: A Matlab Package for 2-Photon Data

## Contents

* [Begonia (for imaging)]()

* [Yucca (for recrig trials)](#yucca)

* [Xylobium (GUI apps)](#xylobium)

* [Unstable)](#unstable)

## Dependencies
Alignment of calcium videos uses NoRMCorre. 
Some old stack formats that uses multipage tiffs need the TIFFStack library.

## Unittests and examples
Some of the functions have unittests, which also provide good examples on how to use the code. From root directory, run all tests with:

```Matlab
runtests('./tests', 'IncludeSubfolders', 1);
```

## Logging

Begonia is quiet by default.  Change the logging level to alter the amount of output the system provides globally:

```Matlab
% causes begonia to be more or less verbose in the command window:
begonia.util.logging.set_global_level(0) % no output
begonia.util.logging.set_global_level(1) % some output
begonia.util.logging.set_global_level(2) % full output
```

### Example 
With the terminal, go to your matlab directory and download the dependencies:
```
$ git clone https://github.com/flatironinstitute/NoRMCorre
$ git clone https://github.com/DylanMuir/TIFFStack
```


## Scantypes

![Alt text](docs/diagrams/CORE.png "Core classes")

In august 2018 we added a reimplementation of the core classes to clean up the inheritance and remove old work-arounds. A series of abstract classes provide a clean API that can be supported for multiple output types, including future microscopes. 

* Data management API: DataInfo, DataLocation
* Scan types API: TSeries, Linescan, ZStack, SingleImage

Future classes that adds supports for new microscopes and data sources need to inherit from the **scan type APIs, DataLocation and DataInfo** to be principally compatible with all begonia functions. 

The classes are mostly compatible with the old begonia, meaning you can instance them on the same paths.

```Matlab
% reading data from a raw tseries
import begonia.scantype.prairie.*;
path = '/path/to/tseries_folder';
TSeriesPrairie(path);
```

A new search function that outputs the new classes can be used to find all data in a directory structure.

```Matlab
% reading data from a raw tseries
path = '/path/to/tseries_folder';
begonia.scantype.find_scans(path);

% produces an output similar to this:

%{ 
scans = 

  1×24 heterogeneous DataLocation (SingleImagePrairie, TSeriesPrairie, TSeriesH5, ...) array with properties:

    path
    saved_vars
    dl_storage_engine
    dl_readonly
    dloc_metadata_dir
    adapter
    dl_unique_id
    dl_changelog
%}
```

## Alignment
Alignment of tseries are most simply done by specifying a folder. The alignment uses NoRMCorre.

### Example
```Matlab
folder_to_align = '/Volumes/tseries_folder'
begonia.stack_operations.align(folder_to_align)
```

## Stacks
Calcium recordings, either raw or aligned format, can be found using 
`begonia.scantypes.find_Scans`. 
All recording objects has a method `get_mat(cycle, channel)` which returns a lazy
3d matrix of the cycle and channel input. 
    The calsium recordings have other properties such as recording date, number of frames, etc. 
Look in `begonia.scantypes.Stack` to see all the properties

### Example
```Matlab
path = 'path_to_tseries';
stacks = begonia.scantypes.find_scans(path);
for stack = stacks
    for cy = 1:stack.cycles
        for ch = 1:stack.channels
            mat = stack.get_mat(cy,ch);
            
            % Make the matrix not lazy by extracting all dimensions.
            mat = mat(:,:,:);
        end
    end
end
```



## Filters
`begonia` includes some simple filters that can be used on N-dimensional matricies. 
Filters works on 2d matricies along the columns, but `begonia` also 
includes a function, `apply_filt`, 
that transforms any matrix to a 2d matrix such that the chosen dimension is put
along the columns.

### List of Filters
* `fft_one_side` - fft (fourier) transform.
* `gaussian` - Gaussian smoothing, similar to `imgaussfilt`, but not limited to 2d matricies. 
* `window` - Window smoothing, aka. running average.
* `histogram_peak` - Returns the most frequent value along the dimension, can be used to find 
    a baseline in signals such as df_f0 in calcium recordings.

### Example
```Matlab
% Create a matrix and filter along the columns.
mat = zeros(11,3);
mat(6,:) = 1;
sigma = 3;
mat_out = begonia.vector_functions.gaussian(mat,sigma)

% Run a filter along the rows (or any dimension)
window_in_pixels = 10;
filter_func = @(x) begonia.vector_functions.window(x,window_in_pixels);
mat = imread('cameraman.tif');
dimension = 2;
mat_out = begonia.vector_functions.apply_filt(mat,filter_func,2);
imagesc(mat_out)
```

## Data Management
`begonia` has methods to easily store data to files, across matlab sessions.
Variables can be saved using a `DataLocation` object that first must be initiated
by inputting a path. 
The returned object can then save and load variables using the `DataLocation.save_var` or
`DataLocation.load_var` functions. In the case the variables are very large matricies
the alternative functions `save_array` or `load_array` can be used instead. 

To view all the variables that have been saved, simply view the `DataLocation.saved_vars` 
property. 

Another tool is `Stash`, a class that can be used to store large objects and recover them later. Like DataLocation, stashing stores these objects to the disk as .mat files, but will not be dependen on a file system path. It can also take large datafiles out of tables, and repopulate the tables later. 

### Practical purpose

The idea behind `DataLocation` is that you can add data to any directory on your filesystem.  If you have a recording device that outputs a directory of files, you can then create a DataLocation with that path and add variables to it.  E.g., you may want to tag some directories to be included in your data analysis, while some may be skipped.  

### Example: Saving Data
2 ways of saving a variable.
```Matlab
import begonia.data_management.*

path = 'directory_to_save_variables';
obj = DataLocation(path);

% Saves variable_to_save_1 with the value 1.
variable_to_save_1 = 1;
obj.save_var(variable_to_save_1)
% Saves variable_to_save_2 with the value 2.
obj.save_var('variable_to_save_2',2)
```

Data locations keep a record of when variables were last updated. Using this information, they can sync with eachother to update variables based on wether they have been updated:

```Matlab
import begonia.data_management.*

dloc_1 = DataLocation(path_1);  
dloc_2 = DataLocation(path_2);

% save two vars:
dloc_1.save_var('first_var', 'first');
pause(1/100);
dloc_2.save_var('first_var', 'last');
dloc_2.save_var('second_var', 42);

% see write time of a variable:
dloc_2.dl_changelog('second_var') % = a datetime 

% this will not change dloc_2, since it's variables are newer:
dloc_2.dl_update_from(dloc_1);

% this should overwrite 'first_var', and include 'second var':
dloc_1.dl_update_from(dloc_2);
dloc_1.load_var('first_var')  % = 'last'
dloc_1.load_var('second_var')  % = 42
```

In a similar fashion, you can forcefully update a datalocation with variables from another. The result of this line, is to make dloc_2 an exact copy of dloc_1, regardless of wether dloc_1's variables were older:

```Matlab
% overwrite all variables in dloc_2 with variables from dloc_
% (and add missing):
dloc_2.dl_overwrite_from(dloc_1);
```

You can also store the actual data in your variables in an external directory instead of the data location's path.  This will then only generate an ID file at the path, and and otherwise leave the dir at 'path' unchanged.  This keeps the original dirs cleaner, and eases sharing of the data.  **Please note** that this unique ID is created when the first variable is saved.  It can be overridden by setting `dl_unique_id` on the data location object.

```Matlab
import begonia.data_management.*;

% create a data location:
dl = DataLocation('C:\dltest1');

% use an alternate off-path storage engine to store varilables in another dir:
ope = OffPathEngine('C:\variables');
dl.dl_storage_engine = ope;

% saving variables will now write them to 'C:\variables', not 'C:\dltest1':
dl.save_var('my_var', 42)
my_var_retrieved = dl.load_var('my_var');
```

### Example: Loading Data
3 ways of load a variable.
```Matlab
path = 'directory_to_save_variables';
obj = begonia.data_management.DataLocation(path);

% Load variable_to_save_1
obj.load_var('variable_to_save_1');
disp(variable_to_save_1);
% Load variable_to_save_2
variable_to_save_2 = obj.load_var('variable_to_save_2');
% Load variable_to_save_3, if it does not exist load a default variable instead.
default_var = 3;
variable_to_save_3 = obj.load_var('variable_to_save_3',default_var);
```


## Stash

Use stash to reduce the size of tables or stashes of data you mig

```Matlab
% stash some data:
stash = begonia.data_management.Stash("C:\Users\knuta\Desktop\stash");
thing = rand(100,100,100);
url = stash.put('randcol', 'random-1', thing);

% recover the data:
thing = stash.get(url);

% providing a default value prevents crashing on url missing:
words = stash.get(urls, "nsfw");
```
You can do this for multiple items at a time. `Stash.put` is not vectorized because you could not tell if the programmer wants to stash the list, or each item within it.

```Matlab
% stash multiple:
words = ["hello", "world"];

% if you dont want to create your own idents, make them with uuid. You can
% however supply a series of ids you have made yourself to ensure data does
% not get written more than once:
idents = begonia.util.make_uuids(length(words));
urls = stash.puts('dictionary', idents, words);

% get them back: 
words = stash.gets(urls);

% you can define a default value as well. This prevents crashing on url missing.
words = stash.gets(urls, "fubar");
```

Even more powerful, you can use Stash to take large heavy data out of a matlab table, such as RoI traces or matricies. Note that this operation generates *random IDs*, meaning running the operation twice will duplicate data on disk. 

You might want to keep a namespace for each table

```Matlab
% make a dummy table:
ts_name(1,:) = "TS1";
ts_name(2,:) = "TS2";
ts_name(3,:) = "TS3";

data(1,:) = {rand(100,100,100)};
data(2,:) = {rand(100,100,100)};
data(3,:) = {rand(100,100,100)};

meta(1,:) = {"<thing>alpha</thing>"};
meta(2,:) = {"<thing>beta</thing>"};
meta(3,:) = {"<thing>gamma</thing>"};

tab = table(ts_name, data, meta);

% stash the columns 'data' and 'meta':
% you need to specify a namespace and what columns you want to stash:
stash.clear_namespace('mytable');
tab_stashed = stash.stash_table('mytable', tab, ["data", "meta"]);

% get the data back - no parameters needed beyond table:
tab = stash.unstash_table(tab_stashed);
```

## Megatable

The megatable enables alignment and resampling of multiple traces such as electrophysiology, calsium data, motion data and more. 

It also enables filtering of these traces based on e.g. sleep or motion states.

To see the example and unittest, run the following code in matlab.
```Matlab
% please dont actually edit this file ;)
% make a copy if you want to experiment
edit tests.begonia.megatable

% how to get episodes from multiple traces from different groups:
edit tests.begonia.megatable_group_episodes
```


## RoA heatmaps



This module creates roa heatmaps and includes a simple visualization of the map and a tool to configure the default color range of the map manually.

![Alt text](docs/screenshots/heatmap_config.PNG "Core classes")

```Matlab
% creates heatmap and a suggested caxis range:
[hmap, cax] = begonia.stack_modules.roa_heatmap.read(stack)

% same, but writes to datalocation (can be multiple stacks)
begonia.stack_modules.roa_heatmap.write(stacks)

% note: you can also do this for e.g. neuron roa masks by specifying mask name:
begonia.stack_modules.roa_heatmap.write(stacks, 'roa_mask_neu');

% configure the heatmap color range manually:
% (supply one or more stacks)
begonia.stack_modules.roa_heatmap.configure(stacks, 'roa_mask');

% plot the results:
begonia.stack_modules.roa_heatmap.plot(stack);
```


## RoI Pixel Activity (RPA)

The RPA module will read pixel activity in a roi based on the RoA mask. It requires a roi array and a roa mask to be present.

WARNING : As RoAs arent generally used for neurons at this point, this function assumes rois are from the RoA channel. This should be fixed... 

```Matlab
% read to variable (takes a long time):
rpa = begonia.stack_modules.rpa.read(stack)

% directly save to data location:
begonia.stack_modules.rpa.write(stack)

% plot the results:
mark_above = 30;
shade_non_zero_dominant = true; % mode != 0 for trace
begonia.stack_modules.rpa.plot_traces(stack, mark_above, shade_non_zero_dominant);
```


# Xylobium

The *+xylobium* namespace contains GUI tools that use *+begonia* and *+yucca*.

## Editor for DataLocations and DataLocationAdapters

![Alt text](docs/screenshots/xylo_editor_example.png "Title")

Better documentation will be written in the summer. Quick setup of an editor:
```Matlab
% dummy data locations:
path1 = tempname(); mkdir(path1);
path2 = tempname(); mkdir(path2);
dloc1 = begonia.data_management.DataLocation(path1);
dloc2 = begonia.data_management.DataLocation(path2);

% an action. They recieve dloc, model and editor.
% second action uses model to save a value. This will give
% safe on-the-fly updates in the editor and uses the normal
% save_var under the hood:
ac1 = xylobium.dledit.Action('Show path' ...
    , @(d, m, e) disp (d.path) ...
    , false ...     % not queueable
    , true);        % can take multiple dlocs

ac2 = xylobium.dledit.Action('Set demo to 42' ...
    , @(d, m, e) m.save(d, 'demo', 42) ...
    , true ...      % queueable
    , false);       % one task in queue pr. dloc

% create editor with editable notes field:
xylobium.dledit.Editor([dloc1, dloc2] ...
    , [ac1 ac2] ...
    , {'path', 'demo', '!notes'})
```


## DataLocationAdapter

Mixing in this class with your own allows your class to be read in the Xylobium data editor:

**person.m**
```Matlab
classdef Person < begonia.data_management.DataLocationAdapter    
    properties
        name
        nationality
    end
    
    methods
        function obj = Person(name, nationality)
            obj.name = name,
            obj.nationality = nationality;
        end
    end
end
```

**test.m**

```Matlab
%% Non dloc objects:
a = Person('Knut', 'Norvegia');
b = Person('Teresa', 'Bacalhau');

a.save_var('name', 'Canute');   % saves to the object's name property
b.load_var('nationality');   % gives 'Bacalhau'
a.has_var('pets');      % gives false
b.saved_vars;   % gives 'name' and 'nationality' 

% saving to a non-property will not work:
b.save_var('pet', 'Mr. Cuddles the Cat');    % throws exception 

% open with editor, as if it was a data location:
xylobium.dledit.Editor([a, b]
```

## Stack Processor

The stack processor takes raw prairie (and in the future Oslo Scope) tseries, and applies motion correction, RoA extraction and proxy stack generation without altering the raw data. It's based on the DataLocation editor. This is the **inteded way** to generate motion corrected proxy stacks

It can apply these operations **in batches**, allowing the processor to work e.g. over night, or in the background for extended periods of time.

![Alt text](docs/screenshots/stackproc.PNG "Title")

### First time setup

This step is only necessary if you want to use a setup sheet to select the tseries, and ocnfigure an output directory for aligned series and proxy stacks.  It's an easy way to use the processor, but you can also launch it with arguments from the Matlab command line and scripts (se code example in next section):

1. Copy (dont edit) the config file from +xylobium.+stackproc called "Processing list template.xls" to a location of your choosing.  
2. Add the path to the recorded tseries you want to process to the first sheet
3. Set the output path to proxy stacks in the second sheet

The configuration key for proxy stack can have these values:
* Path to a folder (e.g. C:\Data\Proxystacks) : will output all proxy stacks to this path
* !PARENT : will output to same folder as parent stack

![Alt text](docs/screenshots/proclist.PNG "Title")

### Using

To start the processor, you can either supply a processing-config sheet, a path to scan or a series of stacks.

```Matlab
% launch the processor with a prompt for processing list:
xylobium.stackproc.StackProcessor();

% OR launch processor with a specific config gile
xylobium.stackproc.StackProcessor('/path/to/list.xslx');

% Or, you can launch it with variables:
config = xylobium.stackproc.StackProcessorConfig();
config.alignment_dir = '_SAME_';    % in parent folder (default)
config.alignment_dir = 'C:/AlignedStacks';    % specific folder
config.proxy_output_dir = '_SAME_';    % output to parent folder  (default)
config.proxy_output_dir = 'C:/ProxyStacks';    % output to specific 

% launch with a dir:
xylobium.stackproc.StackProcessor('/path/to/data', config);

% OR launch with stacks:
xylobium.stackproc.StackProcessor(array_of_stacks, config));

```
Suggested workflow (other orders might fail):

1. Select stacks to process from data grid
2. Config phase offset correction (if needed)
3. Config normcorre
4. *Enqueue* Alignment (press "Q" next to button)
5. *Enqueue* RoA extraction
6. *Enqueue* Making proxy stack
7. *Enqueue* Sync stack vars 
8. Press *Do Actions* in Queue window

Here is a screenshot of this process in action:

![Alt text](docs/screenshots/stackproc_processing.png "Title")


### Additional screenshots

Vieving selected matrix:
![Alt text](docs/screenshots/stackproc_phase_matview.PNG "Title")

Phase correction utility:
![Alt text](docs/screenshots/stackproc_phase_correct.PNG "Title")

Configuring alignment channel and type in normcorre:
![Alt text](docs/screenshots/stackproc_phase_normcorre.PNG "Title")

Inspecting selected stack as a variable in matlan (press "-> workspace")
![Alt text](docs/screenshots/stackproc_to_workspace.PNG "Title")

## Vessel tool

### Basic operation

* Select a tseries
* Open config tool and configure analysis for that tseries, marking interesting vessels
* Run analysis
* View results
* Generate video

The tool uses shortcuts to add and removed vessel markers. Each vessel marker gets analysed separately for the green and red channel. Each vessel added gets a randomly generated name.

*Shift-a* adds artery

*Shift-v* adds vein

*Shift-c* adds a capilary

*Shift-s* saves the lines (happens automatically on close)

*Shift-n* toggles visibility of names

*Backspace* removed the last changed line


### Launching:

```Matlab
% use your weapong of choise to get tseries:
tss_info = dcat.data_types('TSeries Timed Element')
tss = dcat.get_data(tss);

% or using scans:
tss = begonia.scantype.find_tseries(uigetdir);

% load the vessel tool:
video_out_dir = uigetdir();
xylobium.vesseltool2.VesselTool(tss, video_out_dir);
```


### Programatically processing marked tseries

Once marked, tseries gets a **vestool2_markers** variable saved. TSeries already analysed get a **vestool2_result** array of objects that contains the data from the analysis.

```Matlab
import xylobium.vesseltool2.*;

% analyse a array of tseries:
analyse(tss);

% export all data to tables:
export_as_table(tss);
```

### Methods

The tool uses two separate methods: intercept-valley and intercept-hill. 



# Yucca

The *+yucca* namespace complements *+begonia* by providing access to  Letten Center recording rig outputs and synchronizing this data with stacks from +begonia.

## Reading trial and experiment data

```Matlab
% find tseries i in a folder:
stacks = begonia.scantypes.find_scans('./data');

% find trials in same folder:
trials = yucca.trial_search.find_trials('./data');
```

## Merging trials after crashes

In the event that the recording ring crashes during an experiment, and trial from the same experiment ends up in two experiment folders, you can try merging the directories using this experimental fix:

```Matlab
% command takes three directories, the last a new directory:
yucca.trial.merge_experiments(first_exp, second_exp, merged_exp)
```

This will merge the logs, and copy the files found in the experiment directories. The format is not guaranteed to be 100% compatible, but trial will load.


## DataCat

Provides a data cataloging system to keep track of data accross multiple disks, and also a way to look up data from notes. It is a begonia.data_management.uuid_resolver, meaning it can look up raw data from proxy stacks.

In this example, we assume we have some data on the local disk (might be a series of proxy disks and trials), and some raw data on two external disks. The catalog will remember the second external disk, even if it has been disconnected from the computer:

```Matlab
% create a catalog
dcat = yucca.datacat.DataCat();

% connect disks to catalog - assume they are pluged in:
dcat.connect('C:\Users\Knut\Local data', 'My proxy stazzzes');
a = dcat.connect('D:\', 'Data disk 1 - first batch');
b = dcat.connect('E:\', 'Data disk 2 - second mouse');

% will no give any entries as sources are not indexed!
dcat.data_available   

% find data on disks (this will take a while IF the disks are not )
dcat.sources(1).rescan();
dcat.sources(2).rescan();   % can also do a.rescan();
dcat.sources(3).rescan();   % can also do b.rescan();

% (note that you can also connect and rescan in one line):
c = dcat.connect('F:\', 'Disk 3 - speedy gonzales').rescan();

 % now lists DataInfo for all data known to cat
dcat.data_available  

% imagine we now unplug disk 2
data.data_available     % data from disk 2 will not show
data.data_unavailable   % data from disk 2
```
The 'cat' DataCat object can then be saved to a small .mat file and kept for future use, or used to look up data - including data on disks that are not currently avilable.

#### Reading data back from the catalogue

```Matlab
% get a data info object for a stack wioth ID 
info = dcat.get_info('30fe4d76-f608-4bb2-9b4a-ae0a6bca7bdb')

% if you want the stack, use this command instead to re-create it (time-consuming):
stack = dcat.get_data('30fe4d76-f608-4bb2-9b4a-ae0a6bca7bdb')

% you can also use one or more Entry or begonia.data_management.DataInfo object directly:
info = dcat.get_info('30fe4d76-f608-4bb2-9b4a-ae0a6bca7bdb')
stack = dcat.get_data(info);
stack = dcat.get_data([dinfo1, dinfo2, dinfo3]);    % three stacks:

% you can also get data directly from an entry:
info.get_data();
dcat.data(1).get_data();

% look up a piece of data by elements you know are in it's name
% (could be more than one...)
stacks = dcat.get_by_fuzzyname('LineScan', '01122018', '29')

% datacat caches lookups by default:
tic; dcat.get_data(ts_info); toc; % Elapsed time is 3.910596 seconds.
tic; dcat.get_data(ts_info); toc; % Elapsed time is 0.002614 seconds.

% you can turn off cache (will not clear it however):
dcat.cache_mode = 'off'
dcat.clear_cache();
```



#### Clustering data (e.g. protocol runs)

It's often important to see data in context, such as several types or recordings that came out of the same trial. For this, DataCat offers RelatedSets. With these objects you can assign data into sets, and assign a role for the data in that set:

In this example, we imagine that the first two tseries int he catalogue were recorded as part of the same trial, but at different wavelengths:

```Matlab
tseries = dcat.data_types('TSeries Timed Element')
ts_990nm = tseries(1); 
ts_890nm = tseries(2);

trial = yucca.datacat.RelatedSet('Trial #1', 'Double wavelength recording', dcat)
trial.add(ts_990nm, '990nm tseries');   % assigns a role in the relationship
trial.add(ts_890nm, '890nm tseries');   % assigns a role in the relationship

dcat.add_related_set(trial);

% when added, this provides automatic reverse lookup abilities:
relset = tseries(1).get_related_sets();

relset.participants % table for participants
relset.get_info_by_role('990nm tseries');
```

#### Resolving original data for proxy stacks

DataCat is a resolver than can be used to hotswap matricies from proxy stack:
```Matlab
% open a proxy stack:
proxy = begonia.scantype.find_scan(path_to_proxystack);

% get proxy frame nr. 500, then the raw one:
frame_proxy = proxy.get_mat(:,:,500);   % proxy
frame_raw = proxy.get_mat(:,:,500, dcat); % original
```

When a source is added or reindexed, the entries in that source is always correalted in time to all other data in the catalogue. To see how things overlap, use the following example. For some projects that only used tseries <-> trial, this could be sufficient to correlate all data in the experiment.
```Matlab
% : lists all other entries that overlap in this with entry 65:
dcat.data(65).overlaping_in_time
```

#### Listing data by type
```Matlab
% returns a map containing the type names as keys, and entries as values:
types = dcat.data_types;

% lists all types in catalogue
types.keys 

% get info on all trials:
trial_infos = types('Recording rig output');

% or, in a single line:
trial_infos = dcat.data_types('Recording rig output');
trials = dcat.get_data(trial_infos);    % all trial objects
```

#### Exporting and importing sets of variables

DataCat can export variables from DataLocation (or DataLocationAdapter) objects. It can be done for an individual source, or the entire catalogue.

```Matlab
% exports two variables (and metadata):
dcat.export_vars(["img_f0", "roi_array"])

% individual source can also be exported separately:
dcat.sources(2).export_vars(["roa_array"]);

% you can also export form a specific seelction of entries:
entries = dcat.data_types('TSeries Timed Element');
dcat.export_vars(["img_f0"], entries);
```
#### Copying (specific data)

DataCat can assist selectively copying data. Let's say you want to copy only Linescans and ZStack from a catalogue to a folder for export. PLEASE NOTE: Matlab copyfile function is very slow on HDD -> same HDD copy. Disk to disk is much faster.

```Matlab
% select lienscans and zstacks, then copy them to a USB stick on E:\
entries = [dcat.data_types('Linescan') dcat.data_types('ZSeries')];
entries.copy_to('E:\Transfer');
```


## Connecting trials to imaging stacks

[ **NOTE!** this is an older method - consider DataCat from the previosu section as an alternative ]

Correalating trials and recrig recordings requires a way to determin the correlation.  The default is to make a function that looks at when either recording was done. The machines' clocks are often misaligned.  To adjust, provide **offset** and **leeway**.  Yucca sees all time related to the time of the trial.

I this example, we use a lag of -20 seconds, and a leeway of 5 seconds. We make a correlation function, then apply it to he trials and stacks lists.  

```Matlab
cf = @(t, s) yucca.trial.corrfuncs.by_offset(-20000, 5000, t, s);
corrs = yucca.trial.correlate_stacks(trials, stacks, cf, 1)
```

If the final argument is 1 the trials and stacks will get a new property called *.assoicated* that contains the correlated object. In this case you do not need the returned array as the information is already in the objects you supplied. 

```Matlab
% dynamic properties avalable if final parameter is 1:
found_stacks = trials(5).associated_stacks;
found_trials = stacks(5).associated_trials;

% since they are dynamic, this DOES NOT work:
all_trials = stacks.associated_trials;
```


## Working with trials and stacks in a graphical user interface

```Matlab

% open a GUI:
yucca.gui.dview.DataViewer(trials, [], {'Name', 'DateRecorded', 'Path', 'MyVariable'});

% works for stacks too (any child of rig.sup.DataLocation):
yucca.gui.dview.DataViewer(stacks, [], {'Name', 'Path', 'MyVariable'})

% open a GUI with an action:
a1 = yucca.gui.dview.DataViewerAction('Do science!');
a1.Callback = @(data) disp(data);   % any function goes here
yucca.gui.dview.DataViewer(stacks, [a1], {'Name', 'DateRecorded', 'Path', 'MyVariable'});
```

% Adding ! to a variable name makes it editable in table (works for text only):
yucca.gui.dview.DataViewer(trials, [], {'Name', 'DateRecorded', 'Path', '!MyVariable'});

% Adding ? to a variable name makes the editor check if the variable exists rather than
% loading the variable into the table. This is useful if variable contains large data.
yucca.gui.dview.DataViewer(trials, [], {'Name', 'DateRecorded', 'Path', '?MyVariable'});

## Writers

The +yucca.frame_writers namespace ontains multiple "writers".  These allow the output of yucca's functions to be captured in many formats. E.g. you may want to write the average of a tseries as a PNG one day, but keep it as a matrix in-memory another.

Yucca's functions takes a writer as an argument, and the output is sendt to the writer.  If the writer is a *MultiPNGWriter*, it will write each output as a file to disk.  If the writer is a *MemoryWriter*, it will contain a struct with those outputs instead.  

## Tseries tools


### Marking RoAs
```Matlab
% get a lazy reader that marks regions of activity (extracted using begonia)
roas = stack.load_var('roa_array'));
mat_raw = stack.get_mat(1,1);
mat = yucca.roa.visualize.MarkingReader(mat_raw, roas);
begonia.util.matview(mat);  % view result
```
### CombiningReader

For comparing stacks with equal dimentions, you can do a lazy merge of the readers:

```Matlab
% in this case, we'll load both the proxy and the proxy source mat in a single view:
mat = yucca.util.CombiningReader(...
    stack.get_mat(1,1) ...
    , stack.get_mat(1,1, cat)...
    );

% the reader reads the data from the stacks in a lazy manner:
begonia.util.matview(mat)
```
![Alt text](docs/screenshots/Combining_reader.PNG "Title")

### Phase offset correction 

Yucca provides a visual tool to correct phase offset for a tseries before aligning.  You need to provide a series of begonia.stack.StackPrairie objects.  The offset parameter is stored as  a variable in the tseries using DataLocation.  The key is *RGOffsetCorrection*.

% get a teseries (or more - you can also make a list)
tseries = begonia.stack.StackPrarie(uigetdir)

% launch GUI tool to visually correct offset:
yucca.gui.phasecorrec.PhaseCorrector([tseries])
```

### Averaging

Write every 1000th image from channel 1 / cycle 1 from stack averaged by running window of 30 frames:

```Matlab
outputdir = 'C:\Users\Knut\Desktop\output_week_5'
ch = 1;
cy = 1;

% first output as individual PNGs at 16bits (the last flag)
writer = yucca.frame_writers.MultiPNGWriter(outputdir, 1)
yucca.tseries.running_average_channel(stack, writer, ch, cy, 30, 1000)

% you can switch output format by changing writer:
mp4_writer = yucca.frame_writers.MP4Writer(outputdir)
yucca.tseries.running_average_channel(stack, mp4_writer, ch, cy, 30, 1000)
```

Write the average of the entire channel as a (**single**) PNG:

```Matlab
png_writer = yucca.frame_writers.SinglePNGWriter(outputdir, 1)
yucca.tseries.channel_average(stack, png_writer, ch, cy);
```

Write the max values of the entire channel as a (single) PNG:

```Matlab
png_writer = yucca.frame_writers.SinglePNGWriter(outputdir, 1)
yucca.tseries.channel_max(stack, png_writer, ch, cy);
```

## Reading data from specific modules

Code to read output of modules are located in *yucca.mod.XXX*.  They generally contains a **read()** function that extracts the data to and object, and a **write_vars()** method that writes the extracted data as variables using *DataLocation*.

In addition, the various *yucca.mod.XXX* namespaces contains module-specific functions to plot the datatypes, manipulate them etc. 

### Mouse

```Matlab
% read mouse information:
mdata = yucca.mod.mouse.read(trial);

% transfer mouse information from log to data location
yucca.mod.mouse.write_vars(trial)
trial.saved_vars;
```

### Air puff

TBD

### Wheel

The wheel module returns a tscollection with two timeseries: DeltaAngle and Distance.  The following example show how to grab the data.

```Matlab
% read mouse information + the first tick (for synchronization purposes)
[wheel_data, first_tick] = yucca.mod.wheel.read(trial)

% plot the data:
figure();
yucca.mod.wheel.plot_da_dist(wheel_data);
```

### EEG

TBD

### Camera

Camera module has no *write_vars()* function.

```Matlab
% get a video reader for a trial:
reader = yucca.mod.cam.read(trial);

% get a single frame for a specific time:
t = datetime('22-Jan-2018 08:22:13.571');
frame = yucca.mod.cam.frame_by_abstime(trial, t);
```

### Pupil tracking data 

See separate ptrack library for parsing this data.

Loading the data:
```Matlab
% get a video reader for a trial. Has allot of time info:
reader = yucca.mod.ptrack.PupilVideoReader(path);

% reading the data:
frame = reader.read_frame(1);
frame = reader.read_frame_at_second(20);
```

#### Handling large raw data

The raw data output from the pupil tracker can be several gigabytes, depending on the recording rate due to high quality capture. This data can be archived out of the trial directories if needed, using the following function. 

This will make the trial directories more mobile. The moved directory gets a metadata package to enable finding it's source trial easily.

```Matlab
% move raw images to an archive location:
yucca.mod.ptracker.move_raw_data(trial, archive_path)
```

## HTML outputs

You can generate a HTML file from a table for easy in-browser viewing like this:

```Matlab
% get HTML from a table
html = yucca.html.table_to_html(tab);
```

If you want to wrap the table in a template, or customize how the cells values are written, you can add renderers and a template.

Templates attempt to read from the begonia "html_templates" directory if the path is not found.

```Matlab
% function to create an image tag from the image value:
img_func = @(v) "<img src='"  + v + ".png' />";

% add renderers. You dont need these, but can be used to modify output
% in this case we cant to create image tags from the given strings:
renderers = containers.Map;
renderers('ts_img_avg') = img_func;
renderers('ts_img_std') = img_func;
renderers('ts_img_pup') = img_func;

% create a HTML table from the matlab table:
html = yucca.html.table_to_html(tab, renderers, 'table_template.html');
```

The output is a HTML formated string with headings.
```HTML
<table><thead><tr><th>index</th><th>ts_name</th><th>...
```

# Unstable

These are experimental code sets that will change name, but are included for experimentation. DO NOT BECOME DEPENDENT on these.

Use these to provide feedback on new APIs.

## RoA API 2.0

```Matlab
% import roa namespace (reduces issue if namespace changes name):
import begonia.analysis.roa.*;

% extract RoAs from a tseries with baseline at frame 1-70:
roas_astrocytes = extract_roa(ts,1,1, 1:70);
roas_neurons = extract_roa(ts,1,2, 1:70);

% save is no longer integrated:
ts.save_var('roas_astrocytes');
ts.save_var('roas_neurons');

% export table for analysis:
tab = make_roa_table([roas_astrocytes,roas_neurons]);

% plot overview of the RoAs:
plot_roa_overview(roas_astrocytes);

% Visualize RoIs on tseries:
plot_roa_activity_interactive(ts, roas_astrocytes);
plot_roa_activity_interactive(ts, roas_neurons) 
% picks channel automatically
```

## RoI API 2.0

This API is connvenient but inefficient. During testing signal extraction from a single tseries of 9000 frames takes 160 seconds. The point of this API is to collect the calendula and alyssium code into a new wrapper/integration for easy access.

```Matlab
% import roi namespace (reduces issue if namespace changes name):
import begonia.analysis.roi.*;

% extract calsium signals from a tseries marked in roimanager.
% The resulting object contains channel trace, drift correction,
% raw signal and dff signal. Note that this function overwrites 
% previously extracted signals in the DataLocation:
signals = extract_roi_signals(ts);

% example plots:
figure();
subplot(2,2,1); plot(rois.dff_signals);
subplot(2,2,2); plot(rois.channel_trace);
subplot(2,2,3); plot(rois.raw_signals);
subplot(2,2,4); plot(rois.drift_correction); title('Drift corr.');

```

## 3D matrix spin view

Useful to visualize 3D matricies, like RoA mask bases etc. before tresholding. Looks neat too.

```Matlab
import begonia.analysis.tseries.*;

% arguments: mat, title, time to spin, fps, degrees, delta time:
plot_3dspin(mat, stack.name, 5, 60, 45, stack.dt);

% optionally add an argument to save as MP4 video:
plot_3dspin(mat, stack.name, 5, 60, 45, stack.dt, '3D Animation.mp4');
```
