classdef RoI < handle
    
    properties
        % Needs to be set manually.
        group           % (char)    E.g. 'ND' for neuron dendrite, 'AS' for astrocyte soma. 
        channel         % (int)     Channel it is connected to.
        cycle           % (int)     Cycle it is connected to.
        shape           % (char)    Eg. 'polygon', 'circle', 'autodetect'
        
        id_connected_roi    % (char) The RoI.id of another roi.
        
        % Is set in the constructor and should never be empty.
        id              % (char) Unique identifier.
        id_shared       % (char) ID that can be shared between multiple rois
                        %        to identify them as the same. 
        area            % (int) 
        center          % (double 1x2)
        x               % (int 1xN)
        y               % (int 1xN)
        mask            % (logical NxM)
        boundary        % (cell array)
        
        version = '2018-04'
    end
    
    methods
        function obj = RoI(mask_or_struct)
            % RoI object.
            %
            %   roi = calendula.RoI(mask)
            %   roi = calendula.RoI(struct)
            % 
            %   REQUIRED
            %   mask        - (logical mat)
            %                   Creates a new roi based on the positive
            %                   values in mask.
            %   struct      - (struct(s) of a calendula.RoI or calendula.RoI)
            %                   A valid struct is created by RoI.to_struct().
            %                   The output is an exact copy of the rois in
            %                   the struct/objects, just instantiated as seperate 
            %                   roi objects.
            %
            %   EXAMPLE 1
            %   roi = calendula.RoI(mask);
            %
            %   EXAMPLE 2
            %   % A list of RoI objects created with masks. 
            %   roi_array = ...
            %   roi_structs = roi_array.to_struct();
            %   % Make the structs back into RoIs
            %   roi_array = calendula.RoI(roi_structs);
            if nargin == 0
                return
            end
            if isa(mask_or_struct,'calendula.RoI') 
                mask_or_struct = mask_or_struct.to_struct();
            end
            switch class(mask_or_struct)
                case 'struct'
                    obj = calendula.RoI.empty;
                    for i = 1:length(mask_or_struct)
                        s = mask_or_struct(i);
                        obj(i).group = s.group;
                        obj(i).channel = s.channel;
                        obj(i).cycle = s.cycle;
                        obj(i).shape = s.shape;
                        obj(i).id_connected_roi = s.id_connected_roi;
                        obj(i).id = s.id;
                        obj(i).area = s.area;
                        obj(i).center = s.center;
                        obj(i).x = s.x;
                        obj(i).y = s.y;
                        obj(i).mask = s.mask;
                        obj(i).boundary = s.boundary;
                        obj(i).version = s.version;
                        if isfield(s,'id_shared') && ~isempty(s.id_shared)
                            obj(i).id_shared = s.id_shared;
                        else
                            obj(i).id_shared = s.id;
                        end
                    end
                otherwise
                    obj.set_mask(mask_or_struct);
                    obj.id = begonia.util.make_uuid();
                    obj.id_shared = obj.id;
            end
        end
        
        
        function create_new_ids(rois)
            % Give all the rois new ids. 
            for roi = rois
                rois.replace_id(roi.id,begonia.util.make_uuid());
            end
        end
        
        
        function replace_id(rois,id_old,id_new)
            % Replaces RoI.id and RoI.id_connected_roi from id_old with id_new.
            
            old = strcmp({rois.id},id_old);
            [rois(old).id] = deal(id_new);
            
            old = strcmp({rois.id_connected_roi},id_old);
            [rois(old).id_connected_roi] = deal(id_new);
            
            old = strcmp({rois.id_shared},id_old);
            [rois(old).id_shared] = deal(id_new);
        end
        
        
        function structs = to_struct(obj)
            if isempty(obj)
                structs = struct([]);
            end
            warning('off','all')
            for i = 1:length(obj)
                structs(i) = struct(obj(i));
            end
            warning('on','all')
        end
        
        function new_roi = merge(rois)
            % Create a new roi with the union of the masks of the input
            % rois.
            mask = rois(1).mask;
            for i = 2:length(rois)
                mask = mask | rois(i).mask;
            end
            
            new_roi = calendula.RoI(mask);
            new_roi.channel = rois(1).channel;
            new_roi.cycle = rois(1).cycle;
            new_roi.shape = rois(1).shape;
            new_roi.group = rois(1).group;
        end
        
        function set_mask(obj, mask)
            assert(length(size(mask)) == 2);
            mask = logical(mask);
            obj.mask = mask;
            [obj.y,obj.x] = find(mask);
            obj.boundary = bwboundaries(mask);
            cx = mean(obj.x);
            cy = mean(obj.y);
            obj.center = [cx, cy];
            obj.area = length(obj.y);
        end
        
        function move(rois, shift)
            shift = round(shift);
            for roi = rois
                mask = circshift(roi.mask, [shift(2), shift(1)]);
                roi.set_mask(mask);
            end
        end
        
        function bool = is_in_roi(roi, x, y)
            %ISINROI check if the point (x,y) is a part of the roi.
            %   val = isInRoi(roi,x,y,)
            %
            % roi       - Single RoI object.
            % x         - (int) Position in image as pixels.
            % y         - (int) Position in image as pixels.
            %
            % See also: RoI

            v1 = uint32([x, y]);
            v2 = [roi.x, roi.y];
            bool = ismember(v1,v2,'rows');
        end
        
        function vec = extract_signal(roi, mat)
            frames = size(mat,3);
            minX = min(roi.x);
            minY = min(roi.y);
            maxX = max(roi.x);
            maxY = max(roi.y);
            
            minX = max(minX,1);
            minY = max(minY,1);
            maxX = min(maxX,size(mat,2));
            maxY = min(maxY,size(mat,1));
            imChunk = mat(minY:maxY, minX:maxX, :);
            roiMask = repmat(roi.mask(minY:maxY, minX:maxX), 1, 1, frames);
            imChunk(~roiMask) = 0;
            vec = squeeze(sum(sum(imChunk, 1), 2)) / length(roi.x);
        end
        
        
        function signal = get_signal(rois,stack)
            % Assume all cycles have equally many frames.
            frames = stack.frames_in_cycle(1);
            signal = NaN(frames, length(rois));
            
            cnt = 1;
            for ch = 1:stack.channels
                for cy = 1:stack.cycles
                    % Get the rois of the correct cycle and channel.
                    idx_cycle = [rois.cycle] == cy;
                    idx_channel = [rois.channel] == ch;
                    roi_indices = find(idx_cycle & idx_channel);
                    
                    if isempty(roi_indices)
                        continue;
                    end
                    
                    % Get the matrix.
                    mat = stack.get_mat(cy,ch);

                    for roi_idx = roi_indices
                        vec = rois(roi_idx).extract_signal(mat);
                        signal(:,roi_idx) = vec;
                    end
                end
            end
            
            signal_t = 0:(frames-1);
            signal_t = signal_t * stack.dt;
            signal_t = signal_t';
            signal = timeseries(signal,signal_t);
        end
    end
end

