classdef roimanager_lite < handle
    %A lite version of the roimanager GUI.
    %   ROIMANAGER_LITE opens a GUI for drawing rois
    %
    %   Usage:
    %       - Load a tiff stack
    %       - Use draw button (or press 'd' on the keyboard) 
    %         to activate a polygon selection tool for drawing rois in the
    %         image. Press the finish button (or 'f' on the keyboard) to complete the roi.
    %       - Or: Use Autodetect rois ('a') to point and click to select rois.
    %
    %   Extra features:
    %       - Autodetect rois by point and click.
    %       - RoIs can be moved around by dragging them in the image using
    %         the mouse
    %       - Keyboard zoom shortcuts ('z' - zoom in, 'Z' - zoom out)
    %       - Keyboard pan shortcuts (use arrow keys)
    %       - Other keyboard shortcuts:
    %           'e' - toggle edit mode
    %           'q' - toggle zoom in mouse tool
    %           'w' - toggle zoom out mouse tool
    %           'r' - reset zoom (zoom out to show entire image)
    %           's' - go to select mode
    %           'escape' - cancel drawing or editing of roi
    %           'backspace' - remove selected rois
    %       - Play video (speed is not very precise)
    
    properties (Constant)
        
        % Different roi classes, their abbreviations, and color for plots.
        roi_groups_long = {'Neuronal Soma', 'Neuronal Dendrite', 'Neuronal Axon', 'Neuropill','Astrocyte Soma','Astrocyte Endfoot','Astrocyte Process','Gliopill', 'Artery', 'Vein', 'Capillary', 'Unknown glia soma'};
        roi_groups = {'NS', 'ND', 'NA','Np','AS', 'AE', 'AP','Gp','Ar','Ve','Ca', 'Xg'}
        roi_colors = {'Red', [.96 .45 .027], [.96 .65 .027], [.75 .5 0], 'Green', [0.016 .61 .51], [.63 .90 .02], [.067 .48 0], [.24 .09 .66], [.43 .051 .64], [.76, .02 .47], [0.75, 0, 0.75]}
        
    end
    
    properties
        
        % User settings
        
        initPath = '';                  % Path to start browsing from
        
        % Figure and figure settings
        
        fig                             % GUI Figure Window
        margins                         % Side margins within the GUI window
        padding                         % Padding between objects in the GUI window
        
        % Stacks
        stacks                          % Stack objectes found during load.
        stack                           % Currently selected stack
        img_stacks                      % Cell array of lazy 3d matrices, one matrix for each channel.
        img_avg                         % 2d Cell array of 2d images, one for each cycle and channel.
        img_std
        img_max
        
        clim_orig                       % Cell array of clim, one for each cycle and channel.
        clim_avg                        % 
        clim_std
        clim_max 
        
        dloc_engine
        
        % UI panels
        
        panel_left                      % Left sidepanel
        imPanel                         % Image Panel for displaying tiffstack (center)
        panel_right                      % Right sidepanel
        panel_roi_info
        
        % Ax and image for displaying the imagestack
        
        axStackDisplay                  % Axes to display images from the tiffstack
        image_obj                       % Image object for the displayed image
        panFactor = 0.1;                % Determines image shift when using arrow keys
        zoomFactor = 0.25;              % Determines zoom when using zoom keys (z and Z)
        binningSize = 9;                % Bin size used for moving averages or maximums.
        scale_bar_plot_handle
        scale_bar_text_handle
        
        axSignalPlot
        signalPlotHandle
        plotFrameMarker
        
        % Roi handles for creating and displaying rois
        
        RoiTmpPlotHandle                % A plot handle for temporary lines of roi polygon
        RoiTmpEdges                     % A list of impoints for the temporary roi polygon
        RoiTmpPosX                      % X coordinate values of the temporary roi
        RoiTmpPosY                      % Y coordinate values of the temporary roi
        roi_plot_handles                % A list of plot handles for all finished rois
        roi_text_handles                % A list of text handles for all finished rois
        roiOuterDiameter = 10;          % Roi diameter in pixels used for autodetection
        roiInnerDiameter = 5;           % Roi inner diameter in pixels used for autodetection
        roiTemplatePlotHandle           % Plot handle of roitemplate for autodetection
        roiDisplacement = 0;            % Temporary "keeper" of roi displacement if rois are moved
        
        % Data which is loaded into gui
        roi_array = []                       % Array of RoI objects
        roi_tags = {}                        % Cell array of roi names.
        selected_rois = []                   % Numbers of the selected/active rois       
        selected_connected_roi = ' '         % name of currently selected connecter RoI 
        
        % GUI state properties                         
        cycle
        channel
        channelDisplayMode = 'single'
        frame                           % Number of the current frame which is displayed
        playbackspeed = 1
        display_mode = 'Stack'          % eg. 'Avg', 'Max', 'MovingAvg'
        
        % Mouse state and recorded cursor positions
        mouse_mode = 'Select'           % MouseMode ('Draw' or 'Select')
        mouse_mode_prev = 'Select'      % Previous active mousemode
        mouseDown = 0                   % Indicates if mouse button is pressed
        prevMouseClick                  % Axes coordinates when mouse was last clicked.
        prevMousePointAx                % Needed for moving RoIs
        prevMousePointFig               % Needed for panning image display in select mode
        roiTemplateCenter               % Needed when setting the roi diameter
        zoomRectPlotHandle              % Handle for rectangle plot when drag-zooming
        
        editor_display
        
        
        % Buttons
        btn_find_stacks
        btnRunAutoSegmentation          % Run automatic roi detection (Paninski code)
        btn_show_stack                  % Show current frame of tiff stack in image display
        btn_show_avg                    % Show average projection of tiff stack in image display
        btn_show_moving_avg             % Togglebutton for showing moving average projection
        btn_show_moving_std             % Togglebutton for showing moving average projection
        btn_show_moving_max             % Togglebutton for showing moving maximum projection
        btn_show_max                    % Show maximum projection of tiff stack in image display
        btn_show_std
        btn_draw_roi                    % Button used for drawing a new RoI
        btn_draw_roi_circular           % Button used for drawin circular rois.
        btn_auto_detect                 % Button used to activate "autodetection clicktool"
        btn_edit_roi                    % Button used for editing an existing RoI
        btn_show_signal                 % Button used for extracting signal from a roi
        btn_remove_roi                  % Button used for removing selected rois
        btn_load_rois                   % Button used for loading rois from file
        btn_save_rois                   % Button used for exporting rois to file
        btn_export_signal               % Button used for exporting signal to file
        btn_clear_all_rois              % Button used for removing all rois
        btn_show_tags                   % Button used for showing or hiding text/tags of rois
        btn_set_roi_size                % Button to quickly set the size of roi to autodetect
        btn_mark
        btn_mark_remove
        btn_mark_split
        btn_mark_add
        btn_undo
        btn_redo
        btnPlayVideo                    % Button to start/stop video.
        btn2x
        btn4x
        btn8x
        btn_show_single_channel
        toggle_scale_bar
        toggle_show_rois
        
        % Temporary stored actions
        action_display
        
        % Other UIcontrols
        
        popup_stacks
        popup_cycle
        popup_channel
        popup_roi_connected
        popup_colormap
        text_clim_low
        text_clim_high
        text_stack_processing_state
        text_stack_dx
        text_stack_dy
        text_stack_fps
        text_roi_tag
        text_roi_size
        edit_scroll_speed
        edit_roi_radius
        popup_roi_groups                % Popupmenu for selecting roi type/Groupname
        edit_frame                      % Input field to display specified frame in image display
        edit_frame_window               % Input field to set binning size for moving averages
        text_current_frame           % Textstring that shows which frame is displayed
        text_stack_name                 % Textstring with current filename
        listbox_rois                      % A listbox containing the name of all rois
        frameslider                     % A sliderbar for scrolling through frames
        roiSizeSlider                   % A sliderbar to set the roisize for autodetection
        roiSizeSliderContainer          % A container for the roi size sliderbar
        path_prev_load_rois = ''
    end
    
    
    
    methods
        
        
        function obj = roimanager_lite(varargin)
            % roimanager_lite gui for marking rois.
            %
            %   obj = roimanager_lite()
            %   obj = roimanager_lite(stacks)
            %   obj = roimanager_lite(...,NAME,VALUE)
            %
            %   PARAMETERS
            %   stacks          - (begonia.stack.Stack) 
            %                       Array of stacks that will be
            %                       displayed/used when the roimanager
            %                       starts. 
            %   dloc_engine     - (begonia.data_management.DataLocationEngine)
            %                       The storage engine that will be used to
            %                       save rois and signal of stacks that are
            %                       loaded with the roimanager.
            %
            %   RETURNED
            %   obj             - (roimanager_lite)
            %                       Instance of the roimanager_lite gui.
            p = inputParser;
            p.addOptional('stacks',[],...
                @(x) validateattributes(x,{'begonia.scantype.TSeries'},{}));
            p.addParameter('dloc_engine', [], ...
                @(x) validateattributes(x,{'begonia.data_management.DataLocationEngine'},{}));
            p.parse(varargin{:});
            begonia.util.dump_inputParser_vars_to_caller_workspace(p);
            %%
            obj.dloc_engine = dloc_engine;
            %% Needed for update frame. 
            tic
            %% Constructs the GUI window and places all objects within it.


            % Set up figure. Default is to cover the whole screen
            screenSize = get(0, 'Screensize');

            obj.fig = figure(...
                          'Name', 'Roimanager Lite', ...
                          'NumberTitle', 'off', ...
                          'Visible','off', ...
                          'Position', screenSize, ...
                          'WindowScrollWheelFcn', @obj.callback_scroll_wheel, ...
                          'WindowKeyPressFcn', @obj.callback_key_press, ...
                          'WindowButtonUpFcn', @obj.callback_mouse_release);

            % Make figure visible to get the right figure size when setting up.
            obj.fig.Visible = 'On';
            set(obj.fig, 'menubar', 'none');
            pause(0.5)

            % Find aspectratio of figure 
            figsize = get(obj.fig, 'Position');
            aspectRatio = figsize(3)/figsize(4);

            % Specify obj.margins (for figure window) and obj.padding (space between objects)
            obj.margins = 0.05 ./ [aspectRatio, 1];  %sides, top/bottom
            obj.padding = 0.07 ./ [aspectRatio, 1];  %sides, top/bottom

            % Set up UI panel positions
            imPanelSize = [0.9/aspectRatio, 0.9]; % This panel determines the rest.
            freeSpaceX = 1 - imPanelSize(1) - obj.margins(1)*2 - obj.padding(1)*2;
            lSidePanelSize = [freeSpaceX/3*1, 0.9];
            rSidePanelSize = [freeSpaceX/3*2, 0.9];

            lSidePanelPos = [obj.margins(1), obj.margins(2), lSidePanelSize];
            imPanelPos = [obj.margins(1) + lSidePanelSize(1) + obj.padding(1), obj.margins(2), imPanelSize];
            rSidePanelPos = [imPanelPos(1) + imPanelSize(1) + obj.padding(1), obj.margins(2), rSidePanelSize];

            % Create left side UI panel for image controls
            obj.panel_left = uipanel('Title','Image Controls', 'Parent', obj.fig, ...
                              'FontSize', 12, ...
                              'units', 'normalized', 'Position', lSidePanelPos);

            % Create center UI panel for image display             
            obj.imPanel = uipanel('Title','Image Stack', 'Parent', obj.fig, ...
                              'FontSize', 12, ...
                              'units', 'normalized', 'Position', imPanelPos);

            set(obj.imPanel, 'units', 'pixel')
            panelAR = (obj.imPanel.Position(3) / obj.imPanel.Position(4));
            set(obj.imPanel, 'units', 'normalized')

            % Add the ax for image display to the image panel. (Use 0.03 units as margins)
            obj.axStackDisplay = axes('Parent', obj.imPanel, 'xtick',[], 'ytick',[], ...
                                      'Position', [0.03/panelAR, 0.03, 0.94/panelAR, 0.94] );

            obj.axSignalPlot = axes('Parent', obj.imPanel, 'xtick',[], 'ytick',[], ...
                                    'Position', [0.03, 0.01, 0.96, 0.15], 'Visible', 'off', ...
                                    'box', 'on', ...
                                    'ButtonDownFcn', @obj.callback_mouse_press_signal);
            hold(obj.axSignalPlot, 'on')                    

            % Create right side UI panel for roimanager controls
            obj.panel_right = uipanel('Title','RoiManager', 'Parent', obj.fig, ...
                              'FontSize', 12, ...
                              'units', 'normalized', 'Position', rSidePanelPos);   
            
            obj.panel_roi_info = uipanel(...
                'Title','RoI Info', ...
                'Parent', obj.panel_right, ...
                'FontSize', 12, ...
                'units', 'normalized', ...
                'Position', [0,0,0.5,0.20]);  

            % Activate callback function for resizing panels when figure changes size
            set(obj.fig, 'SizeChangedFcn', @obj.figsizeChanged );

            %% Create ui controls of left side panel
            padding = 0.1;

            width = 0.8;
            width_half = width/2 - padding/2;
            width_third = width/3 - padding/2;

            height_btn = 0.04;
            height_popup = 0.01;
            height_text = 0.025;
            height_edit = 0.025;
            height_slider = 0.03;

            x = padding;
            x_half = padding + width_half + padding;
            x_third_1 = padding + width_third + padding;
            x_third_2 = padding + width_third + padding + width_third + padding;

            y = 0.9;
            
            obj.btn_find_stacks = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Find Stacks', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_find_stacks);

            y = y - 0.02;
            obj.popup_stacks = uicontrol(...
                'Style', 'popupmenu', ...
                'Parent', obj.panel_left, ...
                'String', 'Stacks', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_popup], ...
                'Callback', @obj.callback_popup_load_stack);

            y = y - 0.05;
            uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'Cycle', ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [x,y,width_half,height_text]...
                );
            uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'Channel', ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [x_half,y,width_half,height_text]...
                );

            y = y - 0.005;
            obj.popup_cycle = uicontrol(...
                'Style', 'popup', ...
                'Parent', obj.panel_left, ...
                'String', '1', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Callback', @obj.callback_popup_cycle, ...
                'Position', [x,y,width_half,height_popup]);
            obj.popup_channel = uicontrol(...
                'Style', 'popup', ...
                'Parent', obj.panel_left, ...
                'String', '1', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Callback', @obj.callback_popup_channel, ...
                'Position', [x_half,y,width_half,height_popup]);

            y = y - 0.08;
            obj.btn_show_single_channel = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show All Channels', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_single_channel);

            y = y - 0.08;
            obj.btn_show_stack = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show Stack', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_show_stack);                         

            y = y - 0.05;
            obj.btn_show_avg = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show Avg', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_avg);

            y = y - 0.05;
            obj.btn_show_max = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show Max', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_max);
            
            
            y = y - 0.05;
            obj.btn_show_std = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show Std', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_std);

            y = y - 0.04;
            jSlider = com.jidesoft.swing.RangeSlider(0, 255, 0, 255);
            [slider, slider_container] = javacomponent(jSlider);
            slider = handle(slider, 'CallbackProperties');
            set(slider, 'StateChangedCallback', @obj.callback_slider_brightness);
            set(slider_container, 'Parent', obj.panel_left, 'units', 'normalized', 'Position', [x,y,width,height_slider])

            y = y - 0.02;
            obj.text_clim_low = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', '0', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x,y,width_third,height_text]);

            uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'Contrast', ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [x_third_1,y,width_third,height_text]);

            obj.text_clim_high = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', '1', ...
                'HorizontalAlignment', 'right', ...
                'Units', 'normalized', ...
                'Position', [x_third_2,y,width_third,height_text]);

            
            y = y - 0.04;
            h = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'Scroll Speed', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_text]);
            jh = begonia.util.findjobj(h);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)


            y = y - 0.005;
            obj.edit_scroll_speed = uicontrol(...
                'Style', 'edit', ...
                'Parent', obj.panel_left, ...
                'String', '1', ...
                'Units', 'normalized', ...
                'Position', [x_half,y,width_half,height_edit]);

            y = y - 0.04;
            h = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'Go to frame', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_text]);
            jh = begonia.util.findjobj(h);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)


            y = y - 0.005;
            obj.edit_frame = uicontrol(...
                'Style', 'edit', ...
                'Parent', obj.panel_left, ...
                'String', 'N/A', ...
                'Units', 'normalized', ...
                'Position', [x_half,y,width_half,height_edit], ...
                'Callback', {@obj.changeFrame, 'jumptoframe'});

            y = y - 0.08;
            obj.btn_show_moving_avg = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show Moving Average', ...
                'Value', 0, ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_moving_avg);     

            y = y - 0.05;
            obj.btn_show_moving_std = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show Moving Std', ...
                'Value', 0, ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_moving_std); 

            y = y - 0.05;
            obj.btn_show_moving_max = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show Moving Maximum', ...
                'Value', 0, ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_moving_max); 

            y = y - 0.05;
            h = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'Bin Size', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_text]);
            jh = begonia.util.findjobj(h);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)

            y = y - 0.005;
            obj.edit_frame_window = uicontrol(...
                'Style', 'edit', ...
                'Parent', obj.panel_left, ...
                'String', num2str(obj.binningSize), ...
                'Units', 'normalized', ...
                'Position', [x_half,y,width_half,height_edit], ...
                'Callback', @obj.callback_edit_frame_window);
            
            y = y - 0.05;
            obj.toggle_scale_bar = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_left, ...
                'String', 'Show Scale Bar', ...
                'Value', 0, ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_scale); 
            
            y = y - 0.05;
            obj.text_stack_dx = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'dx : N/A', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x,y,width_third,height_btn]);
            jh = begonia.util.findjobj(obj.text_stack_dx);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)
            
            obj.text_stack_dy = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'dy : N/A', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x_third_1,y,width_third,height_btn]);
            jh = begonia.util.findjobj(obj.text_stack_dy);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)
            
            obj.text_stack_fps = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_left, ...
                'String', 'fps : N/A', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x_third_2,y,width_third,height_btn]);
            jh = begonia.util.findjobj(obj.text_stack_fps);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)

            y = y - 0.005;
            obj.popup_colormap = uicontrol(...
                'Style', 'popup', ...
                'Parent', obj.panel_left, ...
                'String', {'Colormap','Viridis','Turbo','Magma','Plasma','Gray'}, ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Callback', @obj.callback_popup_colormap, ...
                'Position', [x,y,width,height_popup]);
            
            
            %% UI controls middle pane

            obj.frameslider = uicontrol('Style', 'slider', 'Visible', 'off', ...
                                        'Min', 1, 'Max', 1, 'Value', 1,...
                                        'units', 'normalized', ...
                                        'Position', [imPanelPos(1), 0.02, imPanelSize(1), 0.02],...
                                        'Callback',  {@obj.changeFrame, 'slider'});

            obj.btnPlayVideo = uicontrol('Style', 'togglebutton', 'Parent', obj.imPanel, ...
                                        'String', 'Play', ...
                                        'units', 'normalized', ...
                                        'HorizontalAlignment', 'right', ...
                                        'Position', [0.1, 0.975, 0.05, 0.02], ...
                                        'Callback', @obj.playVideo );
            obj.btn2x = uicontrol('Style', 'togglebutton', 'Parent', obj.imPanel, ...
                                        'String', '2x', ...
                                        'units', 'normalized', ...
                                        'HorizontalAlignment', 'right', ...
                                        'Position', [0.17, 0.975, 0.05, 0.02], ...
                                        'Callback', @obj.btnPlaybackSpeed );

            obj.btn4x = uicontrol('Style', 'togglebutton', 'Parent', obj.imPanel, ...
                                        'String', '4x', ...
                                        'units', 'normalized', ...
                                        'HorizontalAlignment', 'right', ...
                                        'Position', [0.24, 0.975, 0.05, 0.02], ...
                                        'Callback', @obj.btnPlaybackSpeed );

            obj.btn8x = uicontrol('Style', 'togglebutton', 'Parent', obj.imPanel, ...
                                        'String', '8x', ...
                                        'units', 'normalized', ...
                                        'HorizontalAlignment', 'right', ...
                                        'Position', [0.31, 0.975, 0.05, 0.02], ...
                                        'Callback', @obj.btnPlaybackSpeed );
                                    
            
            obj.text_stack_processing_state = uicontrol('Style', 'text', 'Parent', obj.imPanel, ...
                                        'String', 'Processing state: N/A', ...
                                        'units', 'normalized', ...
                                        'HorizontalAlignment', 'right', ...
                                        'Position', [0.3, 0.9725, 0.10, 0.025]);

            obj.text_stack_name = uicontrol('Style', 'text', 'Parent', obj.imPanel, ...
                                        'String', 'Current file: N/A', ...
                                        'units', 'normalized', ...
                                        'HorizontalAlignment', 'right', ...
                                        'Position', [0.4, 0.9725, 0.36, 0.025]);

            obj.text_current_frame = uicontrol('Style', 'text', 'Parent', obj.imPanel, ...
                                        'String', 'Current frame: N/A', ...
                                        'units', 'normalized', ...
                                        'HorizontalAlignment', 'right', ...
                                        'Position', [0.78, 0.9725, 0.185, 0.025]);

            %% Create ui controls of right side panel

            obj.listbox_rois = uicontrol('Style', 'listbox', 'Parent', obj.panel_right, ...
                                        'Min', 0, 'Max', 2, ...
                                        'FontSize', 12, ...
                                        'units', 'normalized', ...
                                        'Position', [0.05, 0.25, 0.4, 0.70], ...
                                        'Callback', @obj.selectListBoxObj);

            
            btn_spacing = 0.02;
            padding = 0.05;

            width = 0.4;
            width_half = width/2 - padding/2;

            height_btn = 0.02;
            height_popup = 0.1;
            height_slider = 0.03;

            x = 0.55;
            x_half = x + width_half + padding;

            y = 0.92;
            h = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_right, ...
                'String', 'RoI Group', ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn]);
            jh = begonia.util.findjobj(h);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)
            
            y = y - 0.10;
            obj.popup_roi_groups = uicontrol(...
                'Style', 'popupmenu', ...
                'Parent', obj.panel_right, ...
                'String', obj.roi_groups_long, ...
                'Value', 1, ...
                'units', 'normalized', ...
                'Position', [x,y,width,height_popup], ...
                'Callback', @obj.callback_popup_roi_group);
            
            y = y + 0.04;
            h = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_right, ...
                'String', 'Connected RoI', ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn]);
            jh = begonia.util.findjobj(h);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)
            
            y = y - 0.10;
            obj.popup_roi_connected = uicontrol(...
                'Style', 'popupmenu', ...
                'Parent', obj.panel_right, ...
                'String', {' '}, ...
                'Value', 1, ...
                'units', 'normalized', ...
                'Position', [x,y,width,height_popup], ...
                'Callback', @(~,~) obj.callback_connected_roi_changed() );

            y = y + 0.04;
            obj.btn_draw_roi = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Draw Rois', ...
                'Value', 0, ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_draw_roi);

            y = y - btn_spacing;
            obj.btn_draw_roi_circular = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Draw Circular Rois', ...
                'Value', 0, ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_draw_roi_circular);
            
            y = y - btn_spacing;
            h = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_right, ...
                'String', 'Radius (um)', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x,y,width_half,height_btn]);
            jh = begonia.util.findjobj(h);
            jh.setVerticalAlignment(javax.swing.JLabel.CENTER)
            
            obj.edit_roi_radius = uicontrol(...
                'Style', 'edit', ...
                'Parent', obj.panel_right, ...
                'String', '10', ...
                'Units', 'normalized', ...
                'Position', [x_half,y,width_half,height_btn]);

            y = y - btn_spacing;
            obj.btn_auto_detect = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Autodetect Rois', ...
                'Value', 0, ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_auto_detect_roi);

            y = y - btn_spacing;
            obj.btn_edit_roi = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Edit Rois', ...
                'Value', 0,...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_edit_roi);

            y = y - btn_spacing;
            obj.btn_remove_roi = uicontrol( ...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Remove Roi', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_remove_roi);

            y = y - btn_spacing;
            obj.btn_show_signal = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Show Signal', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_signal);

            y = y - btn_spacing;
            obj.btn_load_rois = uicontrol(...
                'Style', 'pushbutton',...
                'Parent', obj.panel_right, ...
                'String', 'Load Rois', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_load_rois);

            y = y - btn_spacing;
            obj.btn_save_rois = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Save Rois', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_save_rois);

            y = y - btn_spacing;
            obj.btn_export_signal = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Save Signal', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_export_signal);

            y = y - btn_spacing;
            obj.btn_clear_all_rois  = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Clear All Rois', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_clear_rois);

            y = y - btn_spacing;
            obj.btn_show_tags = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Show Tags', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_show_tags);
            
            y = y - btn_spacing;
            obj.toggle_show_rois = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_right, ...
                'Value', 1, ...
                'String', 'Show RoIs', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_toggle_show_rois);

            y = y - btn_spacing;
            obj.btn_set_roi_size = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Set Roi Size for Autodetection', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_set_roi_size);

            y = y - 0.04;          
            jSlider = com.jidesoft.swing.RangeSlider(0, 30, obj.roiInnerDiameter, obj.roiOuterDiameter);
            [obj.roiSizeSlider, obj.roiSizeSliderContainer] = javacomponent(jSlider);
            obj.roiSizeSlider = handle(obj.roiSizeSlider, 'CallbackProperties');
            set(obj.roiSizeSlider, 'StateChangedCallback', @obj.callback_brush_size);
            set(obj.roiSizeSliderContainer, 'Parent', obj.panel_right, 'units', 'normalized', 'Position', [x,y,width,height_slider])
%             set(obj.roiSizeSliderContainer, 'Visible', 'off')

            y = y - btn_spacing;
            obj.btn_mark = uicontrol(...
                'Style', 'togglebutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Mark', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_mark);
            
            y = y - btn_spacing;
            obj.btn_mark_add = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Add', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_mark_add);
            
            y = y - btn_spacing;
            obj.btn_mark_remove = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Remove', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_mark_remove);
            
            y = y - btn_spacing;
            obj.btn_mark_split = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Split', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_mark_split);
            
            y = y - btn_spacing;
            obj.btn_undo = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Undo', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_undo);
            
            y = y - btn_spacing;
            obj.btn_redo = uicontrol(...
                'Style', 'pushbutton', ...
                'Parent', obj.panel_right, ...
                'String', 'Redo', ...
                'Units', 'normalized', ...
                'Position', [x,y,width,height_btn], ...
                'Callback', @obj.callback_btn_redo);
            
            %% RoI info panel
            spacing = 0.10;
            padding = 0.05;

            width = 1.0 - 2*padding;
            width_half = width/2 - padding/2;
            
            height = 0.15;
            
            x = padding;
            x_half = x + width_half + padding;
            y = 1 - padding - height;
            
            uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_roi_info, ...
                'String', 'Tag : ', ...
                'HorizontalAlignment', 'right', ...
                'Units', 'normalized', ...
                'Position', [x,y,width_half,height]);
            
            obj.text_roi_tag = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_roi_info, ...
                'String', 'N/A', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x_half,y,width_half,height]);
            
            y = y - spacing;
            uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_roi_info, ...
                'String', 'Area (um) : ', ...
                'HorizontalAlignment', 'right', ...
                'Units', 'normalized', ...
                'Position', [x,y,width_half,height]);
            
            
            obj.text_roi_size = uicontrol(...
                'Style', 'text', ...
                'Parent', obj.panel_roi_info, ...
                'String', 'N/A', ...
                'HorizontalAlignment', 'left', ...
                'Units', 'normalized', ...
                'Position', [x_half,y,width_half,height]);

            %% Set some initial gui state varibles
            obj.mouse_mode = 'Select';
            obj.frame = 0;  

            %% Activate callback functions
            set(obj.fig, 'WindowButtonMotionFcn', {@obj.callback_mouse_move})

            set(obj.fig, 'CurrentObject', obj.fig)
            
            %% Load input stacks if not empty.
            obj.load_stacks(stacks);
        
        end
        
        function callback_toggle_show_scale(obj,source,~)
            obj.update_scale_bar(source.Value);
        end
        
        function callback_toggle_show_rois(obj,source,~)
            if isempty(obj.stack); return; end;
            obj.update_roi_visibility();
        end
        
        function callback_popup_colormap(obj,source,~)
            if isempty(obj.stack); return; end
            
            % {'Viridis','Turbo','Magma','Plasma','Gray'}
            switch source.String{source.Value}
                case 'Viridis'
                    colormap(obj.axStackDisplay,begonia.colormaps.viridis);
                case 'Turbo'
                    colormap(obj.axStackDisplay,begonia.colormaps.turbo);
                case 'Magma'
                    colormap(obj.axStackDisplay,begonia.colormaps.magma);
                case 'Plasma'
                    colormap(obj.axStackDisplay,begonia.colormaps.plasma);
                case 'Gray'
                    colormap(obj.axStackDisplay,gray);
            end
                
        end
        
        % called when the index of the connected rois list changes:
        function callback_connected_roi_changed(obj)
            i = obj.popup_roi_connected.Value;
            obj.selected_connected_roi = obj.popup_roi_connected.String{i};
        end
        
        function figsizeChanged(obj, ~, ~)
        % Callback function to resize/move ui panels if figure size is changed
        
            figsize = get(obj.fig, 'Position');
            aspectRatio = figsize(3)/figsize(4);
            
            obj.margins(1) = obj.margins(2) / aspectRatio; 
            obj.padding(1) = obj.padding(2) / aspectRatio;
            
            % Calculate new panel positions
            imPanelSize = [0.9/aspectRatio, 0.9]; % This panel determines the rest.
            freeSpaceX = 1 - imPanelSize(1) - obj.margins(1)*2 - obj.padding(1)*2;
            lSidePanelSize = [freeSpaceX/3*1, 0.9];
        	rSidePanelSize = [freeSpaceX/3*2, 0.9];
                
            lSidePanelPos = [obj.margins(1), obj.margins(2), lSidePanelSize];
            imPanelPos = [obj.margins(1) + lSidePanelSize(1) + obj.padding(1), obj.margins(2), imPanelSize];
            rSidePanelPos = [imPanelPos(1) + imPanelSize(1) + obj.padding(1), obj.margins(2), rSidePanelSize];

            % Reset panel positions
            set(obj.panel_left, 'Position', lSidePanelPos);
            set(obj.imPanel, 'Position', imPanelPos)
            set(obj.panel_right, 'Position', rSidePanelPos);
            
            % Scale frameslider to keep same width as impanel
            fslidePos = get(obj.frameslider, 'Position');
            fslidePos(1) = imPanelPos(1);
            fslidePos(3) = imPanelPos(3);
            set(obj.frameslider, 'Position', fslidePos)
            
            
            if ~isempty(obj.btn_show_signal.Value) && obj.btn_show_signal.Value
                set(obj.axStackDisplay, 'Position', [0.12, 0.12, 0.76, 0.94])
            end
        end
        
        
        function set_display_mode(obj, new_display_mode)
            obj.display_mode = new_display_mode;
            
            % Take care of togglebuttons. Only one active at the same time
            obj.deselect_display_mode_buttons();
            switch obj.display_mode
                case 'Stack'
                    set(obj.frameslider, 'Visible', 'on');
                case 'Avg'
                    obj.btn_show_avg.Value = 1;
                    set(obj.frameslider, 'Visible', 'off');
                    set(obj.text_current_frame, 'String', 'Current Frame: Avg Image')
                case 'Max'
                    obj.btn_show_max.Value = 1;
                    set(obj.frameslider, 'Visible', 'off');
                    set( obj.text_current_frame, 'String', 'Current Frame: Max Image')
                case 'Std'
                    obj.btn_show_std.Value = 1;
                    set(obj.frameslider, 'Visible', 'off');
                    set(obj.text_current_frame, 'String', 'Current Frame: Std Image')
                case 'MovingAvg'
                    obj.btn_show_moving_avg.Value = 1;
                case 'MovingStd'
                    obj.btn_show_moving_std.Value = 1;
                case 'MovingMax'
                    obj.btn_show_moving_max.Value = 1;
            end
            
            % Show updated contrast and update the shown image.
            obj.update_clim_text();
            obj.update_image_display();
        end
        
        
        function set_mouse_mode(obj, new_mouse_mode)
            % Change the mode of mouseclicks in the GUI
            
            % Cancel rois if draw is unselected
            switch new_mouse_mode
                case {'Select', 'Autodetect', 'AutoRoiDiameter'}
                    if obj.btn_draw_roi.Value && ~isempty(obj.RoiTmpEdges)
                        obj = cancelRoi(obj);
                    elseif obj.btn_edit_roi.Value && ~isempty(obj.RoiTmpEdges)
                        obj = cancelRoi(obj);
                    end
                case 'EditSelect'
                    if obj.btn_draw_roi.Value && ~isempty(obj.RoiTmpEdges)
                        obj = cancelRoi(obj);
                    end
                case 'Draw'
                    if obj.btn_edit_roi.Value && ~isempty(obj.RoiTmpEdges)
                        obj = cancelRoi(obj);
                    end
            end
            
            % Set mousemode
            illegal_prev_mouse_modes = {'Zoom In', 'Zoom Out'};
            switch new_mouse_mode
                case 'Previous'
                    obj.mouse_mode = obj.mouse_mode_prev;
                case illegal_prev_mouse_modes
                    obj.mouse_mode = new_mouse_mode;
                otherwise
                    obj.mouse_mode_prev = obj.mouse_mode;
                    obj.mouse_mode = new_mouse_mode;
            end
            
            % Take care of togglebuttons. Only one active at the same time
            switch obj.mouse_mode
                case 'Draw'
                    obj.select_draw_roi_button(obj.btn_draw_roi);
                case 'DrawCircular'
                    obj.select_draw_roi_button(obj.btn_draw_roi_circular);
                case 'Autodetect'
                    obj.select_draw_roi_button(obj.btn_auto_detect);
                case 'EditSelect'
                    obj.select_draw_roi_button(obj.btn_edit_roi);
                case 'Select'
                    % Deselects draw buttons.
                    obj.select_draw_roi_button();
            end
            
            obj.updatePointer();
        
        end
        
        
        function obj = updatePointer(obj)
            % Change pointers according to mousemode

            if obj.insideImageAx()
                mousePoint = get(obj.axStackDisplay, 'CurrentPoint');
                mousePoint = mousePoint(1, 1:2);
                
                switch obj.mouse_mode
                    case 'Draw'
                        xlim = get(obj.axStackDisplay, 'XLim');
                        impoint_extent = diff(xlim)/100;
                        if any(abs(obj.RoiTmpPosX - mousePoint(1)) < impoint_extent)
                            idx = find(abs(obj.RoiTmpPosX - mousePoint(1)) < impoint_extent);
                            if any(abs(obj.RoiTmpPosY(idx) - mousePoint(2)) < impoint_extent)
                                set(obj.fig,'Pointer', 'fleur');
                            else
                                set(obj.fig, 'Pointer', 'crosshair');
                            end
                        else
                           set(obj.fig, 'Pointer', 'crosshair');
                        end
                    case 'DrawCircular'
                        set(obj.fig, 'Pointer', 'cross');
                    case 'Autodetect'
                        set(obj.fig, 'Pointer', 'cross');
                    case 'EditSelect'
                        set(obj.fig, 'Pointer', 'hand');
                    case 'AutoRoiDiameter'
                        set(obj.fig, 'Pointer', 'circle');
                    case 'Select'
                        set(obj.fig,'Pointer','hand');
                    case 'Zoom In'
                        setptr(obj.fig, 'glassplus');
                    case 'Zoom Out'
                        setptr(obj.fig, 'glassminus');
                end
                
            else
                set(obj.fig, 'Pointer', 'arrow');
            end
            
        end
        
        
        function imageZoom(obj, direction)
        % Zoom in image    
            
            switch direction
                case 'in'
                        zoomF = -obj.zoomFactor;
                case 'out'
                        zoomF = obj.zoomFactor;
            end
                    
            xLim = get(obj.axStackDisplay, 'XLim');
            yLim = get(obj.axStackDisplay, 'YLim');

            mp_f = get(obj.fig, 'CurrentPoint');
            mp_a = get(obj.axStackDisplay, 'CurrentPoint');
            mp_a = mp_a(1, 1:2);

            % Find ax position and limits in figure units.
            figsize = get(obj.fig, 'Position');
            panelPos = get(obj.imPanel, 'Position') .* [figsize(3:4), figsize(3:4)];  % pixel units
            axPos = get(obj.axStackDisplay, 'Position') .* [panelPos(3:4), panelPos(3:4)];  % pixel units
            axPos = [panelPos(1:2), 0, 0] + axPos;
            axLim = axPos + [0, 0, axPos(1), axPos(2)];

            % Check if mousepoint is within axes limits.
            insideImageAx = ~any(any(diff([axLim(1:2); mp_f; axLim(3:4)]) < 0));

            xLimNew = xLim + [-1, 1] * zoomF * diff(xLim);
            yLimNew = yLim + [-1, 1] * zoomF * diff(yLim);

            if insideImageAx
                mp_f = mp_f - [axPos(1), axPos(2)];

                shiftX = (axPos(3)-mp_f(1)) / axPos(3)               * diff(xLimNew) - (xLim(1) + diff(xLim)/2 + diff(xLimNew)/2 - mp_a(1)) ;
                shiftY = (axPos(4)-abs(axPos(4)-mp_f(2))) / axPos(4) * diff(yLimNew) - (yLim(1) + diff(yLim)/2 + diff(yLimNew)/2 - mp_a(2)) ;
                xLimNew = xLimNew + shiftX;
                yLimNew = yLimNew + shiftY;
            end
            
            width = size(obj.img_stacks{obj.channel},2);
            if diff(xLimNew) > width
                xLimNew = [0, width];
            elseif xLimNew(1) < 0
                xLimNew = xLimNew - xLimNew(1);
            elseif xLimNew(2) > width
                xLimNew = xLimNew - (xLimNew(2) - width);
            end

            height = size(obj.img_stacks{obj.channel},1);
            if diff(yLimNew) > height
                yLimNew = [0, height];
            elseif yLimNew(1) < 0
                yLimNew = yLimNew - yLimNew(1);
            elseif yLimNew(2) > height
                yLimNew = yLimNew - (yLimNew(2) - height);
            end

            set(obj.axStackDisplay, 'XLim', xLimNew, 'YLim', yLimNew)
        end
        
        
        function imageZoomRect(obj)
        % Zoom in image according to rectangle coordinates.
        
            xData = get(obj.zoomRectPlotHandle, 'XData');
            yData = get(obj.zoomRectPlotHandle, 'YData');
            
            xLimNew = [min(xData), max(xData)];
            yLimNew = [min(yData), max(yData)];
                        
            if diff(xLimNew) > diff(yLimNew)
                yLimNew = yLimNew + [-1, 1] * (diff(xLimNew) - diff(yLimNew)) / 2;
            elseif diff(xLimNew) < diff(yLimNew)
                xLimNew = xLimNew + [-1, 1] * (diff(yLimNew) - diff(xLimNew)) / 2;
            end
            
            set(obj.axStackDisplay, 'XLim', xLimNew, 'YLim', yLimNew)
            
        end
        
        
        function bool = insideImageAx(obj)
            % Check if mousepoint is within axes limits of stack display
        
            currentPoint = get(obj.fig, 'CurrentPoint') ;
            
            % Find ax position and limits in figure units.
            figsize = get(obj.fig, 'Position');
            panelPos = get(obj.imPanel, 'Position') .* [figsize(3:4), figsize(3:4)];  % pixel units
            axPos = get(obj.axStackDisplay, 'Position') .* [panelPos(3:4), panelPos(3:4)];  % pixel units
            axPos = [panelPos(1:2), 0, 0] + axPos;
            axLim = axPos + [0, 0, axPos(1), axPos(2)];

            % Check if mousepoint is within axes limits.
            bool = ~any(any(diff([axLim(1:2); currentPoint; axLim(3:4)]) < 0));
            
        end
            
            
        function obj = callback_key_press(obj, ~, event)
            % Function to handle keyboard shortcuts. 
            
            % Pan and zoom functions are modified from axdrag (https://github.com/gulley/Ax-Drag)
            switch event.Key
                case '1'
                    obj.change_cycle_and_channel(obj.cycle,1);
                case '2'
                    obj.change_cycle_and_channel(obj.cycle,2);
                case '3'
                    obj.change_cycle_and_channel(obj.cycle,3);
                case 'd'
                    switch obj.mouse_mode
                        case 'Draw'
                            obj.set_mouse_mode('Select');
                        otherwise
                            obj.set_mouse_mode('Draw');
                    end
                case 'a'
                    switch obj.mouse_mode
                        case 'Autodetect'
                            obj.set_mouse_mode('Select');
                        otherwise
                            obj.set_mouse_mode('Autodetect');
                    end
                case 's'
                    obj.set_mouse_mode('Select');
                case 'f'
                    obj.finishRoi();
                case 'escape'
                    obj.cancelRoi();
                case 'delete'
                    obj.remove_selected_rois();
                case 'backspace'
                    obj.callback_btn_remove_roi();
                case 'leftarrow'
                    xLim = get(obj.axStackDisplay, 'XLim');
                    xLimNew = xLim - obj.panFactor * diff(xLim);
                    if xLimNew(1) > 0 && xLimNew(2) < obj.stack.original_img_dim(2)
                        set(obj.axStackDisplay, 'XLim', xLimNew);
                    end
                case 'rightarrow'
                    xLim = get(obj.axStackDisplay, 'XLim');
                    xLimNew = xLim + obj.panFactor * diff(xLim);
                    if xLimNew(1) > 0 && xLimNew(2) < obj.stack.original_img_dim(2)
                        set(obj.axStackDisplay, 'XLim', xLimNew);
                    end
                case 'uparrow'
                    yLim = get(obj.axStackDisplay, 'YLim');
                    yLimNew = yLim - obj.panFactor * diff(yLim);
                    if yLimNew(1) > 0 && yLimNew(2) < obj.stack.original_img_dim(1)
                        set(obj.axStackDisplay, 'YLim', yLimNew);
                    end
                case 'downarrow'
                    yLim = get(obj.axStackDisplay, 'YLim');
                    yLimNew = yLim + obj.panFactor * diff(yLim);
                    if yLimNew(1) > 0 && yLimNew(2) < obj.stack.original_img_dim(1)
                        set(obj.axStackDisplay, 'YLim', yLimNew);
                    end
                case {'z', 'Z'}
                    if event.Character == 'z'
                        obj.imageZoom('in');
                    else
                        obj.imageZoom('out');
                    end
                case 'q'
                    switch obj.mouse_mode
                        case 'Zoom In'
                            obj.set_mouse_mode('Previous');
                        otherwise
                            obj.set_mouse_mode('Zoom In');
                    end
                case 'w'
                    switch obj.mouse_mode
                        case 'Zoom Out'
                            obj.set_mouse_mode('Previous');
                        otherwise
                            obj.set_mouse_mode('Zoom Out');
                    end
                case 'r'
                    set(obj.axStackDisplay, 'XLim', [0, obj.imWidth], 'YLim', [0, obj.imHeight])
            end
            
        end
        
        
        function obj = callback_mouse_press(obj, ~, event)
            % Callback function to handle mouse presses on image obj
        
            % Record mouse press and current mouse position
            obj.mouseDown = true;
            
            % Get current mouse position in ax
            x = event.IntersectionPoint(1);
            y = event.IntersectionPoint(2);
            
            obj.prevMousePointAx = [x, y];
            obj.prevMouseClick = [x, y];
            obj.prevMousePointFig = get(obj.fig, 'CurrentPoint');
            
            
            switch obj.mouse_mode 
                case {'Draw', 'EditDraw'}        % Convert mouseclick to roi vertex
                    switch get(obj.fig, 'SelectionType')
                        case 'normal'
                            obj.RoiTmpPosX(end+1) = x;
                            obj.RoiTmpPosY(end+1) = y;
                            obj.addTmpRoiVertex(obj.axStackDisplay, x, y);
                            obj.drawTmpRoi(obj.axStackDisplay);
                        case {'extend', 'alt'}
                            obj.finishRoi();
                    end
                    
                case 'DrawCircular'
                    % Create a circular roi.
                    roi = obj.make_roi_circular(x,y);
                    obj.add_roi(roi);
                    
                case 'Autodetect'
                    newRoi = make_roi_autodetect(obj, x, y);
                    was_in_roi = 0;
                    for i = 1:length(obj.roi_array)
                        if obj.roi_array(i).is_in_roi(x,y)
                            was_in_roi = 1;
                            oldRoi = obj.roi_array(i);
                            break
                        end
                    end
                    
                    if was_in_roi
%                         newRoi = obj.inheritRoiProperties(newRoi, oldRoi);
%                         newRoi.Shape = oldRoi.Shape;
%                         % Replot the roi
%                         obj.roi_array(i) = newRoi;
%                         updateRoiPlot(obj, i, newRoi);
                         
                    elseif ~isempty(newRoi)
                        obj.add_roi(newRoi);
                    end
                    
                case 'AutoRoiDiameter'
                    obj.roiTemplateCenter = [x, y];
                    obj.plotRoiTemplate();
                case 'Select'         
                    
                    % Interrupt old code if this new feature is selected.
                    if strcmp(obj.editor_display.mode,'mark')
                        action = calendula.actions.DisplayAction(obj.editor_display);
                        return
                    end

                    % Change status of roi if it was clicked
                    selected_roi = [];
                    for i = 1:length(obj.roi_array)
                        if obj.roi_array(i).is_in_roi(x, y)
                            selected_roi = i;
                            break
                        end
                    end
                    
                    switch get(obj.fig, 'SelectionType')
                        case 'normal'
                            obj.select_rois(selected_roi);
                        case 'extend'
                            selected_rois = obj.selected_rois; %#ok<*PROPLC>
                            if any(ismember(obj.selected_rois,selected_roi)) ...
                                    && ~obj.mouseDown
                                % Remove the selected roi.
                                selected_rois(selected_rois == selected_roi) = [];
                            else
                                % Add the selected_roi.
                                selected_rois = [selected_rois,selected_roi];
                            end
                            obj.select_rois(selected_rois);
                    end
                case 'Zoom In'
                    axes(obj.axStackDisplay)
                    hold on
                    if isempty(obj.zoomRectPlotHandle)
                        obj.zoomRectPlotHandle = plot(nan, nan);
                    else
                        set(obj.zoomRectPlotHandle, 'XData', nan, 'Ydata', nan)
                    end
                    set(obj.zoomRectPlotHandle, 'Visible', 'on')
                case 'Zoom Out'
                    obj.imageZoom('out');
            end
            
        end
        
        
        function obj = callback_mouse_press_signal(obj, ~, event)
            % Callback function for mousepress within signal plot ax
            x = event.IntersectionPoint(1);
            
            source.String = num2str(round(x));
            obj.changeFrame(source, [], 'jumptoframe');
        end
        
        
        function obj = callback_mouse_release(obj, ~, ~)
            % Callback function when mouse button is released over figure.
            
            
            if obj.insideImageAx() && ~isempty(obj.editor_display)
                action = calendula.actions.DisplayAction(obj.editor_display);
                obj.editor_display.do_and_record_action(action);
            end
            
            obj.mouseDown = false;
            
            switch obj.mouse_mode
                case 'Zoom In'
                    currentPoint = get(obj.axStackDisplay, 'CurrentPoint');
                    currentPoint = currentPoint(1, 1:2);
                    
                    set(obj.zoomRectPlotHandle, 'Visible', 'off')
                    
                    if all((abs(obj.prevMouseClick - currentPoint)) < 1) % No movement
                        obj.imageZoom('in');
                    else
                        obj.imageZoomRect(); % Set new limits based on new and old point
                    end
                    
                    axes(obj.axStackDisplay)
                    hold off
                    
                case 'Select'
                    if obj.roiDisplacement ~= 0
                        obj.roi_array(obj.selected_rois).move(obj.roiDisplacement);
                        obj.roiDisplacement = 0;
                    end
            end   
        end
        
        
        function callback_btn_undo(obj,~,~)
            obj.editor_display.undo();
        end
        
        function callback_btn_redo(obj,~,~)
            obj.editor_display.redo()
        end
        
        
        
        function callback_mouse_move(obj, ~, ~)
            % Callback funtion to handle mouse movement over figure
            if isempty(obj.image_obj); return; end;
            
            % Get mousepoint coordinates in ax and figure units
            newMousePointFig = get(obj.fig, 'CurrentPoint');
            newMousePointAx = get(obj.axStackDisplay, 'CurrentPoint');
            newMousePointAx = newMousePointAx(1, 1:2);
            
            obj.updatePointer();
            
            if obj.insideImageAx() && ~isempty(obj.editor_display)
                action = calendula.actions.DisplayAction(obj.editor_display);
                action.mouse_clicked = obj.mouseDown;
                action.mouse_coordinates = newMousePointAx;
                action.do();
            end
            
            if obj.insideImageAx() && obj.mouseDown   % "Click and Drag"
                set(obj.fig, 'CurrentObject', obj.imPanel)
                
                switch obj.mouse_mode
                    case 'Select'
                        % TODO:
                        % Interrupt old code for new feature.
                        if ~isempty(obj.editor_display) && strcmp(obj.editor_display.mode,'mark')
                            return
                        end

                        % Move roi to new position
                        if ~isempty(obj.selected_rois)
                            shift = newMousePointAx - obj.prevMousePointAx;
                            obj.roiDisplacement = obj.roiDisplacement + shift;
                            obj.shiftRoiPlot([shift, 0]);
                            
                        % Move image to new position (pan). 
                        else
                            % Need to use figure coordinates, because axes coordinates are
                            % continuously changed
                            shift = newMousePointFig - obj.prevMousePointFig;
                            obj.moveImage(shift);
                        end
                        
                        % Reset previous coordinates to current
                        obj.prevMousePointAx = newMousePointAx;
                        obj.prevMousePointFig = newMousePointFig;
                    
                    case 'Zoom In'
                        % Plot a rectangle 
                        x1 = obj.prevMouseClick(1);
                        x2 = newMousePointAx(1);
                        y1 = obj.prevMouseClick(2);
                        y2 = newMousePointAx(2);
                        set(obj.zoomRectPlotHandle, 'XData', [x1, x1, x2, x2, x1], ...
                                                    'YData', [y1, y2, y2, y1, y1])

                end
         
            else % Release mouseDown if mouse moves out of image.
                set(obj.fig, 'CurrentObject', obj.fig)
                if obj.mouseDown
                    obj.mouseDown = false;
                end
            end
        end
        
        
        function obj = moveImage(obj, shift)
            % Move image in ax according to shift
            
            % Get ax position in figure coordinates
            figsize = get(obj.fig, 'Position');
            panelPos = get(obj.imPanel, 'Position') .* [figsize(3:4), figsize(3:4)];  % pixel units
            axPos = get(obj.axStackDisplay, 'Position') .* [panelPos(3:4), panelPos(3:4)];  % pixel units
            axPos = [panelPos(1:2), 0, 0] + axPos;
        
            % Get current axes limits
            xlim = get(obj.axStackDisplay, 'XLim');
            ylim = get(obj.axStackDisplay, 'YLim');
            
            % Convert mouse shift to image shift
            imshift = shift ./ axPos(3:4) .* [diff(xlim), diff(ylim)];
            xlim = xlim - imshift(1);
            ylim = ylim + imshift(2);

            % Dont move outside of image boundaries..
            ch = obj.channel;
            if xlim(1) > 0 && xlim(2) < size(obj.img_stacks{ch},2)
                set(obj.axStackDisplay, 'XLim', xlim);
            end
            if ylim(1) > 0 && ylim(2) < size(obj.img_stacks{ch},1)
                set(obj.axStackDisplay, 'YLim', ylim);
            end 
        end
        
        
        function callback_btn_find_stacks(obj,~,~)
            path = uigetdir('','Find Stacks');
            if path == 0
                return
            end
            stacks = begonia.scantype.find_tseries(path,'show_waitbar',true);
            
            if isempty(stacks)
                disp('No stacks found.')
                return
            end
            obj.load_stacks(stacks);
        end
        
        
        function callback_popup_load_stack(obj, ~, ~)
            obj.load_stack(obj.popup_stacks.Value);
        end
        
        
        function load_stacks(obj,stacks)
            if isempty(stacks); return; end;
            if ~isempty(obj.dloc_engine)
                [stacks.dl_storage_engine] = deal(obj.dloc_engine);
            end
            obj.stacks = stacks;
            obj.popup_stacks.String = {stacks.name};
            obj.popup_stacks.Value = 1;
            obj.load_stack(1);
        end
        
        
        function load_stack(obj, index)
            % Load an image stack into the GUI
            if isempty(obj.stacks)
                disp('No stacks loaded.')
                return
            end
            if index > length(obj.stacks)
                return
            end
            
            obj.stack = obj.stacks(index);
            
            % Set current channel to the first one.
            obj.channel = 1;
            obj.cycle = 1;
            
            % Get img stacks of each cycle and channel.
            h = waitbar(0,'Loading cycles and channels');
            cnt = 1;
            for cy = 1:obj.stack.cycles
                for ch = 1:obj.stack.channels
                    waitbar(cnt/obj.stack.channels/obj.stack.cycles,h,'Loading cycles and channels');
                    img_stacks{cy,ch} = obj.stack.get_mat(cy,ch);
                    cnt = cnt + 1;
                end
            end
            obj.img_stacks = img_stacks;
            close(h);
            
            % (Re)set current frame to first frame
            obj.frame = 1;
            
            
            obj.load_image_projections();
            
            % Update cycle and channel popupmenus. 
            obj.popup_channel.String = obj.stack.channel_names;
            obj.popup_channel.Value = 1;
            
            obj.popup_cycle.String = arrayfun(@num2str,1:obj.stack.cycles,'UniformOutput',false);
            obj.popup_cycle.Value = 1;
                         
            % Set up frame indicators and frameslider
            frames = obj.stack.frames_in_cycle(obj.cycle);
            if frames < 11
                sliderStep = [1/(frames-1), 1/(frames-1)];
            else
                sliderStep = [1/(frames-1), 10/(frames-10)];
            end
            set(obj.text_current_frame, 'String', ['Current Frame: 1/' num2str(frames)] )
            set(obj.text_stack_name, 'String',  obj.stack.name)
            set(obj.edit_frame, 'String', '1')
            set(obj.frameslider, ...
                'Max', frames, ...
                'Value', obj.frame, ...
                'SliderStep', sliderStep, ...
                'Visible', 'on');
            obj.text_stack_processing_state.String = obj.stack.processing_state;
            obj.text_stack_dx.String = sprintf('dx : %05.2f',obj.stack.dx);
            obj.text_stack_dy.String = sprintf('dy : %05.2f',obj.stack.dy);
            obj.text_stack_fps.String = sprintf('fps : %05.2f',1/obj.stack.dt);
            
            % Set limits of signal Plot
            set(obj.axSignalPlot, 'Xlim', [1, frames])
                             
            obj.editor_display = [];
                             
            % Reset some buttons
            set(obj.btn_show_moving_avg, 'Value', 0)
            set(obj.btn_show_moving_std, 'Value', 0) 
            set(obj.btn_show_moving_max, 'Value', 0)
            
            obj.update_clim_text();
                             
            % Display image
            obj.update_image_display();
            
            obj.update_scale_bar();
            
            obj.callback_btn_clear_rois();
            
            % Load saved rois.
            roi_array = obj.stack.load_var('roi_array',calendula.RoI.empty);
            % If the class is unknown it will be loaded as an int. 
            if ~isempty(roi_array) && ~isnumeric(roi_array)
                % Add and plot the rois.
                % Cast the structs back to roi objects.
                roi_array = calendula.RoI(roi_array);
                obj.add_roi(roi_array)
                obj.update_roi_visibility();
            end
        end
        
        
        function update_scale_bar(obj, show)
            % Recreates the scale bare on update, this is not updated very
            % often so it shouldnt be a problem. 
            
            if isempty(obj.stack); return; end;
            
            if ~isempty(obj.scale_bar_plot_handle)
                delete(obj.scale_bar_plot_handle)
            end
            if ~isempty(obj.scale_bar_text_handle)
                delete(obj.scale_bar_text_handle)
            end
            
            if nargin < 2
                show = true;
            end
            
            if show
                obj.toggle_scale_bar.Value = 1;
                % Plot the scale bar. 
                axes(obj.axStackDisplay)

                dim = size(obj.img_stacks{obj.cycle,obj.channel});
                dim_10 = round(dim .* 0.10);
                x1 = dim_10(2);
                x2 = 3*dim_10(2);
                y1 = dim(1) - dim_10(1);
                y2 = dim(1) - dim_10(1);
                hold on
                obj.scale_bar_plot_handle = plot([x1,x2],[y1,y2], '-r', 'LineWidth', 4);
                dx = round(obj.stack.dx*2*dim_10(2));
                obj.scale_bar_text_handle = text(2*dim_10(1), dim(2) - 0.5*dim_10(2), ...
                    [char(num2str(dx)), ' um'], 'HorizontalAlignment','center');
                obj.scale_bar_text_handle.Color = 'red';
                obj.scale_bar_text_handle.FontSize = 20;
                hold off
            else
                obj.toggle_scale_bar.Value = 0;
            end
            
        end
        
        
        function update_clim_text(obj)
            % Change the contrast based on display mode and update the clim
            % fields. 
            switch obj.display_mode
                case 'Stack'
                    clim = obj.clim_orig{obj.cycle,obj.channel};
                case 'Avg'
                    clim = obj.clim_avg{obj.cycle,obj.channel};
                case 'Max'
                    clim = obj.clim_max{obj.cycle,obj.channel};
                case 'Std'
                    clim = obj.clim_std{obj.cycle,obj.channel};
                case 'MovingAvg'
                    clim = obj.clim_avg{obj.cycle,obj.channel};
                case 'MovingStd'
                    clim = obj.clim_std{obj.cycle,obj.channel};
                case 'MovingMax'
                    clim = obj.clim_max{obj.cycle,obj.channel};
            end
            
            obj.text_clim_low.String = num2str(clim(1));
            obj.text_clim_high.String = num2str(clim(2));
            
        end

        
        function update_image_display(obj)
            % Updates the image in the image display
            if isempty(obj.stack)
                return
            end
            
            frames = obj.stack.frames_in_cycle(obj.cycle);
            
            if strcmp(obj.display_mode,'Stack') || contains(obj.display_mode,'Moving') 
                set( obj.text_current_frame, 'String', ...
                      ['Current Frame: ' num2str(obj.frame) '/' num2str(frames)] )
            end

            % Get framenumbers to run window on
            if contains(obj.display_mode,'Moving')
                if obj.frame < ceil(obj.binningSize/2)
                    bin_idx = 1:obj.binningSize;
                elseif (frames - obj.frame)  < ceil(obj.binningSize/2)
                    bin_idx = frames - obj.binningSize+1:frames;
                else
                    bin_idx = obj.frame - floor(obj.binningSize/2):obj.frame + floor(obj.binningSize/2);
                end
            end
            
            % Show one or multiple channels?
            switch obj.channelDisplayMode
                case 'single'
                    ch_indices = obj.channel;
                case 'multi'
                    ch_indices = 1:min(obj.stack.channels,3);
            end
             
            % Load images for specified channels 
            
            dim = size(obj.img_stacks{1});
            if strcmp(obj.channelDisplayMode, 'multi')
                image = zeros(dim(1),dim(2),3,'uint8');
            else
                image = zeros(dim(1),dim(2),'uint8');
            end
            
            for ch  = ch_indices
                img_stack = obj.img_stacks{obj.cycle,ch};
                
                % Get the image and clim.
                switch obj.display_mode
                    case 'Stack'
                        img = img_stack(:, :, obj.frame);
                        clim = obj.clim_orig{obj.cycle,ch};
                    case 'Avg'
                        if isempty(obj.img_avg); return; end
                        img = obj.img_avg{obj.cycle,obj.channel};
                        clim = obj.clim_avg{obj.cycle,obj.channel};
                    case 'Max'
                        if isempty(obj.img_max); return; end
                        img = obj.img_max{obj.cycle,obj.channel};
                        clim = obj.clim_max{obj.cycle,obj.channel};
                    case 'Std'
                        if isempty(obj.img_std); return; end
                        img = obj.img_std{obj.cycle,obj.channel};
                        clim = obj.clim_std{obj.cycle,obj.channel};
                    case 'MovingAvg'
                        img = img_stack(:, :, bin_idx);
                        img = mean(single(img), 3);
                        clim = obj.clim_avg{obj.cycle,obj.channel};
                    case 'MovingStd'
                        img = img_stack(:, :, bin_idx);
                        img = std(single(img), [], 3);
                        clim = obj.clim_std{obj.cycle,obj.channel};
                    case 'MovingMax'
                        img = img_stack(:, :, bin_idx);
                        img = max(img, [], 3);
                        clim = obj.clim_max{obj.cycle,obj.channel};
                end
                
                img = begonia.mat_functions.normalize(single(img),single(clim));
                switch obj.channelDisplayMode
                    case 'single'
                        image(:,:) = uint8(img*255);
                    case 'multi'
                        image(:,:,ch) = uint8(img*255);
                end
            end
            
            if isempty(obj.image_obj) 
                % First time initialization. Create image object
                obj.image_obj = imshow(...
                    image, [0, 255], ...
                    'Colormap',begonia.colormaps.viridis,...
                    'Parent', obj.axStackDisplay, ...
                    'InitialMagnification', 'fit');
                set(obj.image_obj, 'ButtonDownFcn', @obj.callback_mouse_press)
            else
                set(obj.image_obj, 'cdata', image);
            end
            if isempty(obj.editor_display)
                obj.editor_display = calendula.editors.EditorDisplay(obj.axStackDisplay,obj.image_obj);
                obj.editor_display.addlistener('OnModeChanged',@(s,e)obj.listener_mode_changed());
            end

            obj.updateFrameMarker();
            
        end
        
        
        function callback_scroll_wheel(obj, source, event)
            if isempty(obj.stack); return;end
            factor = str2double(obj.edit_scroll_speed.String);
            if isnan(factor); factor = 1; end
            frame = obj.frame + round(event.VerticalScrollCount)*factor;
            obj.update_frame(frame);
        end
        
        
        function obj = changeFrame(obj, source, event, action)
            % Callback from different sources to change the current frame.
            if isempty(obj.stack)
                return
            end
            switch action
                case {'slider', 'buttonclick'}
                    newValue = source.Value;
                    i = newValue -  obj.frame;
                    i = round(i);
                case {'jumptoframe'}
                    i = str2double(source.String) -  obj.frame;
                    i = round(i);
                case 'playvideo'
                    i = 1;
                otherwise
                    i = 0;   
            end
            
            obj.update_frame(obj.frame+i);
        end
        
        
        function update_frame(obj, frame)
            time_since_last_call = toc;
            if time_since_last_call < 1/30
                return; 
            end
            frames = obj.stack.frames_in_cycle(obj.cycle);
            % Check that new value is within range and update current frame/slider info
            if frame >= 1  && frame <= frames
                obj.frame = round(frame);
                set(obj.frameslider, 'Value', obj.frame );
                set(obj.edit_frame, 'String', num2str(obj.frame))
                obj.frameslider.Visible = 'on';
            end

            if ~isempty(obj.stack)
                obj.update_image_display();
            end
            tic
        end
        
        
        function obj = updateFrameMarker(obj)
            % Update line indicating current frame in plot. 
            frameNo = obj.frame;
            if isempty(obj.plotFrameMarker)
                obj.plotFrameMarker = plot(obj.axSignalPlot, [1, 1], get(obj.axSignalPlot, 'ylim'), 'r', 'Visible', 'off', 'HitTest', 'off');
            else
                set(obj.plotFrameMarker, 'XData', [frameNo, frameNo]);
            end
        end
        
        function select_draw_roi_button(obj, btn)
            if nargin < 2
                btn = [];
            end
            buttons = [...
                obj.btn_draw_roi, ...
                obj.btn_draw_roi_circular, ...
                obj.btn_edit_roi, ...
                obj.btn_auto_detect, ...
                obj.btn_mark];
            button_vals = ismember(buttons,btn);
            for i = 1:length(buttons)
                buttons(i).Value = button_vals(i);
            end
        end
              
        
        function callback_toggle_draw_roi(obj, source, ~)
            % Button Callback. Start or finish marking of a roi.
            if source.Value
                obj.set_mouse_mode('Draw');
            else
                obj.cancelRoi();
                obj.set_mouse_mode('Select');
            end 
        end
        
        function callback_toggle_draw_roi_circular(obj,source,~)
            if source.Value
                source.Value = 1;
                obj.set_mouse_mode('DrawCircular');
            else
                obj.cancelRoi();
                obj.set_mouse_mode('Select');
            end 
        end
        
        function add_roi(obj, rois)
            if isempty(rois); return;end
            for i = 1:length(rois)
                roi = rois(i);
                
                obj.roi_array(end+1) = roi;
                
                % Plot the roi in the ax.
                axes(obj.axStackDisplay);
                hold on

                roiColor = obj.getRoiColor(roi);
                % Add each boundary to the plot
                for j = 1:length(roi.boundary)
                    boundary = roi.boundary{j};
                    if j == 1
                        h = plot(boundary(:,2), boundary(:,1));
                        h.Color = roiColor;
                    else
                        h.XData = horzcat(h.XData, nan, boundary(:,2).');
                        h.YData = horzcat(h.YData, nan, boundary(:,1).');
                    end
                end

                if isempty(obj.roi_plot_handles)
                       obj.roi_plot_handles = gobjects(0);
                end

                % Add the plot handle to gui obj and set some properties
                obj.roi_plot_handles(end+1) = h;
                h.HitTest = 'off';
                h.PickableParts = 'none';
                h.LineWidth = 2;

                 % Add the text tag of the roi.
                h = text(0,0,'');
                if isempty(obj.roi_text_handles)
                    obj.roi_text_handles = gobjects(0);
                end

                % Add text handle to gui obj and set some properties
                obj.roi_text_handles(end+1) = h;
                h.HitTest = 'off';
                h.PickableParts = 'none';
                h.HorizontalAlignment = 'center';
                % Set this one later when all the rois are created.
                h.String = 'placeholder';
                h.Position = [roi.center(1), roi.center(2), 0];
                h.Color = roiColor;

                % Set visibility of text based on button "Show/Hide Tags"
                switch obj.btn_show_tags.String
                    case 'Show Tags'
                        h.Visible = 'off';
                    case 'Hide Tags'
                        h.Visible = 'on';
                end
                hold off
            end
            
            obj.update_roi_tags_everywhere();
            
            selected_rois = (length(obj.roi_array) - length(rois) + 1):length(obj.roi_array);
            obj.select_rois(selected_rois);
            
            obj.reorder_rois()
        end
        
        
        function finishRoi(obj)
            % Finish a "drawn" roi object.
            switch obj.mouse_mode
                case 'Draw'
                    new_roi = obj.make_roi();
                    obj.add_roi(new_roi);
                    obj.removeTmpRoi();

                case 'EditDraw'
                    roi = obj.roi_array(obj.selected_rois);
                    % Make new roi with some inherited properties
                    new_roi = make_roi(obj);
                    new_roi = obj.inheritRoiProperties(new_roi, roi);
                    obj.roi_array(obj.selected_rois) = new_roi;
                    obj.update_roi_tags();

                    % Replot the roi
                    updateRoiPlot(obj, obj.selected_rois, new_roi);

                    % Remove the temporary roi polygon 
                    obj = removeTmpRoi(obj);
                    obj.set_mouse_mode('EditSelect');
            end
        end
        
        function update_roi_tags_everywhere(obj)
            % Updates:
            %   The roi_tags property
            %   The roi tags in the list box.
            %   The roi tags in every text handle.
            % Also updates the selected values in the listbox.
            
            % Create incrementing numbers for each group.
            groups = {obj.roi_array.group};
            nums = zeros(1,length(groups));
            [C,IA,IC] = unique(groups);
            for i = 1:length(C)
                ind = IC == i;
                nums(ind) = 1:sum(ind);
            end
            
            % Create tags based on group and number.
            tags = {};
            for i = 1:length(obj.roi_array)
                tag = sprintf([obj.roi_array(i).group,'%02d'],nums(i));
                h = obj.roi_text_handles(i);
                h.String = tag;
                tags{i} = tag;
            end
            % Add the tag of a connected roi
            for i = 1:length(obj.roi_array)
                roi = obj.roi_array(i);
                if isempty(roi.id_connected_roi); continue; end;
                % Find the connected roi.
                id = roi.id_connected_roi;
                ids = {obj.roi_array.id};
                connected_tag = tags{strcmp(id,ids)};
                tags{i} = [tags{i},'_',connected_tag];
            end
            
            obj.roi_tags = tags;
            obj.update_listbox();
        end
        
        function update_listbox(obj)
            % Find the indices of the rois that are in the current channel.
            ind_channel = [obj.roi_array.channel] == obj.channel;
            ind_cycle = [obj.roi_array.cycle] == obj.cycle;
            ind = find(ind_channel & ind_cycle);
            
            % Of those rois, find the ones which are selected.
            selected = ismember(ind,obj.selected_rois);
            selected = find(selected);
            
            % Update the listbox with the rois and select them.
            obj.listbox_rois.String = obj.roi_tags(ind);
            obj.listbox_rois.Value = selected;
            
            % Update the connected rois popup, first value is empty.
            % Knut update: roi list would change in list depending on the
            % sorting of the names of the rois. This should ensure the
            % actually last selected connected roi gets picked, rather than
            % restoring the index of the preveious one in the list (which
            % could have changed as RoIs are added or removed:
            con_rois = {' ', obj.roi_tags{:}};
            obj.popup_roi_connected.String = con_rois;
            ind_con_roi = find(con_rois == string(obj.selected_connected_roi));
            if isempty(ind_con_roi)
                obj.popup_roi_connected.Value = 1;
            else
                obj.popup_roi_connected.Value = ind_con_roi;
            end
        end
        
        
        function obj = cancelRoi(obj)
            % Cancel a "drawn" roi object
        
            % Remove Temporary Roi Object
            obj = removeTmpRoi(obj);
            ch = obj.channel;
            
            % Replot Roi if edit mode is active
            if strcmp(obj.mouse_mode, 'EditDraw')
                updateRoiPlot(obj, obj.selected_rois, obj.roi_array{ch}(obj.selected_rois));
                obj.set_mouse_mode('EditSelect');
            end
        end
        
        
        function obj = callback_toggle_auto_detect_roi(obj, source, ~)
        % Button Callback. Start or finish autodetection of roi by clicking on it. 
            
            if source.Value
                obj.set_mouse_mode('Autodetect');
            else
                obj.set_mouse_mode('Select');
            end  
        end
            
        
        function obj = callback_toggle_edit_roi(obj, source, ~)
        % Button Callback. Start or finish editing of rois.

%             if source.Value
%                 obj.set_mouse_mode('EditSelect');
%                 set(obj.btn_auto_detect, 'Value', 0)
%                 
%             else
%                 obj.cancelRoi();
%                 obj.set_mouse_mode('Select');
%                 obj.unFocusButton(obj.btn_draw_roi)
%             end  
            
        end
        
        
        function callback_btn_remove_roi(obj, ~, ~)
            % Button callback. Remove selected rois from gui
            obj.remove_selected_rois();
        end
        
        
        function remove_selected_rois(obj)  
            if isempty(obj.stack); return; end
            obj.listbox_rois.Value = [];
            
            % Loop through selected rois, remove roi objects and associated
            % plots. Also clear the roi from the listbox
            for i = fliplr(obj.selected_rois)
                obj.roi_array(i) = [];
                delete(obj.roi_plot_handles(i))
                delete(obj.roi_text_handles(i))
                obj.roi_plot_handles(i) = [];
                obj.roi_text_handles(i) = [];
                obj.roi_tags(i) = [];
                obj.listbox_rois.String(i) = [];
            end
            obj.selected_rois = [];
            obj.reorder_rois();
        end

        
        function callback_btn_clear_rois(obj, ~, ~)
            % Button callback. Remove all rois from gui and listbox, and remove
            % all plots and labels.

            obj.listbox_rois.Value = [];
            obj.roi_array = calendula.RoI.empty;
            
            for i = 1:length(obj.roi_plot_handles)
                delete(obj.roi_plot_handles(i));
            end
            obj.roi_plot_handles = [];
            
            for i = 1:length(obj.roi_text_handles)
                delete(obj.roi_text_handles(i));
            end
            obj.roi_text_handles = [];

            obj.listbox_rois.String = {};
            obj.selected_rois = [];
            obj.roi_tags = {};
        end
        
              
        function obj = drawTmpRoi(obj, ax)
            % Draw the lines between the impoints of tmp roi. 
        
            % Get list of vertex points
            x = obj.RoiTmpPosX;
            y = obj.RoiTmpPosY;
            
            if length(x) < 2 || length(y) < 2
                return
            end
            
            % Close the circle
            x(end+1) = x(1);
            y(end+1) = y(1);
            
            % There should only be one instance of the tmp roi plot. 
            if isempty(obj.RoiTmpPlotHandle) || ~isvalid(obj.RoiTmpPlotHandle)
                axes(ax);
                hold on
                obj.RoiTmpPlotHandle = plot(0,0);
                obj.RoiTmpPlotHandle.HitTest = 'off';
                obj.RoiTmpPlotHandle.PickableParts = 'none';
                hold off
            end
            set(obj.RoiTmpPlotHandle,'XData',x,'YData',y);
            
            obj.text_roi_tag.String = 'Temporary RoI';
            obj.text_roi_size.String = num2str(round(polyarea(x,y)*obj.stack.dx*obj.stack.dy));
        end
        
        
        function obj = addTmpRoiVertex(obj, ax, x, y)
        % ADDTMPROIVERTEX adds a new tmp roi vertex to the axes.  
        % After the impoint is created it is also configured.
        %   addTmpRoiVertex(obj, ax, x, y)
        %   x, y       - Coordinates in pixels. 
        %
        %   See also configImpoint, impoint

            % Find the index of this edge.
            i = length(obj.RoiTmpEdges) + 1;
            
            % The vertices are impoints that can be moved around. 
            tmpRoiVertex = impoint(ax, x, y);
            configImpoint(obj, ax, tmpRoiVertex, i);
            obj.RoiTmpEdges{end+1} = tmpRoiVertex;
        end
        
        
        function obj = moveTmpRoiVertex(obj, ax, x, y, i)
            %MOVETMPROIVERTEX moves a vertex (impoint) of the tmproi. Will also redraw
            %the lines between the vertices. 
            %   moveTmpRoiVertex(obj, ax, x, y, i)
            %   x, y          - New coordinates in pixels. 
            %   i             - Index of the vertex in obj.RoiTmpPos
            %
            % See also callbackRoiPosChanged.

            obj.RoiTmpPosX(i) = x;
            obj.RoiTmpPosY(i) = y;
            % Draw the lines between the edges.
            obj = drawTmpRoi(obj, ax);
            % Move the impoint objects to the edges. 
            obj.RoiTmpEdges{i}.setPosition([x,y]);
            
        end

        
        function configImpoint(obj, ax, impointObj, i)
        %CONFIGIMPOINT configures an impoint. 
        % Sets the new position callback of impoints. They are responsible for
        % updating the plot when a vertex is moved. 
        %   configImpoint(obj, ax, impointObj, i)
        %   impointObj    - impoint to configure. 
        %   i             - Sent to the move callback. Index of the impoint. 
        %
        % See also impoint, moveTmpRoiVertex

            impointObj.addNewPositionCallback(@(pos)callbackRoiPosChanged(obj, pos, ax, i));
            impointObj.setPositionConstraintFcn(@(pos)lockImpointInZoomMode(obj, pos, i))
            impointObj.Deletable = false;
        end
        
        
        function constrained_pos = lockImpointInZoomMode(obj, new_pos, i)
        % Callback function when dragging impoint. Locks impoint in place
        % during zoom mode.
            switch obj.mouse_mode
                case {'Zoom In', 'Zoom Out'}
                	x = obj.RoiTmpPosX(i);
                    y = obj.RoiTmpPosY(i);
                    constrained_pos = [x, y];
                otherwise
                    constrained_pos = new_pos;
            end   
        end
        
        
        function obj = callbackRoiPosChanged(obj, pos, ax, i)
            % callback function of impoint. 
            % This function is called whenever a impoint is moved (Tmp RoI vertex). 
            %
            % See also configImpoint, impoint, moveTmpRoiVertex
            
            x = pos(1);
            y = pos(2);
            
            obj.moveTmpRoiVertex(ax, x, y, i);
        end
        
        function roi = make_roi_circular(obj,x,y)
            r = str2num(obj.edit_roi_radius.String)/obj.stack.dx;
            dim = size(obj.img_stacks{obj.cycle,obj.channel});
            mask = false(dim(1:2));
            for i = 1:dim(1)
                for j = 1:dim(2)
                    if sqrt((i - y)^2 + (j-x)^2) < r
                        mask(i,j) = true;
                    end
                end
            end
            
            group = obj.roi_groups{obj.popup_roi_groups.Value};
            roi = calendula.RoI(mask);
            roi.group = group;
            roi.channel = obj.channel;
            roi.cycle = obj.cycle;
            roi.shape = 'circle';
            if obj.popup_roi_connected.Value > 1
                roi.id_connected_roi = obj.roi_array(obj.popup_roi_connected.Value-1).id;
            end
        end
        
        function roi = make_roi(obj)
            x = obj.RoiTmpPosX;
            y = obj.RoiTmpPosY;
            if length(x) < 3 || length(y) < 3
                roi = calendula.RoI.empty;
                return
            end
            
            % Get the selected group.
            group = obj.roi_groups{obj.popup_roi_groups.Value};
            
            % Create a RoI object from the polygon.
            dim =  size(obj.img_stacks{obj.channel});
            m = dim(1);
            n = dim(2);
            bw = poly2mask(x, y, m, n);
            
            roi = calendula.RoI(bw);
            roi.group = group;
            roi.channel = obj.channel;
            roi.cycle = obj.cycle;
            roi.shape = 'polygon';
            if obj.popup_roi_connected.Value > 1
                roi.id_connected_roi = obj.roi_array(obj.popup_roi_connected.Value-1).id;
            end
        end
        
        
        function obj = removeTmpRoi(obj)
        %REMOVETMPROI clear the obj.RoiTmpPos or obj.RoiTmpEdges.

            obj.RoiTmpPosX = [];
            obj.RoiTmpPosY = [];
            len = length(obj.RoiTmpEdges);
            for i = 1:len
                delete(obj.RoiTmpEdges{i});
            end
            
            delete(obj.RoiTmpPlotHandle)

            obj.RoiTmpEdges = cell(0);
            
        end
        
        
        function select_rois(obj, selected_rois)
            if nargin < 2
                selected_rois = obj.selected_rois;
            end
            selected_rois = unique(selected_rois);
            obj.selected_rois = selected_rois;
            
            for i = 1:length(obj.roi_array)
                if any(i == obj.selected_rois)
                    % Selected
                    obj.roi_plot_handles(i).Color = 'White';
                    obj.roi_text_handles(i).Color = 'White';
                else
                    % Not selected
                    color = obj.getRoiColor(obj.roi_array(i));
                    obj.roi_plot_handles(i).Color = color;
                    obj.roi_text_handles(i).Color = color;
                end
            end
            
            % Plot signal.
            if obj.btn_show_signal.Value
                obj.resetSignalPlot();
                obj.updateSignalPlot(obj.selected_rois, 'append');
            end
            
            % Roi info
            if isempty(obj.selected_rois)
                obj.text_roi_tag.String = 'N/A';
                obj.text_roi_size.String = 'N/A';
            else
                idx = obj.selected_rois(1);
                roi = obj.roi_array(idx);
                obj.text_roi_tag.String = obj.roi_tags{idx};
                obj.text_roi_size.String = num2str(round(roi.area*obj.stack.dx*obj.stack.dy));
            end
            
            obj.update_listbox();
        end
         
         
        function selectListBoxObj(obj, source, ~)
            if isempty(obj.stack); return; end;
            % Change selection status of roi that are selected in listbox
            if ~strcmp(obj.mouse_mode, 'EditSelect')
                % Convert the selected values of the list box to
                % indices of all rois. This must be done because the list
                % box is usually not showing all of the rois.
                selected = source.Value;
                indices = find([obj.roi_array.channel] == obj.channel);
                selected = indices(selected);
                obj.select_rois(selected);
                
%                 newlySelected  = setdiff(selected, obj.selected_rois);
%                 unselected_rois = setdiff(obj.selected_rois, selected);
% 
%                 if ~isempty(newlySelected)
%                     for i = newlySelected
%                         obj.roi_plot_handles(i).Color = 'White';
%                         obj.roi_text_handles(i).Color = 'White';
%                     end
%                 end
% 
%                 if ~isempty(unselected_rois)
%                     for i = unselected_rois
%                         color = obj.getRoiColor(obj.roi_array(i));
%                         obj.roi_plot_handles(i).Color = color;
%                         obj.roi_text_handles(i).Color = color;
%                     end
%                 end
% 
%                 obj.selected_rois = selected;
%                 
%                 if obj.btn_show_signal.Value
%                     obj.resetSignalPlot();
%                     for i = selected
%                         obj.updateSignalPlot(i, 'append');
%                     end
%                 end
                
            else 
                source.Value = [];
            end
            
        end
        
        
        function obj = deselectRois(obj)
            % Deselect all selected rois
            for i = obj.selected_rois
                color = obj.getRoiColor(obj.roi_array(i));
                obj.roi_plot_handles(i).Color = color;
                obj.roi_text_handles(i).Color = color;
            end
            
            obj.selected_rois = [];
        
        end
         
        
        function obj = updateRoiPlot(obj, idx, newRoi)
            % Replot the roi at idx in roi_array
            hold on
            for j = 1:length(newRoi.Boundary)
                boundary = newRoi.Boundary{j};
                if j == 1
                    set(obj.roi_plot_handles(idx), 'XData', boundary(:,2), 'YData', boundary(:,1));
                else
                    obj.roi_plot_handles(idx).XData = horzcat(obj.roi_plot_handles(idx).XData, nan, boundary(:,2)');
                    obj.roi_plot_handles(idx).YData = horzcat(obj.roi_plot_handles(idx).YData, nan, boundary(:,1)');
                end

            end

            % Move roi label/tag to new center position
            set(obj.roi_text_handles(idx), 'Position', [newRoi.Center, 0])
            hold off
            
        end
        
        
        function shiftRoiPlot(obj, shift)
        % Shift Roi plots according to a shift [x, y, 0]
            % Get active roi
            for i = obj.selected_rois
                
                xData = get(obj.roi_plot_handles(i), 'XData');
                yData = get(obj.roi_plot_handles(i), 'YData');
            
                % Calculate and update position 
                xData = xData + shift(1);
                yData = yData + shift(2);
                set(obj.roi_plot_handles(i), 'XData', xData)
                set(obj.roi_plot_handles(i), 'YData', yData)
            
                % Shift text to new position
                textpos = get(obj.roi_text_handles(i), 'Position');
                textpos = textpos + shift;
                set(obj.roi_text_handles(i), 'Position', textpos);
            end
        end
        
        
        function obj = callback_btn_load_rois(obj, ~, ~)
            % Open filebrowser in same location as tiffstack was loaded from
            path = uigetdir(obj.path_prev_load_rois,'Find Stacks');
            if path == 0
                return
            end
            
            obj.path_prev_load_rois = path;
            
            stack = begonia.scantype.find_tseries(path);
            if ~isempty(obj.dloc_engine)
                [stack.dl_storage_engine] = deal(obj.dloc_engine);
            end
            if length(stack) > 1
                disp('Error: Multiple stacks found.')
                return;
            end
            if isempty(stack)
                disp('Error: No stack found.')
                return
            end
            if ~stack.has_var('roi_array')
                disp('No RoIs found.');
            end
            roi_array = stack.load_var('roi_array');
            % This will update old rois to the current version.
            roi_array = calendula.RoI(roi_array);
            % Update the ids of the rois so we will have no collisions
            % later. 
            roi_array.create_new_ids();
            
            cycles = {};
            channels = stack.channel_names;
            for i = 1:stack.cycles
                cycles{i} = num2str(i);
            end
            
            f = figure(...
                'Name', 'Load RoIs', ...
                'NumberTitle', 'off');
            f.Position(3) = 500;
            f.Position(4) = 200;
            set(f, 'menubar', 'none');
            
            n_btns = 6;
            
            uicontrol(...
                'Style','text',...
                'String','Specify which channel and cycle to load rois from', ...
                'Parent', f, ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [0.2, 5/n_btns, 0.6, 1/n_btns])
            
            uicontrol(...
                'Style','text',...
                'String','Cycle', ...
                'Parent', f, ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [0.2, 4/n_btns, 0.6, 1/n_btns]);
            
            cycle = uicontrol(...
                'Style','popupmenu',...
                'String',cycles, ...
                'Parent', f, ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [0.2, 3/n_btns, 0.6, 1/n_btns]);
            
            uicontrol(...
                'Style','text',...
                'String','Channel', ...
                'Parent', f, ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [0.2, 2/n_btns, 0.6, 1/n_btns]);
            
            channel = uicontrol(...
                'Style','popupmenu',...
                'String',channels, ...
                'Parent', f, ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Position', [0.2, 1/n_btns, 0.6, 1/n_btns]);
            
            btn = uicontrol(...
                'Style','pushbutton',...
                'String','Load RoIs', ...
                'Parent', f, ...
                'HorizontalAlignment', 'center', ...
                'Units', 'normalized', ...
                'Callback', @sub_load_rois, ...
                'Position', [0.2, 0/n_btns, 0.6, 1/n_btns]);
            
            function sub_load_rois(o, ~, ~)
                idx_1 = [roi_array.cycle] == cycle.Value;
                idx_2 = [roi_array.channel] == channel.Value;
                rois = roi_array(idx_1 & idx_2);
                for roi = rois
                    % Change properties of the rois so they fit the current
                    % stack.
                    roi.cycle = obj.cycle;
                    roi.channel = obj.channel;
                end
                obj.add_roi(rois);
                close(f)
            end
            
        end
        
        
        function callback_btn_save_rois(obj, ~, ~)
            % Export rois to file. Save to sessionfolder if available
            roi_array = obj.roi_array.to_struct();
            obj.stack.save_var(roi_array);
            save(fullfile(obj.stack.path, 'roi_array.mat'), 'roi_array')
            disp('RoIs saved.')
        end
        
        
        function callback_btn_export_signal(obj, ~, ~)
            % Export signals to file. Save to sessionfolder if available
            disp('Saving signal ...');
            tic
            signal_data = obj.roi_array.get_signal(obj.stack,'waitbar')
            
            obj.stack.save_var(signal_data);
            
            save_path = fullfile(obj.stack.path,[obj.stack.name,'_signal.mat']);
            save(save_path,'signal_data')
            toc
        end
        
        
        function callback_btn_show_tags(obj, source, ~)
            % Change visibility of roi tags based on state of button "Show Tags"

            indices_on = [obj.roi_array.channel] == obj.channel;
            switch source.String
                case 'Show Tags'
                    set(obj.btn_show_tags, 'String', 'Hide Tags')
                    set(obj.roi_text_handles(indices_on),'Visible','on');
                case 'Hide Tags'
                    set(obj.btn_show_tags, 'String', 'Show Tags')
                    set(obj.roi_text_handles(indices_on),'Visible','off');
            end

        end
        
        
        function callback_btn_mark_add(obj,~,~)
            if isempty(obj.editor_display); return; end
            
            mask = obj.editor_display.image_marked;
            
            action = calendula.actions.DisplayAction(obj.editor_display);
            action.mode = 'select';
            obj.editor_display.do_and_record_action(action);
            
            if sum(mask(:)) == 0
                return; 
            end
            
            
            rois = obj.roi_array(obj.selected_rois);
            if ~isempty(rois)
                roi = rois.merge();
                obj.remove_selected_rois()
                mask_base = roi.mask;
                mask = mask_base | mask;
                roi.set_mask(mask);
            else
                roi = calendula.RoI(mask);
                roi.group = obj.roi_groups{obj.popup_roi_groups.Value};
            end
            
            roi.channel = obj.channel;
            roi.cycle = obj.cycle;
            roi.shape = 'brush';
            if obj.popup_roi_connected.Value > 1
                roi.id_connected_roi = obj.roi_array(obj.popup_roi_connected.Value-1).id;
            end
            obj.add_roi(roi);
        end
        
        
        function callback_btn_mark_remove(obj,~,~)
            if isempty(obj.editor_display); return; end
            
            mask = obj.editor_display.image_marked;
            
            action = calendula.actions.DisplayAction(obj.editor_display);
            action.mode = 'select';
            obj.editor_display.do_and_record_action(action);
            
            rois = obj.roi_array(obj.selected_rois);
            if isempty(rois); return; end
            base_roi = rois.merge();
            mask_base = base_roi.mask;
            mask = mask_base & mask;
            mask = xor(mask,mask_base);
            base_roi.set_mask(mask);
            
            
            if sum(mask(:)) == 0
                return; 
            end
            obj.remove_selected_rois()
            
            CC = bwconncomp(mask);
            new_rois = calendula.RoI.empty(0);
            for i = 1:CC.NumObjects
                m = false(size(mask));
                m(CC.PixelIdxList{i}) = true;
                new_rois(i) = calendula.RoI(base_roi);
                new_rois(i).set_mask(m);
                new_rois(i).id = begonia.util.make_uuid();
            end
            obj.add_roi(new_rois);
        end
        
        
        function callback_btn_mark_split(obj,~,~)
            if isempty(obj.editor_display); return; end
            
            mask = obj.editor_display.image_marked;
            
            action = calendula.actions.DisplayAction(obj.editor_display);
            action.mode = 'select';
            obj.editor_display.do_and_record_action(action);
            
            rois = obj.roi_array(obj.selected_rois);
            if isempty(rois); return; end
            base_roi = rois.merge();
            mask_base = base_roi.mask;
            mask_and = mask_base & mask;
            mask = xor(mask_and,mask_base);
            
            if sum(mask(:)) == 0
                return; 
            end
            obj.remove_selected_rois()
            
            CC = bwconncomp(mask);
            new_rois = calendula.RoI.empty(0);
            for i = 1:CC.NumObjects
                m = false(size(mask));
                m(CC.PixelIdxList{i}) = true;
                new_rois(i) = calendula.RoI(base_roi);
                new_rois(i).set_mask(m);
                new_rois(i).id = begonia.util.make_uuid();
            end
            
            new = calendula.RoI(base_roi);
            new.set_mask(mask_and);
            new.id = begonia.util.make_uuid();
            new_rois(end+1) = new;
            
            obj.add_roi(new_rois);
        end
        
        
        function callback_btn_mark(obj,source,~)
            if isempty(obj.editor_display); return; end
            action = calendula.actions.DisplayAction(obj.editor_display);
            if source.Value
                % Button is down
                action.mode = 'mark';
            else
                action.mode = 'select';
            end
            obj.editor_display.do_and_record_action(action);
        end
        
        
        function obj = callback_btn_show_stack(obj, ~, ~)
            obj.set_display_mode('Stack');
        end
        
        
        function load_image_projections(obj, calculate_if_fail)
            if nargin < 2
                calculate_if_fail = false;
            end
            if isempty(obj.stack)
                return
            end
            
            obj.img_avg = {};
            obj.img_max = {};
            obj.img_std = {};
            obj.clim_orig = {};
            obj.clim_avg = {};
            obj.clim_std = {};
            obj.clim_max = {};
            
            % Load average, max, std image. 
            for i = 1:obj.stack.cycles
                for j = 1:obj.stack.channels
                    obj.img_avg{i,j} = obj.stack.get_avg_img(i,j);
                    obj.img_max{i,j} = obj.stack.get_max_img(i,j);
                    obj.img_std{i,j} = obj.stack.get_std_img(i,j);
                    % Set contrast limits.
                    obj.clim_orig{i,j} = [0,2^12-1];
                    obj.clim_avg{i,j} = [min(obj.img_avg{i,j}(:)),max(obj.img_avg{i,j}(:))];
                    obj.clim_max{i,j} = [min(obj.img_max{i,j}(:)),max(obj.img_max{i,j}(:))];
                    obj.clim_std{i,j} = [min(obj.img_std{i,j}(:)),max(obj.img_std{i,j}(:))];
                end
            end
        end
        
        
        function callback_toggle_show_avg(obj, ~, ~)
            % Shows stack average projection in image display
            if isempty(obj.img_avg)
                obj.load_image_projections(false);
            end
            
            obj.set_display_mode('Avg');
            obj.update_clim_text();
            obj.update_image_display();
        end
        
        
        function obj = callback_toggle_show_max(obj, ~, ~)
            % Shows stack average projection in image display
            if isempty(obj.img_max)
                obj.load_image_projections();
            end
            obj.set_display_mode('Max');
        end      
        
        
        function obj = callback_toggle_show_std(obj, ~, ~)
            % Shows stack average projection in image display
            if isempty(obj.img_std)
                obj.load_image_projections();
            end
            obj.set_display_mode('Std');
        end   

        
        function roiNew = make_roi_autodetect(obj, x, y)
            %autodetect autodetects roi by automatic thresholding.   
            % Set some parameters
            
            d1 = obj.roiInnerDiameter; 
            d2 = obj.roiOuterDiameter; 
            minRoiSize = round(pi*(d2/2)^2/2);
            x = round(x);
            y = round(y);
            % retrieve box 2xroi size around x and y
            
            roiMask = zeros(size(obj.image_obj.CData(:, :,1)));
            
            % /todo take care of image borders
            im = obj.image_obj.CData(y-d2:y+d2, x-d2:x+d2, 1);
            %imChunk = obj.tiffStack(y-d2:y+d2, x-d2:x+d2, :);
            
            
            % Define center coordinates and radius
            imdim = size(im);
            x_mask = imdim(1)/2;
            y_mask = imdim(2)/2;
            r1 = d1/2;
            r2 = d2/2;

            % Generate grid with binary mask representing the outer circle. Credit
            % StackOverflow??
            [xx, yy] = ndgrid((1:imdim(1)) - y_mask, (1:imdim(2)) - x_mask);
            mask1 = (xx.^2 + yy.^2) < r1^2;
            mask2 = (xx.^2 + yy.^2) < r2^2;
            mask3 = logical(mask2-mask1);
            
            nucleus_values = im(mask1);
            soma_values = im(mask2);
            ring_values = im(mask3);
            surround_values = im(~mask2);
            
            im = medfilt2(im, [5, 5]);
            %imshow(im);
            if ~isempty(nucleus_values)
                nucleus_val = median(nucleus_values);
                ring_val = median(ring_values);
                surround_val = median(surround_values(1:round(end*0.6)));
                threshold = double((ring_val - surround_val) / 2 + surround_val);
            else 
            	high_val = median(soma_values);
                low_val = median(surround_values);
            
                threshold = double((high_val - low_val) / 2 + low_val);
            end    
            
            % Create roimask
            localRoiMask = im2bw(im, threshold/255);
            if ~isempty(nucleus_values)
                localRoiMask(mask1) = 1;
            end

            roiMask(y-d2:y+d2, x-d2:x+d2) = localRoiMask;
            
            % remove small "holes"
            roiMask = bwareaopen(roiMask, minRoiSize);
            
            if sum(roiMask(:)) > 0
                % Create a RoI object
                group = obj.roi_groups{obj.popup_roi_groups.Value};
                roiNew = calendula.RoI(roiMask);
                roiNew.cycle = obj.cycle;
                roiNew.channel = obj.channel;
                roiNew.group = group;
                roiNew.shape = 'autodetect';
                if obj.popup_roi_connected.Value > 1
                    roiNew.id_connected_roi = obj.roi_array(obj.popup_roi_connected.Value-1).id;
                end
            else
                roiNew = [];
            end
            
        end
        
  
        function obj = callback_toggle_show_moving_avg(obj, ~, ~)
            % Shows stack running average projection in image display
            obj.set_display_mode('MovingAvg');
        end
        
        
        function obj = callback_toggle_show_moving_std(obj, ~, ~)
            % Shows stack running standard deviation projection in image display
            obj.set_display_mode('MovingStd');
        end
        
        
        function obj = callback_toggle_show_moving_max(obj, ~, ~)
            obj.set_display_mode('MovingMax');
        end
        
        
        function listener_mode_changed(obj)
            switch obj.editor_display.mode
                case 'mark'
                    obj.select_draw_roi_button(obj.btn_mark)
                otherwise
                    obj.select_draw_roi_button()
            end
            obj.set_mouse_mode('Select');
        end
        
        
        function deselect_display_mode_buttons(obj)
            % Untoggle all buttons except for source button.
            showButtons = { obj.btn_show_moving_max, ...
                            obj.btn_show_moving_std, ...
                            obj.btn_show_moving_avg, ...
                            obj.btn_show_avg, ...
                            obj.btn_show_max, ...
                            obj.btn_show_std};
            
            for i = 1:numel(showButtons)
                showButtons{i}.Value = 0;
            end
        end
        
        
        function obj = callback_edit_frame_window(obj, source, ~)
            % Updates the binning size for moving averages. Forces new value to
            % be odd
            newBinningSize = str2double(source.String);
            if ~mod(newBinningSize, 2)
                newBinningSize = newBinningSize - 1;
            end
            obj.binningSize = newBinningSize;
            set(obj.edit_frame_window, 'String', num2str(obj.binningSize))
        
        end
        
        
        function obj = callback_btn_set_roi_size(obj, source,  ~)
        % Callback for button to set the roi diameter to use for autodetection
        
            switch source.String
                case 'Set Roi Size for Autodetection'
                    set(obj.btn_set_roi_size, 'String', 'Confirm')
                    set(obj.roiSizeSliderContainer, 'Visible', 'on')
                    set(obj.btn_set_roi_size, 'TooltipString', 'Move cursor to image and click to reposition the Roi Template')
                    obj.set_mouse_mode('AutoRoiDiameter');
                    if isempty(obj.roiTemplateCenter)
                        xlim = get(obj.axStackDisplay, 'Xlim');
                        ylim = get(obj.axStackDisplay, 'Ylim');
                        obj.roiTemplateCenter = [xlim(1) + diff(xlim)/2, ylim(1) + diff(ylim)/2];
                        obj = plotRoiTemplate(obj);
                    end
                    
                    set(obj.roiTemplatePlotHandle, 'Visible', 'on')
                    
                case 'Confirm'
                    set(obj.btn_set_roi_size, 'String', 'Set Roi Size for Autodetection')
%                     set(obj.roiSizeSliderContainer, 'Visible', 'off')
                    set(obj.btn_set_roi_size, 'TooltipString', '')
                    set(obj.roiTemplatePlotHandle, 'Visible', 'off')
                    obj.set_mouse_mode('Previous');
            end        
        
        end
            
        
        function obj = callback_brush_size(obj, slider, ~)
            % Callback function for value change of roi diameter slider 
            obj.roiInnerDiameter = slider.Low;
            obj.roiOuterDiameter = slider.High;
            plotRoiTemplate(obj);
            
            if ~isempty(obj.editor_display)
                obj.editor_display.brush_radius = slider.High;
            end
        end
        
        
        function obj = callback_slider_brightness(obj, slider, ~)
        % Callback function for value change of brightness slider
            min_brightness = slider.Low;
            max_brightness = slider.High;
            if max_brightness == min_brightness
                max_brightness = min_brightness + 1;
            end
            set(obj.axStackDisplay, 'CLim', [min_brightness, max_brightness])
        end
        
        
        function obj = plotRoiTemplate(obj)
        % Plot a circle with diameter equal to roi template diameter.
            if ~isempty(obj.roiTemplateCenter) && ~isempty(obj.image_obj)
                % Define center coordinates and radius
                
                x = obj.roiTemplateCenter(1);
                y = obj.roiTemplateCenter(2);
                r1 = obj.roiInnerDiameter/2;
                r2 = obj.roiOuterDiameter/2;

                th = 0:pi/50:2*pi;
                xdata1 = r1 * cos(th) + x;
                ydata1 = r1 * sin(th) + y;
                xdata2 = r2 * cos(th) + x;
                ydata2 = r2 * sin(th) + y;
                
                if r1 > 0
                    xData = horzcat(xdata1, nan, xdata2);
                    yData = horzcat(ydata1, nan, ydata2);
                else
                    xData = xdata2;
                    yData = ydata2;
                end
                
                hold(obj.axStackDisplay, 'on')
                if isempty(obj.roiTemplatePlotHandle)
                    obj.roiTemplatePlotHandle = plot(obj.axStackDisplay, xData, yData, 'yellow');
                else
                    set(obj.roiTemplatePlotHandle, 'XData', xData, 'YData', yData)
                end
                hold(obj.axStackDisplay, 'off')
            end
        end
        
        
        function reorder_rois(obj)
            % Sorts rois
            [~,I] = sort(obj.roi_tags);
            obj.roi_tags = obj.roi_tags(I);
            obj.roi_array = obj.roi_array(I);
            obj.roi_plot_handles = obj.roi_plot_handles(I);
            obj.roi_text_handles = obj.roi_text_handles(I);
            obj.update_roi_tags_everywhere();
        end
        
        
        function signal = extractSignal(obj, selectedRoi)
        % Extract signal from currently selected roi
        
            currentRoi = obj.roi_array(selectedRoi);
            minX = min(currentRoi.x);
            minY = min(currentRoi.y);
            maxX = max(currentRoi.x);
            maxY = max(currentRoi.y);
            
            imChunk = obj.img_stacks{obj.cycle,obj.channel}(minY:maxY, minX:maxX, :);
            roiMask = repmat(currentRoi.mask(minY:maxY, minX:maxX), 1,1,obj.stack.frames_in_cycle(obj.cycle));
            imChunk(~roiMask) = 0;
            signal = squeeze(sum(sum(imChunk, 1), 2)) / length(currentRoi.x);
%            signal = smooth(signal, 10);
            sorted = sort(signal);
            
            f0 = median(sorted(1:round(end*0.2)));
            deltaFoverF = (signal - f0) ./ f0;
            signal = deltaFoverF;
%             [minS, minId] = min(signal);
%             [maxS, maxId] = max(signal);

            
        end
        
        
        function obj = callback_toggle_show_signal(obj, source, ~)
            if source.Value
                set(obj.axSignalPlot, 'Visible', 'on')  
                set(obj.signalPlotHandle, 'Visible', 'on')
                set(obj.plotFrameMarker, 'Visible', 'on')
                set(obj.axStackDisplay, 'Position', [0.12, 0.2, 0.76, 0.76])
                drawnow
                set(obj.image_obj, 'ButtonDownFcn', @obj.callback_mouse_press)
            else
                obj.resetSignalPlot();
                set(obj.axSignalPlot, 'Visible', 'off')
                set(obj.signalPlotHandle, 'Visible', 'off')
                set(obj.plotFrameMarker, 'Visible', 'off')
                set(obj.axStackDisplay, 'Position', [0.03, 0.03, 0.94, 0.94])
            end
            
        end
        
        
        function obj = updateSignalPlot(obj, selectedRois, mode)
            for i = 1:length(selectedRois)
                selectedRoi = selectedRois(i);
                % Update signal plot
                df_F = obj.extractSignal(selectedRoi);
                ylim = [min(df_F) * 0.9, max(df_F) * 1.1];
                %ylim=[0,1];
                set(obj.axSignalPlot, 'YLim', ylim)
                set(obj.axSignalPlot, 'YTick', 0:0.5:max(ylim))

                if isempty(obj.signalPlotHandle)
                    obj.signalPlotHandle = plot(obj.axSignalPlot, df_F, 'HitTest', 'off');
                else
                    switch mode
                        case 'append'
                            obj.signalPlotHandle(end+1) = plot(obj.axSignalPlot, df_F, 'HitTest', 'off');
                        case 'overwrite'
                            set(obj.signalPlotHandle(end), 'YData', df_F)
                    end
                end

                set(obj.plotFrameMarker, 'YData', ylim)
            end
        end
        
        
        function obj = resetSignalPlot(obj)
            for i = 1:length(obj.signalPlotHandle)
                delete(obj.signalPlotHandle(i))
            end
            obj.signalPlotHandle=gobjects(0);
        end
        
        
        function playVideo(obj, ~, ~)
            % Callback for play button. Plays calcium images as video
            
            if obj.btnPlayVideo.Value
                                
                while obj.frame < obj.stack.frames_in_cycle(obj.cycle)
                    t1 = tic;
                    obj.changeFrame([], [], 'playvideo');
                    if ~ obj.btnPlayVideo.Value
                        break
                    end
                    t2 = toc(t1);
                    pause(0.033 / obj.playbackspeed-t2)
                end
                
                obj.btnPlayVideo.Value = 0;
            end
            
        end
        
        
        function btnPlaybackSpeed(obj, source, ~)
            switch source.String
                case '2x'
                    if source.Value
                        if obj.btn4x.Value
                            obj.btn4x.Value = 0;
                        elseif obj.btn8x.Value
                            obj.btn8x.Value = 0;
                        end
                        obj.playbackspeed = 2;
                    end
                case '4x'
                    if source.Value
                        if obj.btn2x.Value
                            obj.btn2x.Value = 0;
                        elseif obj.btn8x.Value
                            obj.btn8x.Value = 0;
                        end
                        obj.playbackspeed = 4;
                    end
                case '8x'
                    if source.Value
                        if obj.btn2x.Value
                            obj.btn2x.Value = 0;
                        elseif obj.btn4x.Value
                            obj.btn4x.Value = 0;
                        end
                        obj.playbackspeed = 8;
                    end                    
            end
            
        end
        
                   
        function color = getRoiColor(obj, roi)
        % Return a color for the roi based on which group it belongs to.
            for i = 1:length(obj.roi_colors)
                if contains(obj.roi_groups{i},roi.group)
                    color = obj.roi_colors{i};
                    return
                end
            end
            color = 'red';
        end
            
        
        function obj = callback_popup_roi_group(obj, source, ~)
            % TODO:
%             % Change roiclass of roi if popupmenu is changed.
%             if ~isempty(obj.selected_rois)
%                 for i = obj.selected_rois
%                     groupName = source.String{source.Value};
%                     obj.roi_array(i).Group = groupName;
%                     obj.roi_array(i).Tag = [obj.roi_groups{source.Value}, num2str(obj.roi_array(i).ID,'%03d')];
%                     
%                 end
%                 obj.update_roi_tags_and_list_box();
%          
%             end
                        
        end
        
        
        function obj = callback_btn_single_channel(obj, source, ~)
        % Button callback to change display between showing single or
        % multiple channels.
        
            switch source.String
                case 'Show Single Channel'
                    set(obj.btn_show_single_channel, 'String', 'Show All Channels')
                    obj.channelDisplayMode = 'single';
                    colormap(obj.axStackDisplay, gray(256))
                    set(obj.axStackDisplay, 'CLim', [0, 255])
                    
                case 'Show All Channels'
                    set(obj.btn_show_single_channel, 'String', 'Show Single Channel')
                    obj.channelDisplayMode = 'multi';
                otherwise
                    return
            end
            
            obj.update_image_display();
        
        end
        
        
        function callback_popup_cycle(obj, source, ~)
            obj.change_cycle_and_channel(source.Value,obj.channel);
        end
        
        function change_cycle_and_channel(obj, cycle, channel)
            if isempty(obj.stack); return; end
            if cycle > obj.stack.cycles; return; end
            if channel > obj.stack.channels; return; end
            
            obj.channel = channel;
            obj.cycle = cycle;
            
            obj.popup_channel.Value = channel;
            obj.popup_cycle.Value = cycle;
            
            % Update image
            obj.update_image_display();
            
            if isempty(obj.roi_array)
                return
            end
            
            % Reset selection of rois.
            obj.update_roi_visibility();
            obj.deselectRois();
            obj.resetSignalPlot();
            obj.update_listbox();
        end
        
        
        function callback_popup_channel(obj, source, ~)
            % Callback for changing the current channel
            obj.change_cycle_and_channel(obj.cycle, source.Value);
        end
        
        function update_roi_visibility(obj)
            % Change visibility of rois.
            ind_channel = [obj.roi_array.channel] == obj.channel;
            ind_cycle = [obj.roi_array.cycle] == obj.cycle;
            ind = ind_channel & ind_cycle;
            
            if obj.toggle_show_rois.Value == 0
                ind(:) = 0;
            end
            
            set(obj.roi_plot_handles(ind),'Visible','on')
            set(obj.roi_plot_handles(~ind),'Visible','off')
            
            set(obj.roi_text_handles,'Visible','off');
            if strcmp(obj.btn_show_tags.String, 'Hide Tags')
                set(obj.roi_text_handles(ind),'Visible','on');
            end
        end
        
    end
end

